#!/usr/bin/perl
#
###################################################################################
#	
#	LASTPIECE_P (Local Alignment-STate Probabilities that Insertion-type and dEletion-type gaps Co-Exist, type P):
#	    A package of programs to compute the probabilities of gapped segments in each of which an insertion-type gap and a deletion-type gap co-exist, (which were referred to as case-(iv) gapped segments by Ezawa (2016a?)),
#	    under a stochastic model of sequence evolution with biologically realistic insertions/deletions.
#	   (Written almost exclusively in Perl.)
#
#	Version 0.3: Copyright (C) 2020 Kiyoshi Ezawa
#
#    This package is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This package is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License, "GNU_GPL.txt",
#    along with this package.  If not, see <http://www.gnu.org/licenses/>.
#
#    The author can be contacted by e-mailing to <kezawa dot ezawa3 at gmail dot com>
#    (replace " dot " and " at " with "." and "@", respectively).
#
###################################################################################
#
# compute_raw_frequencies_case_ii_theory.pl:
#	This script computes & outputs RAW expected frequencies of case ii gapped segments
#	from the occurrence probabilities (of the configurations) predicted with our theory.
#	The script was created by modifying 'compute_raw_frequencies_case_iv_theory.pl'.
#	[CREATED on 2020/04/12.]
#
# compute_raw_frequencies_case_iv_theory.pl:
#	This script computes & outputs RAW expected frequencies
#	from the occurrence probabilities (of the configurations) predicted with our theory.
#	The script was created by DISABLING the function to compute cumulative frequencies
#	in 'compute_dbl_cumulative_frequencies_theory.pl'.
#	[CREATED on 2020/04/11.]
#
# compute_dbl_cumulative_frequencies_theory.pl:
#	This script computes & outputs double-cumulative frequencies
#	from the occurrence probabilities (of the configurations) predicted with our theory.
#	The script was created by modifying the basic script, 'compute_dbl_cumulative_frequencies.pl'.
#	[CREATION STARTED on 2020/04/10, ENDED on 2020/04/11.]
#
# compute_dbl_cumulative_frequencies.pl:
#	This Perl script reads in an input file recording frequencies of configurations,
#	each of which is specified by two non-negative integers,
#	and computes & outputs double-cumulative frequencies.
#	[CREATION STARTED on 2020/04/08, ENDED on 2020/04/dd.]
#


use strict;
use warnings;

use MyFile_manipu qw(
	open_input_file2
	open_output_file
	list_sessions_ext
	tfm_labels2indices
	tfm_labels2indices_gen
	);

use MyPWA_Analyses qw(
	readin_1D_array
	cmpt_dbl_cumulative_2Darray
	);
	# REPLACED 'readin_2D_array'  with 'readin_1D_array' on 2020/04/12, in creation of 'compute_raw_frequencies_case_ii_theory.pl'. (US) #


sub create_dir_rw (\$); # IMPORTED from 'lastpiece.alpha.pl' on 2020/04/09. #


$| = 1; # Flush the buffer for standard output as soon as it receives data.

my $program_path = $0;
my ($program_name) = ($program_path =~ m!([^/]+)$!);


	# Controlling the level of test outputs. #
	# (0: silent; 1: very quiet; 2: quiet; 3: medium; 4: verbose; 5: very verbose) 
our $TEST_LEVEL = 3;


		### Reader-Specified Parameters. ###

our @TIME_LAPSES = (0.2, 0.5, 1); # A list of time-lapses to be examined. #


				# (Modified in creation of 'compute_raw_frequencies_case_ii_theory.pl' (on 2020/04/12). (GL) #
##our $LABEL_CORNER = '#{sites}_Init:#{sites}_Fin\Time_Lapse'; # The label at the top-left corner of the input table.
our $LABEL_CORNER = '#{Anc_sites}\Time_Lapse'; # The label at the top-left corner of the input table.
				# End of "(Modified in creation of 'compute_raw_frequencies_case_ii_theory.pl' (on 2020/04/12). (GL)" #

##our @RELV_FLD_LABELS = ('0.2', '0.5', '1');  # A list of labels of relevant fields (except the row-label, i.e., the 0th field). #
our @RELV_FLD_LABELS = @TIME_LAPSES; # A list of labels of relevant fields (except the row-label, i.e., the 0th field). #

our $CT_RLV_FLDS = @RELV_FLD_LABELS;

our $FLD_SEP = qr/\t/; # The Pre-Compiled Regular Expression of the Field separator used for the header. #

our $STR_UNDEF = 'n/a'; # String that represents 'undefined' (or 'not applicable'). #

our $MAX_X = 100; # Maximum value for the $x (i.e., the 1st index for the 2D array) to be examined.
## our $MAX_Y = 100; # Maximum value for the $y (i.e., the 2nd index for the 2D array) to be examined. # This is NO LONGER NECESSARY in the script, 'compute_raw_frequencies_case_ii_theory.pl'. #

=For TESTING the script. (Test1)
  $MAX_X = 10; # Maximum value for the $x (i.e., the 1st index for the 2D array) to be examined.
  $MAX_Y = 10; # Maximum value for the $y (i.e., the 2nd index for the 2D array) to be examined.
=cut # End of "For TESTING the script. (Test1)"


			## ADDED in creation of 'compute_dbl_cumulative_frequencies_theory.pl' (on 2020/04/10). (GL) ##

our $TOTCT_ANC_SITES = 10000 * 400 * 250; # Total number of ancestral sites simulated. #

our ($INS_RATE, $DEL_RATE) = (0.1, 0.1); # Insertion and Deletion Rates (/site/unit time). #
our $DELTA_DEL = 5.32316590760864; # = Sum_{$l = 1}^{$CUTOFF_LEN_DEL} ($l-1) * $FREQ_DEL_LEN[$l].


		## Parameters Determined from the User-Specified Parameters. ##

 	# Coefficient for the Constant Factors.
	# (NOTE: the +2 * $DEL_RATE is necessary to prevent the deletion of the flanking ancestral sites;
	# the + $INS_RATE is necessary to prevent superfluous insertions in between the flanking preserved ancestral sites (PASs); we don't care whether or not insertions occur outside of the flanking PASs) #
our $DELTA_DEL_P =  ($DELTA_DEL + 2) * $DEL_RATE + $INS_RATE ;

our @CONST_FACS = (); # {The $f th Element} = {Constant Factor for the predicted probabilities at time-lapse = $RELV_FLD_LABELS[$f]}.
for (my $f=0; $f <$CT_RLV_FLDS; $f++) {
	$CONST_FACS[$f] = exp (- $RELV_FLD_LABELS[$f] * $DELTA_DEL_P) ;
}

			## END of "ADDED in creation of 'compute_dbl_cumulative_frequencies_theory.pl' (on 2020/04/10). (GL)" ##


		### Input & Output files. ###

			### RESTART FROM HERE ###

			# (MODIFIED in creation of 'compute_dbl_cumulative_frequencies_theory.pl' on 2020/04/10.) (I/O-F) #

				# (Modified in creation of 'compute_raw_frequencies_case_ii_theory.pl' (on 2020/04/12). (I/O-F1) #

my $infile_name = 'case_ii.txt';

my $indir = "../../../../Outputs_LASTPIECE.ver0.3/Example_Outputs_UBC150_lite/Basic/muP_case_i2iii/";

=For the predecessor, 'compute_raw_frequencies_case_iv_theory.pl'. (I/O-F)
#my $infile_name = 'case_iv_2nd.txt';
#my $infile_name = 'case_iv_3rd.txt';
##my $infile_name = 'case_iv_2nd_plus_3rd.txt';

#my $indir = "../../../Outputs_LASTPIECE.ver0.3/Example_Outputs_lite/Summations/PreFinal";
##my $indir = "../../../Outputs_LASTPIECE.ver0.3/Example_Outputs_UBC150_lite/Summations/PreFinal";
my $indir = "../../../../Outputs_LASTPIECE.ver0.3/Example_Outputs_UBC150_lite/Summations/PreFinal";
=cut # END of "For the predecessor, 'compute_raw_frequencies_case_iv_theory.pl'. (I/O-F)"

				# END of "(Modified in creation of 'compute_raw_frequencies_case_ii_theory.pl' (on 2020/04/12). (I/O-F1)" #


my $infile = "$indir/$infile_name";


			# (Revised on 2020/04/11.) (I/O-F1) #
####my $topoutdir = '.';
####my $topoutdir = "X_le_${MAX_X}_Y_le_$MAX_Y";
###my $topoutdir = "X_le_${MAX_X}_Y_le_${MAX_Y}_UBC100";
##my $topoutdir = "X_le_${MAX_X}_Y_le_${MAX_Y}_UBC150";
my $topoutdir = "X_le_${MAX_X}_UBC150"; # Modified in creation of 'compute_raw_frequencies_case_ii_theory.pl' (on 2020/04/12). (I/O-F2) #
			# End of "(Revised on 2020/04/11.) (I/O-F1)" #

##   $topoutdir = 'TESTs/Test1'; # For TESTING the script (Test1). #


my $outfile_name = $infile_name;

			# (Revised on 2020/04/11.) (I/O-F2) #
####my $outdir1 = "$topoutdir/Cum_Probs_X_le${MAX_X}_Y_le$MAX_Y";
####my $outdir2 = "$topoutdir/Cum_ExpFreqs_X_le${MAX_X}_Y_le$MAX_Y";
##my $outdir1 = "$topoutdir/Cum_Probs";
##my $outdir2 = "$topoutdir/Cum_ExpFreqs";
my $outdir1 = "$topoutdir/Raw_Probs";
my $outdir2 = "$topoutdir/Raw_ExpFreqs";
			# End of "(Revised on 2020/04/11.) (I/O-F2)" #

my $outfile1 = "$outdir1/$outfile_name";
my $outfile2 = "$outdir2/$outfile_name";

=For the predecessor, 'compute_dbl_cumulative_frequencies.pl'.

my $infile = '../ancdec_blen0.5/total_frqs_gseg_configs_bulk.txt';

(my $outfile = $infile) =~ s/^(\.\.\/)+//;
$outfile =~ s/frqs/cuml_frqs/;

   $outfile = 'TESTs/Test1/total_cuml_frqs_gseg_configs_bulk.txt'; # For TESTING the script (Test1). #

(my $outdir = $outfile) =~ s/[^\/]+$//;
if ($outdir =~ m/\S/) {
	$outdir =~ s/\/$//;
} else {
	$outdir = '.';
}

if ($TEST_LEVEL >0) {
	print STDOUT "\n[INFORMATION '$program_name'] Output directory: $outdir .\n\n";
}
=cut # END of "For the predecessor, 'compute_dbl_cumulative_frequencies.pl'."

			# END of "(MODIFIED in creation of 'compute_dbl_cumulative_frequencies_theory.pl' on 2020/04/10.) (I/O-F)" #


		### Check the input file. ###

unless (( -f $infile ) and ( -r $infile )) {
    print STDOUT "[ERROR '$program_name'] CanNOT FIND a readable INPUT FILE, '$infile': $? !\n\n";
    print STDERR "[ERROR '$program_name'] $! \n";
    die ;
}

		## Create the output directories, if they do not exist yet. ##

foreach my $outdir ($outdir1, $outdir2) { # ADDED in creation of 'compute_dbl_cumulative_frequencies_theory.pl' (on 2020/04/10). #
    unless ( create_dir_rw ($outdir) ) { 
        print STDOUT "[ERROR '$program_name'] The subroutine, 'create_dir_rw', FAILED on the  output directory, \"$outdir\" : $?\n\n";
        print STDERR "[ERROR '$program_name'] $!\n";
        die ;
    }
} # End of the foreach ADDED in creation of 'compute_dbl_cumulative_frequencies_theory.pl' (on 2020/04/10). #

		### Declare the Beginning of the Script ###

print STDOUT "\n<<< Program name: $program_name (path: $program_path) >>>\n\n";


			# (Modified in creation of 'compute_dbl_cumulative_frequencies_theory.pl' on 2020/04/10.) (DC) #
##print STDOUT "Input file: $infile .\n\n";
print STDOUT "Input file: {in_dir}/$infile_name ,\n";
print STDOUT "            where {in_dir} = $indir .\n\n";

##print STDOUT "Output file: $outfile .\n\n\n";
print STDOUT "Output files:\n\n";
print STDOUT "  (Probabilities)        {out_dir1}/$outfile_name ,\n";
print STDOUT "      where {out_dir1} = $outdir1 ;\n\n";
print STDOUT "  (Expected Frequencies)  {out_dir2}/$outfile_name ,\n";
print STDOUT "      where {out_dir2} = $outdir2 .\n\n";
print STDOUT "\n";

			# End of "(Modified in creation of 'compute_dbl_cumulative_frequencies_theory.pl' on 2020/04/10.) (DC)" #

print STDOUT "<< Important Parameters: >>\n\n";

print STDOUT '@TIME_LAPSES = (', join (', ', @TIME_LAPSES), ") .\n\n"; # ADDED in creation of 'compute_dbl_cumulative_frequencies_theory.pl' on 2020/04/10. (DC2) #

print STDOUT '$LABEL_CORNER = ', "'$LABEL_CORNER' .\n\n";

print STDOUT '@RELV_FLD_LABELS = (', join (', ', @RELV_FLD_LABELS), ") .\n\n";

print STDOUT '$STR_UNDEF = ', "'$STR_UNDEF' .\n\n";

			# (Modified in creation of 'compute_raw_frequencies_case_ii_theory.pl' on 2020/04/12.) (DC) #
##print STDOUT '($MAX_X, $MAX_Y) = ', "($MAX_X, $MAX_Y) .\n\n\n";
print STDOUT '$MAX_X = ', "$MAX_X .\n\n\n";
			# End of "(Modified in creation of 'compute_raw_frequencies_case_ii_theory.pl' on 2020/04/12.) (DC)" #


			# (ADDED in creation of 'compute_dbl_cumulative_frequencies_theory.pl' on 2020/04/10.) (DC1) #

print STDOUT '$TOTCT_ANC_SITES = ', $TOTCT_ANC_SITES, " .\n\n";

print STDOUT '($INS_RATE, $DEL_RATE) = ', "($INS_RATE, $DEL_RATE) .\n\n";

print STDOUT '$DELTA_DEL = ', $DELTA_DEL, " .\n\n";


print STDOUT '$DELTA_DEL_P (= ($DELTA_DEL + 2) * $DEL_RATE + $INS_RATE) = ', $DELTA_DEL_P, " .\n\n";

print STDOUT '@CONST_FACS = (', join (', ', @CONST_FACS), ") .\n\n";

			# END of "(ADDED in creation of 'compute_dbl_cumulative_frequencies_theory.pl' on 2020/04/10.) (DC1)" #



print STDOUT "\n<<< (0) Find Indexes Corresponding to the Field Labels in the input file ('$infile'). >>>\n\n";

local $/ = "\n"; # Set the input-record-separator to be the newline character. #

my $fh_in0 = open_input_file2 ($infile);
unless (defined $fh_in0) {
	print STDOUT "[ERROR '$program_name'] The subroutine, 'open_input_file2', FAILED to open a file-handle for the file \"$infile\": $? !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}

my $len_lcorner = length ($LABEL_CORNER);


my $tbl_header;
while (my $line0 = <$fh_in0>) {
    my $sbstr = substr ($line0, 0, $len_lcorner) ;
    if ($LABEL_CORNER eq $sbstr) {
	$tbl_header = $line0;
	last;
    }
}
close ($fh_in0);

unless (defined $tbl_header) {
	print STDOUT "[ERROR '$program_name'] The input file \"$infile\", has NO LINES starting with '$LABEL_CORNER': $? !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}
$tbl_header =~ s/\s+$//;

if ($TEST_LEVEL >1) { # (0a) #
	print STDOUT "\n";
	print STDOUT "[INFORMATION '$program_name'] Retrieved the Table Header from the input file \"$infile\":\n\n";
	print STDOUT "[INFORMATION '$program_name'] \"$tbl_header\" .\n";
	print STDOUT "\n\n";

} # End of "if ($TEST_LEVEL >1) { # (0a) # ...}"

my $indexes_relv_flds = tfm_labels2indices_gen (@RELV_FLD_LABELS, $tbl_header, $FLD_SEP);
unless (defined $indexes_relv_flds) {
	print STDOUT "[ERROR '$program_name'] The input file \"$infile\", has NO LINES starting with '$LABEL_CORNER': $? !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}

if ($TEST_LEVEL >1) { # (0b) #
	print STDOUT "\n[ Output of 'tfm_labels2indices_gen': ]\n\n";
	print STDOUT join ("\t", "'Label'", 'Index'), "\n\n";

	for (my $f=0; $f <$CT_RLV_FLDS; $f++) {
	    my $label = $RELV_FLD_LABELS[$f];
	    my $indx = $indexes_relv_flds->[$f];
	    print STDOUT join ("\t", "'$label'", $indx), "\n";
	}
	print STDOUT "\n\n";

} # End of "if ($TEST_LEVEL >1) { # (0b) # ...}"
 

####print STDOUT "\n\n<<< (1) Read in Relevant 2D arrays & Compute Their Double-Cumulative Counterparts ... >>>\n\n";
##print STDOUT "\n\n<<< (1) Read in Relevant 2D arrays & Store them into a 3D array ... >>>\n\n";
print STDOUT "\n\n<<< (1) Read in Relevant 1D arrays & Store them into a 2D array ... >>>\n\n";

####my @set_dcum_arrays_2D = ();
##my @set_raw_arrays_2D = ();
my @set_raw_arrays_1D = ();

for (my $f=0; $f <$CT_RLV_FLDS; $f++) { # 1st Main Outer for-loop (over the relevant fields). #

	my $label_fld = $RELV_FLD_LABELS[$f];
	my $indx_fld = $indexes_relv_flds->[$f];

	print STDOUT "< The $f th field (label = '$label_fld', index = $indx_fld) >\n\n";
	
	my $fh_in = open_input_file2 ($infile);
	unless (defined $fh_in) {
		print STDOUT "[ERROR '$program_name'] The subroutine, 'open_input_file2', FAILED to open a file-handle for the file \"$infile\": $? !\n";
		print STDERR "[ERROR '$program_name'] $!\n";
		die ;
	}

	##my $len_lcorner = length ($LABEL_CORNER);

	my $dummy_str_end ;

			# MAJOR MODIFICATION in creation of 'compute_raw_frequencies_case_ii_theory.pl' (on 2020/04/12). (1st Main a) #
	my $array_1D = readin_1D_array (*${$fh_in}, $indx_fld, $STR_UNDEF, $dummy_str_end);
	unless (defined $array_1D) {
		print STDOUT "[ERROR '$program_name'] The subroutine, 'readin_1D_array', FAILED to retrieve the field of '$label_fld' in \"$infile\": $? !\n";
		print STDERR "[ERROR '$program_name'] $!\n";
		die ;
	}
=For the predecessor, 'compute_raw_frequencies_case_iv_theory.pl'. (1st Main a)
	my $array_2D = readin_2D_array (*${$fh_in}, $indx_fld, $STR_UNDEF, $dummy_str_end);
	unless (defined $array_2D) {
		print STDOUT "[ERROR '$program_name'] The subroutine, 'readin_2D_array', FAILED to retrieve the field of '$label_fld' in \"$infile\": $? !\n";
		print STDERR "[ERROR '$program_name'] $!\n";
		die ;
	}
=cut # END of "For the predecessor, 'compute_raw_frequencies_case_iv_theory.pl'. (1st Main a)"
			# END of "MAJOR MODIFICATION in creation of 'compute_raw_frequencies_case_ii_theory.pl' (on 2020/04/12). (1st Main a)" #

	close ($fh_in);

			# MODIFIED in creation of 'compute_raw_frequencies_case_iv_theory.pl' (on 2020/04/11). (1) #
			# => Further MODIFIED in creation of 'compute_raw_frequencies_case_ii_theory.pl' (on 2020/04/12). (1st Main b) #

	##$set_raw_arrays_2D[$f] = $array_2D;
	$set_raw_arrays_1D[$f] = $array_1D;

=DISABLED in creation of 'compute_raw_frequencies_theory.pl'.
	my $dcum_array_2D = cmpt_dbl_cumulative_2Darray (@{$array_2D}, $MAX_X, $MAX_Y);
	unless (defined $dcum_array_2D) {
		print STDOUT "[ERROR '$program_name'] The subroutine, 'cmpt_dbl_cumulative_2Darray', FAILED on the 2D array constructed from the field of '$label_fld' in \"$infile\": $? !\n";
		print STDERR "[ERROR '$program_name'] $!\n";
		die ;
	}

	$set_dcum_arrays_2D[$f] = $dcum_array_2D;
=cut # END of "DISABLED in creation of 'compute_raw_frequencies_theory.pl'."

			# END of "MODIFIED in creation of 'compute_raw_frequencies_case_iv_theory.pl' (on 2020/04/11). (1) #
			# => Further MODIFIED in creation of 'compute_raw_frequencies_case_ii_theory.pl' (on 2020/04/12). (1st Main b)" #

} # End of the 1st Main outer for-loop (over the relevant fields). #



			### MAJOR MODIFICATION in creation of 'compute_dbl_cumulative_frequencies_theory.pl' (on 2020/04/10). (Main 2) ###

##print STDOUT "\n\n<<< (2) Output the Computed Double-Cumulative 2D arrays to a single file (= \"$outfile\") ... >>>\n\n";
##print STDOUT "\n\n<<< (2) Output the Computed Double-Cumulative Probabilities & Expected Frequencies, each to a single file (= \"$outfile1\" or \"$outfile2\") ... >>>\n\n";
print STDOUT "\n\n<<< (2) Output the Computed RAW Probabilities & Expected Frequencies, each to a single file (= \"$outfile1\" or \"$outfile2\") ... >>>\n\n";

my $fh_out1 = open_output_file ($outfile1, 0);
unless (defined $fh_out1) {
	print STDOUT "[ERROR '$program_name'] The subroutine 'open_output_file' FAILED to open a file-handle for the file \"$outfile1\": $? !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}
my $fh_out2 = open_output_file ($outfile2, 0);
unless (defined $fh_out2) {
	print STDOUT "[ERROR '$program_name'] The subroutine 'open_output_file' FAILED to open a file-handle for the file \"$outfile2\": $? !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}



print $fh_out1 join ("\t", $LABEL_CORNER, @RELV_FLD_LABELS), "\n\n";
print $fh_out2 join ("\t", $LABEL_CORNER, @RELV_FLD_LABELS), "\n\n";

####for (my $x=0; $x <$MAX_X; $x++) { # 2nd Main outer for-loop (over $x). 
##for (my $x=0; $x <= $MAX_X; $x++) { # 2nd Main outer for-loop (over $x). # (CORRECTED on 2020/04/10.) (2nd Main outer)
for (my $x=1; $x <= $MAX_X; $x++) { # 2nd Main outer for-loop (over $x). # (=> Further Modified on 2020/04/11.) (2nd Main outer)

		# MAJOR MODIFICATION in creation of 'compute_raw_frequencies_case_ii_theory.pl' (on 2020/04/12). (2nd Main) #

	my @out_flds1 = my @out_flds2 = ();

	for (my $f=0; $f <$CT_RLV_FLDS; $f++) { # The inner for-loop (over relevant fields). #

	     my $raw_prb = $set_raw_arrays_1D[$f]->[$x] * $CONST_FACS[$f] ;
	     $out_flds1[$f] = $raw_prb;

	     my $raw_exp_frq = $raw_prb * $TOTCT_ANC_SITES ;
	     $out_flds2[$f] = $raw_exp_frq;

	} # End of The inner for-loop (over relevant fields). #

	print $fh_out1 join ("\t", $x, @out_flds1), "\n";
	print $fh_out2 join ("\t", $x, @out_flds2), "\n";

=For the predecessor, 'compute_raw_frequencies_case_iv_theory.pl'. (2nd Main)
			# MODIFIED in creation of 'compute_raw_frequencies_case_iv_theory.pl' (on 2020/04/11). (2a) #

	##my @set_rows_dcum = ();
	##for (my $f=0; $f <$CT_RLV_FLDS; $f++) { $set_rows_dcum[$f] = $set_dcum_arrays_2D[$f]->[$x]; } # The 1st middle for-loop (over relevant fields). #

	my @set_rows_raw = ();
	for (my $f=0; $f <$CT_RLV_FLDS; $f++) { $set_rows_raw[$f] = $set_raw_arrays_2D[$f]->[$x]; }

			# END of "MODIFIED in creation of 'compute_raw_frequencies_case_iv_theory.pl' (on 2020/04/11). (2a)" #


	####for (my $y=0; $y <$MAX_Y; $y++) { # The 2nd Middle for-loop (over $y). #
	##for (my $y=0; $y <= $MAX_Y; $y++) { # The 2nd Middle for-loop (over $y). # (CORRECTED on 2020/04/10.) (2nd Middle) #
	for (my $y=1; $y <= $MAX_Y; $y++) { # The 2nd Middle for-loop (over $y). # (=> Further Modified on 2020/04/10.) (2nd Middle) #

	    my @out_flds1 = my @out_flds2 = ();

			# MODIFIED in creation of 'compute_raw_frequencies_case_iv_theory.pl' (on 2020/04/11). (2a) #

	    for (my $f=0; $f <$CT_RLV_FLDS; $f++) { # The inner for-loop (over relevant fields). #

		my $raw_prb = $set_rows_raw[$f]->[$y] * $CONST_FACS[$f] ;
		$out_flds1[$f] = $raw_prb;


		my $raw_exp_frq = $raw_prb * $TOTCT_ANC_SITES ;
		$out_flds2[$f] = $raw_exp_frq;

	    } # End of The inner for-loop (over relevant fields). #

##=DISABLED in creation of 'compute_raw_frequencies_case_iv_theory.pl'. (2)
##	    ##for (my $f=0; $f <$CT_RLV_FLDS; $f++) {$out_flds[$f] = $set_rows_dcum[$f]->[$y]; } # The inner for-loop (over relevant fields). #
##	    for (my $f=0; $f <$CT_RLV_FLDS; $f++) { # The inner for-loop (over relevant fields). #
##
##		my $dcum_prb = $set_rows_dcum[$f]->[$y] * $CONST_FACS[$f] ;
##		$out_flds1[$f] = $dcum_prb;
##
##
##		my $dcum_exp_frq = $dcum_prb * $TOTCT_ANC_SITES ;
##		$out_flds2[$f] = $dcum_exp_frq;
##
##	    } # End of The inner for-loop (over relevant fields). #
##=cut # END of "DISABLED in creation of 'compute_raw_frequencies_case_iv_theory.pl'. (2)"

			# END of "MODIFIED in creation of 'compute_raw_frequencies_case_iv_theory.pl' (on 2020/04/11). (2a)" #

			# Fixed the BUGs (on 2020/04/10). #
	    ##print $fh_out1 join ("$x:$y", @out_flds1), "\n";
	    ##print $fh_out2 join ("$x:$y", @out_flds2), "\n";
	    print $fh_out1 join ("\t", "$x:$y", @out_flds1), "\n";
	    print $fh_out2 join ("\t", "$x:$y", @out_flds2), "\n";
			# End of "Fixed the BUGs (on 2020/04/10)." #

	} # End of the 2nd Middle for-loop (over $y). #

=cut # END of "For the predecessor, 'compute_raw_frequencies_case_iv_theory.pl'. (2nd Main)"
		# END of "MAJOR MODIFICATION in creation of 'compute_raw_frequencies_case_ii_theory.pl' (on 2020/04/12). (2nd Main)" #

}  # 2nd Main outer for-loop (over $x). #

close ($fh_out1);
close ($fh_out2);

			### END of "MAJOR MODIFICATION in creation of 'compute_dbl_cumulative_frequencies_theory.pl' (on 2020/04/10). (Main 2)" ###



	##### DEFINITIONS of the Subroutines. #####


# create_dir_rw:
#	This subroutine receives the path to a directory,
#	examines (1) whether the directory exists.
#	Otherwise, it creates the directory.
#	Then, it examines (2) whether the directory is readable & writable.
#	Otherwise, it changes the permission mode of the directory to make it readable & writable.
#	[CREATED on 2020/01/29; IMPORTED from 'lastpiece.alpha.pl' on 2020/04/09.]
#
sub create_dir_rw (\$) {

	my ($rf_dir_path) = @_;
	my $dir_path = ${$rf_dir_path};

	unless ( -d $dir_path ) {
	   unless (0 == system ("mkdir -p $dir_path") ) {
		print STDOUT "[ERROR '$program_name'] FAILED to CREATE the OUTPUT DIRECTORY, '$dir_path/': $? !\n\n";
		print STDERR "[ERROR '$program_name'] $!\n";
		return 0 ;
	   }
	}
	unless (( -r $dir_path ) and ( -w $dir_path )) {
	    unless (0 == system ("chmod u+rw $dir_path") ) {
		print STDOUT "[ERROR '$program_name'] FAILED to CHANGE permission MODES of the output directory, '$dir_path': $? !\n";
		print STDERR "[ERROR '$program_name'] $!\n";
		return 0 ;
	    }
	}

	return 1;

}	# End of the subroutine, 'create_dir_rw'. #


