#!/usr/bin/perl
#
###################################################################################
#	
#	LASTPIECE_P (Local Alignment-STate Probabilities that Insertion-type and dEletion-type gaps Co-Exist, type P):
#	    A package of programs to compute the probabilities of gapped segments in each of which an insertion-type gap and a deletion-type gap co-exist, (which were referred to as case-(iv) gapped segments by Ezawa (2016a?)),
#	    under a stochastic model of sequence evolution with biologically realistic insertions/deletions.
#	   (Written almost exclusively in Perl.)
#
#	Version 0.3: Copyright (C) 2020 Kiyoshi Ezawa
#
#    This package is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This package is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License, "GNU_GPL.txt",
#    along with this package.  If not, see <http://www.gnu.org/licenses/>.
#
#    The author can be contacted by e-mailing to <kezawa dot ezawa3 at gmail dot com>
#    (replace " dot " and " at " with "." and "@", respectively).
#
###################################################################################
#
# compute_ratios_case_iv_to_iixiii.pl:
#	This script computes & outputs the ratios of
#	case-iv multiplication factors to the products of their counterparts of case-ii and case-iii gapped segments.
#	The script was created by taking advantage of the framework of 'compute_raw_frequencies_case_iv_theory.pl'
#	and by adding the input blocks from 'compute_raw_frequencies_case_ii_theory.pl' & 'compute_raw_frequencies_case_iii_theory.pl'.
#	[CREATED on 2020/04/12.]
#
# compute_raw_frequencies_case_iv_theory.pl:
#	This script computes & outputs RAW expected frequencies
#	from the occurrence probabilities (of the configurations) predicted with our theory.
#	The script was created by DISABLING the function to compute cumulative frequencies
#	in 'compute_dbl_cumulative_frequencies_theory.pl'.
#	[CREATED on 2020/04/11.]
#
# compute_dbl_cumulative_frequencies_theory.pl:
#	This script computes & outputs double-cumulative frequencies
#	from the occurrence probabilities (of the configurations) predicted with our theory.
#	The script was created by modifying the basic script, 'compute_dbl_cumulative_frequencies.pl'.
#	[CREATION STARTED on 2020/04/10, ENDED on 2020/04/11.]
#
# compute_dbl_cumulative_frequencies.pl:
#	This Perl script reads in an input file recording frequencies of configurations,
#	each of which is specified by two non-negative integers,
#	and computes & outputs double-cumulative frequencies.
#	[CREATION STARTED on 2020/04/08, ENDED on 2020/04/dd.]
#


use strict;
use warnings;

use MyFile_manipu qw(
	open_input_file2
	open_output_file
	list_sessions_ext
	tfm_labels2indices
	tfm_labels2indices_gen
	);

use MyPWA_Analyses qw(
	readin_2D_array
	readin_1D_array
	cmpt_dbl_cumulative_2Darray
	);
	# ADDED 'readin_1D_array' on 2020/04/13. #


sub create_dir_rw (\$); # IMPORTED from 'lastpiece.alpha.pl' on 2020/04/09. #


$| = 1; # Flush the buffer for standard output as soon as it receives data.

my $program_path = $0;
my ($program_name) = ($program_path =~ m!([^/]+)$!);


	# Controlling the level of test outputs. #
	# (0: silent; 1: very quiet; 2: quiet; 3: medium; 4: verbose; 5: very verbose) 
our $TEST_LEVEL = 3;


		### Reader-Specified Parameters. ###

our @TIME_LAPSES = (0.2, 0.5, 1); # A list of time-lapses to be examined. #

		# The labels at the top-left corner of the input table.
		# (ENHANCED in creation of 'compute_ratios_case_iv_to_iixiii.pl' (on 2020/04/12).) (GL) #
our $LABEL_CORNER1 = '#{sites}_Init:#{sites}_Fin\Time_Lapse'; # For $infile1.
our $LABEL_CORNER2 = '#{Anc_sites}\Time_Lapse';  # For $infile2.
our $LABEL_CORNER3 = '#{Dec_sites}\Time_Lapse';  # For $infile3.
		# End of "(ENHANCED in creation of 'compute_ratios_case_iv_to_iixiii.pl' (on 2020/04/12).) (GL)" #

##our @RELV_FLD_LABELS = ('0.2', '0.5', '1');  # A list of labels of relevant fields (except the row-label, i.e., the 0th field). #
our @RELV_FLD_LABELS = @TIME_LAPSES; # A list of labels of relevant fields (except the row-label, i.e., the 0th field). #

our $CT_RLV_FLDS = @RELV_FLD_LABELS;

our $FLD_SEP = qr/\t/; # The Pre-Compiled Regular Expression of the Field separator used for the header. #

our $STR_UNDEF = 'n/a'; # String that represents 'undefined' (or 'not applicable'). #

our $MAX_X = 100; # Maximum value for the $x (i.e., the 1st index for the 2D array) to be examined.
our $MAX_Y = 100; # Maximum value for the $y (i.e., the 2nd index for the 2D array) to be examined.

=For TESTING the script. (Test1)
  $MAX_X = 10; # Maximum value for the $x (i.e., the 1st index for the 2D array) to be examined.
  $MAX_Y = 10; # Maximum value for the $y (i.e., the 2nd index for the 2D array) to be examined.
=cut # End of "For TESTING the script. (Test1)"


=For the predecessor, 'compute_raw_frequencies_case_iv_theory.pl'. (GL)
			## ADDED in creation of 'compute_dbl_cumulative_frequencies_theory.pl' (on 2020/04/10). (GL) ##

our $TOTCT_ANC_SITES = 10000 * 400 * 250; # Total number of ancestral sites simulated. #

our ($INS_RATE, $DEL_RATE) = (0.1, 0.1); # Insertion and Deletion Rates (/site/unit time). #
our $DELTA_DEL = 5.32316590760864; # = Sum_{$l = 1}^{$CUTOFF_LEN_DEL} ($l-1) * $FREQ_DEL_LEN[$l].

		## Parameters Determined from the User-Specified Parameters. ##

 	# Coefficient for the Constant Factors.
	# (NOTE: the +2 * $DEL_RATE is necessary to prevent the deletion of the flanking ancestral sites;
	# the + $INS_RATE is necessary to prevent superfluous insertions in between the flanking preserved ancestral sites (PASs); we don't care whether or not insertions occur outside of the flanking PASs) #
our $DELTA_DEL_P =  ($DELTA_DEL + 2) * $DEL_RATE + $INS_RATE ;

our @CONST_FACS = (); # {The $f th Element} = {Constant Factor for the predicted probabilities at time-lapse = $RELV_FLD_LABELS[$f]}.
for (my $f=0; $f <$CT_RLV_FLDS; $f++) {
	$CONST_FACS[$f] = exp (- $RELV_FLD_LABELS[$f] * $DELTA_DEL_P) ;
}

			## END of "ADDED in creation of 'compute_dbl_cumulative_frequencies_theory.pl' (on 2020/04/10). (GL)" ##
=cut # END of "For the predecessor, 'compute_raw_frequencies_case_iv_theory.pl'. (GL)"


		### Input & Output files. ###

			### RESTART FROM HERE (I/O-F) ###

#my $infile_name1 = 'case_iv_2nd.txt';
my $infile_name1 = 'case_iv_3rd.txt';
#my $infile_name1 = 'case_iv_2nd_plus_3rd.txt';

#my $indir1 = "../../../Outputs_LASTPIECE.ver0.3/Example_Outputs_lite/Summations/PreFinal";
my $indir1 = "../../../../Outputs_LASTPIECE.ver0.3/Example_Outputs_UBC150_lite/Summations/PreFinal";

my $infile1 = "$indir1/$infile_name1";

			# (ADDED in creation of 'compute_ratios_case_iv_to_iixiii.pl' on 2020/04/13). (I/O-F2) #

my $indir2 = "../../../../Outputs_LASTPIECE.ver0.3/Example_Outputs_UBC150_lite/Basic/muP_case_i2iii";
my $infile_name2 = 'case_ii.txt';

my $infile2 = "$indir2/$infile_name2";

my $indir3 = $indir2;
my $infile_name3 = 'case_iii.txt';

my $infile3 = "$indir3/$infile_name3";

			# End of "(ADDED in creation of 'compute_ratios_case_iv_to_iixiii.pl' on 2020/04/13). (I/O-F2)" #

##my $topoutdir = '.';
##my $topoutdir = "X_le_${MAX_X}_Y_le_$MAX_Y";
my $topoutdir = "X_le_${MAX_X}_Y_le_${MAX_Y}_UBC150";

##   $topoutdir = 'TESTs/Test1'; # For TESTING the script (Test1). #


my $outfile_name = $infile_name1;


		# MODIFIED in creation of 'compute_ratios_case_iv_to_iixiii.pl' (on 2020/04/12). (I/O-F1) #
my $outdir1 = "$topoutdir/Ratios_muPs";

my $outfile1 = "$outdir1/$outfile_name";

=For the predecessor, 'compute_raw_frequencies_case_iv_theory.pl'. (I/O-F)
			# (Revised on 2020/04/11.) (I/O-F2) #
####my $outdir1 = "$topoutdir/Cum_Probs_X_le${MAX_X}_Y_le$MAX_Y";
####my $outdir2 = "$topoutdir/Cum_ExpFreqs_X_le${MAX_X}_Y_le$MAX_Y";
##my $outdir1 = "$topoutdir/Cum_Probs";
##my $outdir2 = "$topoutdir/Cum_ExpFreqs";
my $outdir1 = "$topoutdir/Raw_Probs";
my $outdir2 = "$topoutdir/Raw_ExpFreqs";
			# End of "(Revised on 2020/04/11.) (I/O-F2)" #

my $outfile1 = "$outdir1/$outfile_name";
my $outfile2 = "$outdir2/$outfile_name";
=cut # END of "For the predecessor, 'compute_raw_frequencies_case_iv_theory.pl'. (I/O-F)"

		# END of "MODIFIED in creation of 'compute_ratios_case_iv_to_iixiii.pl' (on 2020/04/12). (I/O-F1)" #




		### Check the input file. ###

foreach my $infile ($infile1, $infile2, $infile3) { # ADDED in creation of 'compute_ratios_case_iv_to_iixiii.pl' (on 2020/04/12). #
    unless (( -f $infile ) and ( -r $infile )) {
        print STDOUT "[ERROR '$program_name'] CanNOT FIND a readable INPUT FILE, '$infile': $? !\n\n";
        print STDERR "[ERROR '$program_name'] $! \n";
        die ;
    }
} # End of the foreach ADDED in creation of 'compute_ratios_case_iv_to_iixiii.pl' (on 2020/04/12). #


		## Create the output directories, if they do not exist yet. ##

##foreach my $outdir ($outdir1, $outdir2) { # ADDED in creation of 'compute_dbl_cumulative_frequencies_theory.pl' (on 2020/04/10). #
foreach my $outdir ($outdir1) { # => MODIFIED in creation of 'compute_ratios_case_iv_to_iixiii.pl' (on 2020/04/12). #
    unless ( create_dir_rw ($outdir) ) { 
        print STDOUT "[ERROR '$program_name'] The subroutine, 'create_dir_rw', FAILED on the  output directory, \"$outdir\" : $?\n\n";
        print STDERR "[ERROR '$program_name'] $!\n";
        die ;
    }
} # End of the foreach ADDED in creation of 'compute_dbl_cumulative_frequencies_theory.pl' (on 2020/04/10). #

		### Declare the Beginning of the Script ###

print STDOUT "\n<<< Program name: $program_name (path: $program_path) >>>\n\n";


print STDOUT "Input files:\n\n";

print STDOUT "  (multiplication factors of case-iv configurations)  $infile1 ,\n";
print STDOUT "  (multiplication factors of case-ii configurations)  $infile2 ,\n";
print STDOUT "  (multiplication factors of case-iii configurations) $infile3 .\n\n\n";

print STDOUT "Output file:\n";
print STDOUT "  (Ratios of case-iv/[case-ii x case-iii])  {out_dir1}/$outfile_name ,\n";
print STDOUT "      where {out_dir1} = $outdir1 ;\n\n";


=For the predecessor, 'compute_raw_frequencies_case_iv_theory.pl' (DC1)
print STDOUT "Input file: {in_dir}/$infile_name ,\n";
print STDOUT "            where {in_dir} = $indir .\n\n";

print STDOUT "Output files:\n\n";
print STDOUT "  (Probabilities)        {out_dir1}/$outfile_name ,\n";
print STDOUT "      where {out_dir1} = $outdir1 ;\n\n";
print STDOUT "  (Expected Frequencies)  {out_dir2}/$outfile_name ,\n";
print STDOUT "      where {out_dir2} = $outdir2 .\n\n";
print STDOUT "\n";
=cut # END of "For the predecessor, 'compute_raw_frequencies_case_iv_theory.pl' (DC1)"


print STDOUT "<< Important Parameters: >>\n\n";

print STDOUT '@TIME_LAPSES = (', join (', ', @TIME_LAPSES), ") .\n\n"; # ADDED in creation of 'compute_dbl_cumulative_frequencies_theory.pl' on 2020/04/10. (DC2) #

print STDOUT '$LABEL_CORNER1 = ', "'$LABEL_CORNER1' ,\n";
print STDOUT '$LABEL_CORNER2 = ', "'$LABEL_CORNER2' ,\n";
print STDOUT '$LABEL_CORNER3 = ', "'$LABEL_CORNER3' .\n\n";

print STDOUT '@RELV_FLD_LABELS = (', join (', ', @RELV_FLD_LABELS), ") .\n\n";

print STDOUT '$STR_UNDEF = ', "'$STR_UNDEF' .\n\n";

print STDOUT '($MAX_X, $MAX_Y) = ', "($MAX_X, $MAX_Y) .\n\n\n";


=For the predecessor, 'compute_raw_frequencies_case_iv_theory.pl' (DC2)
			# (ADDED in creation of 'compute_dbl_cumulative_frequencies_theory.pl' on 2020/04/10.) (DC1) #

print STDOUT '$TOTCT_ANC_SITES = ', $TOTCT_ANC_SITES, " .\n\n";

print STDOUT '($INS_RATE, $DEL_RATE) = ', "($INS_RATE, $DEL_RATE) .\n\n";

print STDOUT '$DELTA_DEL = ', $DELTA_DEL, " .\n\n";


print STDOUT '$DELTA_DEL_P (= ($DELTA_DEL + 2) * $DEL_RATE + $INS_RATE) = ', $DELTA_DEL_P, " .\n\n";

print STDOUT '@CONST_FACS = (', join (', ', @CONST_FACS), ") .\n\n";

			# END of "(ADDED in creation of 'compute_dbl_cumulative_frequencies_theory.pl' on 2020/04/10.) (DC1)" #
=cut # END of "For the predecessor, 'compute_raw_frequencies_case_iv_theory.pl' (DC2)"


local $/ = "\n"; # Set the input-record-separator to be the newline character. #

print STDOUT "\n<<< (0-1) Find Indexes Corresponding to the Field Labels in the input file ('$infile1'). >>>\n\n";

my $fh_in01 = open_input_file2 ($infile1);
unless (defined $fh_in01) {
	print STDOUT "[ERROR '$program_name'] The subroutine, 'open_input_file2', FAILED to open a file-handle for the file \"$infile1\": $? !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}

my $len_lcorner1 = length ($LABEL_CORNER1);


my $tbl_header1;
while (my $line0 = <$fh_in01>) {
    my $sbstr = substr ($line0, 0, $len_lcorner1) ;
    if ($LABEL_CORNER1 eq $sbstr) {
	$tbl_header1 = $line0;
	last;
    }
}
close ($fh_in01);

unless (defined $tbl_header1) {
	print STDOUT "[ERROR '$program_name'] The input file \"$infile1\", has NO LINES starting with '$LABEL_CORNER1': $? !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}
$tbl_header1 =~ s/\s+$//;

if ($TEST_LEVEL >1) { # (0a-1) #
	print STDOUT "\n";
	print STDOUT "[INFORMATION '$program_name'] Retrieved the Table Header from the input file \"$infile1\":\n\n";
	print STDOUT "[INFORMATION '$program_name'] \"$tbl_header1\" .\n";
	print STDOUT "\n\n";

} # End of "if ($TEST_LEVEL >1) { # (0a-1) # ...}"

my $indexes_relv_flds1 = tfm_labels2indices_gen (@RELV_FLD_LABELS, $tbl_header1, $FLD_SEP);
unless (defined $indexes_relv_flds1) {
	print STDOUT "[ERROR '$program_name'] The subroutine, 'tfm_labels2indices_gen', FAILED on the header of the input file \"$infile1\": $? !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}

if ($TEST_LEVEL >1) { # (0b-1) #
	print STDOUT "\n[ Output of 'tfm_labels2indices_gen' (on the 1st input file): ]\n\n";
	print STDOUT join ("\t", "'Label'", 'Index1'), "\n\n";

	for (my $f=0; $f <$CT_RLV_FLDS; $f++) {
	    my $label = $RELV_FLD_LABELS[$f];
	    my $indx = $indexes_relv_flds1->[$f];
	    print STDOUT join ("\t", "'$label'", $indx), "\n";
	}
	print STDOUT "\n\n";

} # End of "if ($TEST_LEVEL >1) { # (0b-1) # ...}"
 

			### COPIED (w/ modifications) the (0-1) above, in creation of 'compute_ratios_case_iv_to_iixiii.pl' (on 2020/04/12). (0-2 & 0-3) ###

print STDOUT "\n<<< (0-2) Find Indexes Corresponding to the Field Labels in the input file ('$infile2'). >>>\n\n";

my $fh_in02 = open_input_file2 ($infile2);
unless (defined $fh_in02) {
	print STDOUT "[ERROR '$program_name'] The subroutine, 'open_input_file2', FAILED to open a file-handle for the file \"$infile2\": $? !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}

my $len_lcorner2 = length ($LABEL_CORNER2);


my $tbl_header2;
while (my $line0 = <$fh_in02>) {
    my $sbstr = substr ($line0, 0, $len_lcorner2) ;
    if ($LABEL_CORNER2 eq $sbstr) {
	$tbl_header2 = $line0;
	last;
    }
}
close ($fh_in02);

unless (defined $tbl_header2) {
	print STDOUT "[ERROR '$program_name'] The input file \"$infile2\", has NO LINES starting with '$LABEL_CORNER2': $? !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}
$tbl_header2 =~ s/\s+$//;

if ($TEST_LEVEL >1) { # (0a-2) #
	print STDOUT "\n";
	print STDOUT "[INFORMATION '$program_name'] Retrieved the Table Header from the input file \"$infile2\":\n\n";
	print STDOUT "[INFORMATION '$program_name'] \"$tbl_header2\" .\n";
	print STDOUT "\n\n";

} # End of "if ($TEST_LEVEL >1) { # (0a-2) # ...}"

my $indexes_relv_flds2 = tfm_labels2indices_gen (@RELV_FLD_LABELS, $tbl_header2, $FLD_SEP);
unless (defined $indexes_relv_flds2) {
	print STDOUT "[ERROR '$program_name'] The subroutine, 'tfm_labels2indices_gen', FAILED on the header of the input file \"$infile2\": $? !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}

if ($TEST_LEVEL >1) { # (0b-2) #
	print STDOUT "\n[ Output of 'tfm_labels2indices_gen' (on the 2nd input file): ]\n\n";
	print STDOUT join ("\t", "'Label'", 'Index2'), "\n\n";

	for (my $f=0; $f <$CT_RLV_FLDS; $f++) {
	    my $label = $RELV_FLD_LABELS[$f];
	    my $indx = $indexes_relv_flds2->[$f];
	    print STDOUT join ("\t", "'$label'", $indx), "\n";
	}
	print STDOUT "\n\n";

} # End of "if ($TEST_LEVEL >1) { # (0b-1) # ...}"
 


print STDOUT "\n<<< (0-3) Find Indexes Corresponding to the Field Labels in the input file ('$infile3'). >>>\n\n";

my $fh_in03 = open_input_file2 ($infile3);
unless (defined $fh_in03) {
	print STDOUT "[ERROR '$program_name'] The subroutine, 'open_input_file2', FAILED to open a file-handle for the file \"$infile3\": $? !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}

my $len_lcorner3 = length ($LABEL_CORNER3);


my $tbl_header3;
while (my $line0 = <$fh_in03>) {
    my $sbstr = substr ($line0, 0, $len_lcorner3) ;
    if ($LABEL_CORNER3 eq $sbstr) {
	$tbl_header3 = $line0;
	last;
    }
}
close ($fh_in03);

unless (defined $tbl_header3) {
	print STDOUT "[ERROR '$program_name'] The input file \"$infile3\", has NO LINES starting with '$LABEL_CORNER3': $? !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}
$tbl_header3 =~ s/\s+$//;

if ($TEST_LEVEL >1) { # (0a-3) #
	print STDOUT "\n";
	print STDOUT "[INFORMATION '$program_name'] Retrieved the Table Header from the input file \"$infile3\":\n\n";
	print STDOUT "[INFORMATION '$program_name'] \"$tbl_header3\" .\n";
	print STDOUT "\n\n";

} # End of "if ($TEST_LEVEL >1) { # (0a-3) # ...}"

my $indexes_relv_flds3 = tfm_labels2indices_gen (@RELV_FLD_LABELS, $tbl_header3, $FLD_SEP);
unless (defined $indexes_relv_flds3) {
	print STDOUT "[ERROR '$program_name'] The subroutine, 'tfm_labels2indices_gen', FAILED on the header of the input file \"$infile3\": $? !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}

if ($TEST_LEVEL >1) { # (0b-3) #
	print STDOUT "\n[ Output of 'tfm_labels2indices_gen' (on the 3rd input file): ]\n\n";
	print STDOUT join ("\t", "'Label'", 'Index3'), "\n\n";

	for (my $f=0; $f <$CT_RLV_FLDS; $f++) {
	    my $label = $RELV_FLD_LABELS[$f];
	    my $indx = $indexes_relv_flds3->[$f];
	    print STDOUT join ("\t", "'$label'", $indx), "\n";
	}
	print STDOUT "\n\n";

} # End of "if ($TEST_LEVEL >1) { # (0b-3) # ...}"

			### END of "COPIED (w/ modifications) the (0-1) above, in creation of 'compute_ratios_case_iv_to_iixiii.pl' (on 2020/04/12). (0-2 & 0-3)" ###


print STDOUT "\n\n<<< (1-1) Read in Relevant 2D arrays from the 1st input file & Store them into a 3D array ... >>>\n\n";

##my @set_dcum_arrays_2D = ();
my @set_raw_arrays_2D = ();

for (my $f=0; $f <$CT_RLV_FLDS; $f++) { # 1st Main Outer for-loop (over the relevant fields). #

	my $label_fld = $RELV_FLD_LABELS[$f];
	my $indx_fld = $indexes_relv_flds1->[$f];

	print STDOUT "< The $f th field (label = '$label_fld', index = $indx_fld) >\n\n";
	
	my $fh_in = open_input_file2 ($infile1);
	unless (defined $fh_in) {
		print STDOUT "[ERROR '$program_name'] The subroutine, 'open_input_file2', FAILED to open a file-handle for the file \"$infile1\": $? !\n";
		print STDERR "[ERROR '$program_name'] $!\n";
		die ;
	}

	my $dummy_str_end ;

	my $array_2D = readin_2D_array (*${$fh_in}, $indx_fld, $STR_UNDEF, $dummy_str_end);
	unless (defined $array_2D) {
		print STDOUT "[ERROR '$program_name'] The subroutine, 'readin_2D_array', FAILED to retrieve the field of '$label_fld' in \"$infile1\": $? !\n";
		print STDERR "[ERROR '$program_name'] $!\n";
		die ;
	}

	close ($fh_in);

	$set_raw_arrays_2D[$f] = $array_2D;

} # End of the 1st Main outer for-loop (over the relevant fields). #


			### (IMPORTED (w/ modifications) from 'compute_raw_frequencies_case_ii_theory.pl' & 'compute_raw_frequencies_case_iii_theory.pl', on 2020/04/13) (1-2 & 1-3) ###

print STDOUT "\n\n<<< (1-2) Read in Relevant 1D arrays from the 2nd input file & Store them into a 2D array ... >>>\n\n";

my @set_raw_arrays_1D_ii = ();

for (my $f=0; $f <$CT_RLV_FLDS; $f++) { # 1st Main Outer for-loop (over the relevant fields). #

	my $label_fld = $RELV_FLD_LABELS[$f];
	my $indx_fld = $indexes_relv_flds2->[$f];

	print STDOUT "< The $f th field (label = '$label_fld', index = $indx_fld) >\n\n";
	
	my $fh_in = open_input_file2 ($infile2);
	unless (defined $fh_in) {
		print STDOUT "[ERROR '$program_name'] The subroutine, 'open_input_file2', FAILED to open a file-handle for the file \"$infile2\": $? !\n";
		print STDERR "[ERROR '$program_name'] $!\n";
		die ;
	}

	##my $len_lcorner = length ($LABEL_CORNER);

	my $dummy_str_end ;

	my $array_1D = readin_1D_array (*${$fh_in}, $indx_fld, $STR_UNDEF, $dummy_str_end);
	unless (defined $array_1D) {
		print STDOUT "[ERROR '$program_name'] The subroutine, 'readin_1D_array', FAILED to retrieve the field of '$label_fld' in \"$infile2\": $? !\n";
		print STDERR "[ERROR '$program_name'] $!\n";
		die ;
	}

	close ($fh_in);

	$set_raw_arrays_1D_ii[$f] = $array_1D;


} # End of the 1st Main outer for-loop (over the relevant fields). #


print STDOUT "\n\n<<< (1-3) Read in Relevant 1D arrays from the 3rd input file & Store them into a 2D array ... >>>\n\n";

			### RESTART FROM HERE (1-2 & 1-3) ###

my @set_raw_arrays_1D_iii = ();

for (my $f=0; $f <$CT_RLV_FLDS; $f++) { # 1st Main Outer for-loop (over the relevant fields). #

	my $label_fld = $RELV_FLD_LABELS[$f];
	my $indx_fld = $indexes_relv_flds3->[$f];

	print STDOUT "< The $f th field (label = '$label_fld', index = $indx_fld) >\n\n";
	
	my $fh_in = open_input_file2 ($infile3);
	unless (defined $fh_in) {
		print STDOUT "[ERROR '$program_name'] The subroutine, 'open_input_file2', FAILED to open a file-handle for the file \"$infile3\": $? !\n";
		print STDERR "[ERROR '$program_name'] $!\n";
		die ;
	}

	my $dummy_str_end ;

	my $array_1D = readin_1D_array (*${$fh_in}, $indx_fld, $STR_UNDEF, $dummy_str_end);
	unless (defined $array_1D) {
		print STDOUT "[ERROR '$program_name'] The subroutine, 'readin_1D_array', FAILED to retrieve the field of '$label_fld' in \"$infile3\": $? !\n";
		print STDERR "[ERROR '$program_name'] $!\n";
		die ;
	}

	close ($fh_in);

	$set_raw_arrays_1D_iii[$f] = $array_1D;

} # End of the 1st Main outer for-loop (over the relevant fields). #



			### END of "(IMPORTED (w/ modifications) from 'compute_raw_frequencies_case_ii_theory.pl' & 'compute_raw_frequencies_case_iii_theory.pl', on 2020/04/13) (1-2 & 1-3)" ###


			## CREATED taking advantage the framework of Process (2) in 'compute_raw_frequencies_case_iv_theory.pl' (on 2020/04/13). (Main 2) ##

## print STDOUT "\n\n<<< (2) Output the Computed RAW Probabilities & Expected Frequencies, each to a single file (= \"$outfile1\" or \"$outfile2\") ... >>>\n\n"; # ... For the predecessor, 'compute_raw_frequencies_case_iv_theory.pl'. #
##
print STDOUT "\n\n<<< (2) Compute the RATIOs of RAW multiplication factors in case iv to the products of those in cases ii & iii, & Output them to a single file (= \"$outfile1\") ... >>>\n\n";

my $fh_out1 = open_output_file ($outfile1, 0);
unless (defined $fh_out1) {
	print STDOUT "[ERROR '$program_name'] The subroutine 'open_output_file' FAILED to open a file-handle for the file \"$outfile1\": $? !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}

print $fh_out1 join ("\t", $LABEL_CORNER1, @RELV_FLD_LABELS), "\n\n";

for (my $x=1; $x <= $MAX_X; $x++) { # 2nd Main outer for-loop (over $x). # (=> Further Modified on 2020/04/11.) (2nd Main outer)

	my @set_rows_raw_iv = ();
	for (my $f=0; $f <$CT_RLV_FLDS; $f++) { $set_rows_raw_iv[$f] = $set_raw_arrays_2D[$f]->[$x]; }

	my @set_raw_mfacs_ii = ();

	for (my $f=0; $f <$CT_RLV_FLDS; $f++) { $set_raw_mfacs_ii[$f] = $set_raw_arrays_1D_ii[$f]->[$x]; }

	for (my $y=1; $y <= $MAX_Y; $y++) { # The 2nd Middle for-loop (over $y). # (=> Further Modified on 2020/04/10.) (2nd Middle) #

	    my @set_raw_mfacs_iii = ();

	    ##for (my $f=0; $f <$CT_RLV_FLDS; $f++) { $set_raw_mfacs_iii[$f] = $set_raw_arrays_1D_iii[$f]->[$x]; } # A BUG fixed on 2020/04/13.
	    for (my $f=0; $f <$CT_RLV_FLDS; $f++) { $set_raw_mfacs_iii[$f] = $set_raw_arrays_1D_iii[$f]->[$y]; }

	    my @out_flds1 = ();

	    for (my $f=0; $f <$CT_RLV_FLDS; $f++) { # The inner for-loop (over relevant fields). #

		my $raw_mfac_iv = $set_rows_raw_iv[$f]->[$y];
		my $raw_mfac_ii = $set_raw_mfacs_ii[$f];
		my $raw_mfac_iii = $set_raw_mfacs_iii[$f];

		$out_flds1[$f] = $raw_mfac_iv / ($raw_mfac_ii * $raw_mfac_iii);

	    } # End of The inner for-loop (over relevant fields). #


			# Fixed the BUGs (on 2020/04/10). #
	    ##print $fh_out1 join ("$x:$y", @out_flds1), "\n";
	    print $fh_out1 join ("\t", "$x:$y", @out_flds1), "\n";
			# End of "Fixed the BUGs (on 2020/04/10)." #

	} # End of the 2nd Middle for-loop (over $y). #

}  # 2nd Main outer for-loop (over $x). #

close ($fh_out1);

			## CREATED taking advantage the framework of Process (2) in 'compute_raw_frequencies_case_iv_theory.pl' (on 2020/04/13). (Main 2) ##


	##### DEFINITIONS of the Subroutines. #####


# create_dir_rw:
#	This subroutine receives the path to a directory,
#	examines (1) whether the directory exists.
#	Otherwise, it creates the directory.
#	Then, it examines (2) whether the directory is readable & writable.
#	Otherwise, it changes the permission mode of the directory to make it readable & writable.
#	[CREATED on 2020/01/29; IMPORTED from 'lastpiece.alpha.pl' on 2020/04/09.]
#
sub create_dir_rw (\$) {

	my ($rf_dir_path) = @_;
	my $dir_path = ${$rf_dir_path};

	unless ( -d $dir_path ) {
	   unless (0 == system ("mkdir -p $dir_path") ) {
		print STDOUT "[ERROR '$program_name'] FAILED to CREATE the OUTPUT DIRECTORY, '$dir_path/': $? !\n\n";
		print STDERR "[ERROR '$program_name'] $!\n";
		return 0 ;
	   }
	}
	unless (( -r $dir_path ) and ( -w $dir_path )) {
	    unless (0 == system ("chmod u+rw $dir_path") ) {
		print STDOUT "[ERROR '$program_name'] FAILED to CHANGE permission MODES of the output directory, '$dir_path': $? !\n";
		print STDERR "[ERROR '$program_name'] $!\n";
		return 0 ;
	    }
	}

	return 1;

}	# End of the subroutine, 'create_dir_rw'. #


