#!/usr/bin/perl
#
###################################################################################
#	
#	LASTPIECE_P (Local Alignment-STate Probabilities that Insertion-type and dEletion-type gaps Co-Exist, type P):
#	    A package of programs to compute the probabilities of gapped segments in each of which an insertion-type gap and a deletion-type gap co-exist, (which were referred to as case-(iv) gapped segments by Ezawa (2016a?)),
#	    under a stochastic model of sequence evolution with biologically realistic insertions/deletions.
#	   (Written almost exclusively in Perl.)
#
#	Version 0.3: Copyright (C) 2020 Kiyoshi Ezawa
#
#    This package is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This package is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License, "GNU_GPL.txt",
#    along with this package.  If not, see <http://www.gnu.org/licenses/>.
#
#    The author can be contacted by e-mailing to <kezawa dot ezawa3 at gmail dot com>
#    (replace " dot " and " at " with "." and "@", respectively).
#
###################################################################################
#
# cmpt_mu_A2empty.pl:
#	This servant script computes the "extended" multiplication factors
#	for 'A -> {emptyset}':
#	mu_{A->emptyset} [ (L_I, t_I); (0, 0) ; t_2](sigmaBE=2),
#	where L_I is the number of sites in the "A"-region at the 'initial' time t_I,
#	and t_2 is some time-point after the deletion of the "A"-region.
#	(Actually, a "D"-region will be created somewhere in [t_2, t_2 + Delta{}t_2].)
#	The script was created by taking advantage of the main framework of another servant script,
#	'cmpt_mu_Dcr.pl' (in "../../Basic/mu_Dcr/"),
#	as well as of the main process of 'cmpt_mu_Adel.pl' (in "../../Basic/mu_Adel/").
#	[CREATION STARTED on 2020/02/07, ENDED on 2020/02/08.]
#
# cmpt_mu_Dcr.pl:
#	This servant script computes the multiplication factors
#	for the creation of a "D"-region:
#	mu_{D-cr} [ (x, [t_1, t_1 + Delta{}t]) ; (L, t_2) ],
#	where x is the position at which the "D"-region is created,
#	[t_1, t_1 + Delta{}t] is the time-interval within which the D-creation occurred,
#	and L is the number of sites in the "D"-region at the time-point t_2 ( > t_1).
#	The script was created by taking advantage of the main framework of another servant script,
#	'cmpt_muP0.pl' (in "../muP0/") .
#	[CREATION STARTED on 2020/02/03, ENDED on 2020/02/04; Modified on 2020/02/05]
#


	##### (IMPORTED w/ modifications from '../../Basic/mu_Dcr/cmpt_mu_Dcr.pl' on 2020/02/07) (HD & CL & CHK) #####

		### (IMPORTED w/ modifications from '../../../lastpiece.alpha.pl' on 2020/01/30) (HD) ###

use strict;
use warnings;

use MyFile_manipu qw(
	open_input_file2
	open_output_file
	list_sessions_ext
	);

use MyLASTPIECE_IO qw(
	readin_muP_case_i
	readin_mu_Adel
	);
	# print_muP_case_i
	# print_muP0s
	# readin_indel_params
	# readin_indel_rates
	# print_muP_case_ii_or_iii 
	# readin_muP0s
	#
	# Added 'readin_indel_params' & 'readin_indel_rates' on 2020/01/31. # => Removed in creation of 'cmpt_mu_A2empty.pl' (on 2020/02/07). #
	# Added 'print_muP_case_i' & 'print_muP_case_ii_or_iii' on 2020/02/01. # => Removed on 2020/02/03. 
	# => Resurrected 'print_muP_case_ii_or_iii', in creation of 'cmpt_mu_Dcr.pl' (on 2020/02/05). # => Removed Again in creation of 'cmpt_mu_A2empty.pl' (on 2020/02/07). #
	#
	# Added 'print_muP0s' on 2020/02/03. => Removed on 2020/02/04.
	# Added 'readin_muP0s' on 2020/02/04. => Removed in creation of 'cmpt_mu_A2empty.pl' (on 2020/02/07).
	#
	# Added 'readin_mu_Adel' & 'readin_muP_case_i', in creation of 'cmpt_mu_A2empty.pl' (on 2020/02/07).

##use MyLASTPIECE_Sppl qw(
##	cal_delta_exp_del_5p2
##	);


##sub create_dir_rw (\$); # Added on 2020/01/29.


$| = 1; # Flush the buffer for standard output as soon as it receives data.

my $bin_perl = "/usr/bin/perl"; # The Perl interpreter. # (Added on 2020/01/29). #

my $program_path = $0;
my ($program_name) = ($program_path =~ m!([^/]+)$!);

	# Controlling the level of test outputs. #
	# (0: silent; 1: very quiet; 2: quiet; 3: medium; 4: verbose; 5: very verbose) 
our $TEST_LEVEL = 2;
##our $TEST_LEVEL = 5;

$MyLASTPIECE_IO::TEST_LEVEL = $TEST_LEVEL ; # Added on 2020/01/27. # RESURRECTED on 2020/02/04.


		### END of "(IMPORTED w/ modifications from '../../../lastpiece.alpha.pl' on 2020/01/30) (HD)" ###

=For the predecessor, 'cmpt_mu_Dcr.pl'. (GL)
our $EPS = 1E-07; # A "small" number to be used for checking the consistency. # (Added on 2020/02/01.) #

our $IF_DRESSED = 0; # Whether the input indel rates are raw (=0) or dressed (=1). (The default value = 0.)  # Added in creation of 'cmpt_mu_Dcr.pl' (on 2020/02/04). #
=cut # End of "For the predecessor, 'cmpt_mu_Dcr.pl'. (GL)"

	####### Take in the Command-Line Arguments. #######

		# Modified in creation of 'cmpt_mu_A2empty.pl' (on 2020/02/07) (CL1) #
unless (@ARGV > 7) {
	print STDERR "\nUsage:  [perl] $program_name {initial_time (non-negative number)} {final_time (positive number, greater than initial_time)} {number_of_sub_time_intervals (positive integer)} {upper-bound_of_number_of_sites} {SIGMA_BE (MUST ALWAYS BE 2!!)} {input_file_for_mu_Adel (w/ SIGMA_BE=2)} {input_file_for_muP_case_i} {output_file} ";
	print STDERR "\n\n";
	die "$!\n\n";
}
=For the predecessor, 'cmpt_mu_Dcr.pl'. (CL1)
unless (@ARGV > 8) {
	print STDERR "\nUsage:  [perl] $program_name {initial_time (non-negative number)} {final_time (positive number, greater than initial_time)} {number_of_sub_time_intervals (positive integer)} {upper-bound_of_number_of_sites} {SIGMA_BE (MUST ALWAYS BE 2!!)} {input_file_for_indel_parameters} {input_file_for_raw_indel_rates} {input_file for muP0 (w/ SIGMA_BE=2)} {output_file} "; # CORRECTED on 2020/02/06. # 
		# End of "Modified in creation of 'cmpt_muP0.pl' (on 2020/02/03) (CL1)" #
	print STDERR " [ options ] ,\n";
	print STDERR "   where the [options] can be either of the following:\n\n";
	print STDERR "            --RAW      when using the RAW indel rates;\n";
	print STDERR "            --DRESSED      when using the DRESSED indel rates.\n";
	print STDERR "		  (NOTE1: If both options are used, the '--RAW' will be prioritized.)\n"; # (The '\n";' was added in debugging on 2020/02/05.)
	print STDERR "		  (NOTE2: If no options are used, the '--RAW' will be assumed.)\n"; # (The '\n";' was added in debugging on 2020/02/05.)
	##print STDERR "           --XLEVEL=integer       for the level of MSA space exploration (from =1 (performing only multiple-\"shift\"s) to =5 (performing all programed moves) (default=$XLEVEL);\n"; # ADDED on 2019/08/01.
	print STDERR "\n";

	die "$!\n\n";
}
=cut # End of "For the predecessor, 'cmpt_mu_Dcr.pl'. (CL1)"
		# End of "Modified in creation of 'cmpt_mu_A2empty.pl' (on 2020/02/07) (CL1)" #

		# Modified in creation of 'cmpt_mu_A2empty.pl' (on 2020/02/04) (CL2) #
##our ($TIME_INIT, $TIME_FIN, $CT_SB_INTVLS, $UB_CT_SITES_CMPT, $SIGMA_BE) = @ARGV[0 .. 3+1];
##my ($infile_id_prms, $infile_id_rates, $infile_muP0, $ofile_mu_Dcr) = @ARGV[5 .. 8];
our ($TIME_INIT, $TIME_FIN, $CT_SB_INTVLS, $UB_CT_SITES_CMPT, $SIGMA_BE) = @ARGV[0 .. 3+1];
my ($infile_mu_Adel, $infile_muP_case_i, $ofile_mu_A2empty) = @ARGV[5 .. 7];
		# End of "Modified in creation of 'cmpt_mu_A2empty.pl' (on 2020/02/04) (CL2)" #

=For the predecessor, 'cmpt_mu_Dcr.pl'. (CL2)
		# Added in creation of 'cmpt_mu_Dcr.pl' (on 2020/02/04). (CL) #
my ($if_dressed, $if_raw) = (0, 0);

for (my $k=9; $k < @ARGV; $k++) { # Preliminary for-loop (over additional command-line arguments). #

	my $clarg = $ARGV[$k];
	unless ($clarg =~ s/^(\-)+//) { next; } # Skip if the argument lacks preceding '-'s. #

	my ($option, $value) = split /\=/, $clarg;
	unless ((defined $value) and ($value ne '')) {
	     print STDOUT "[CAUTION '$program_name'] Found an UNEXPECTED FORMAT of the commmand-line argument, ", $ARGV[$k], ": $! ! ... I will SKIP it ...\n\n";
	     next; # Skip an argument in an unexpected format.
	}
	my $org_option = $option;
	$option = uc $option; # Capitalize the option, in advance.

	if ($option eq 'RAW') {
	    $if_raw = 1;

	} elsif ($option eq 'DRESSED') {
	    $if_dressed = 1;

	} else {
	    print STDOUT "[CAUTION '$program_name'] Found an UNEXPECTED OPTION, '$org_option', within the cmmmand-line argument, ", $
ARGV[$k], ": $! ! ... I will SKIP it ...\n\n";
	    next; # Skip an argument with an unexpected option.
	}

} # End of the "Preliminary for-loop (over additional command-line arguments)." #

if ($if_dressed and !($if_raw)) {
	$IF_DRESSED = 1;

	print STDOUT "\n[INFORMATION '$program_name'] AS YOU SPECIFIED, we set \$IF_DRESSED to be 1 (true) ...\n\n";
}
		# End of "Added in creation of 'cmpt_mu_Dcr.pl' (on 2020/02/04). (CL)" #
=cut # End of "For the predecessor, 'cmpt_mu_Dcr.pl'. (CL2)"


	####### END of "Take in the Command-Line Arguments." #######

	## Check the value of $SIGMA_BE ##
	## (Added in creation of 'cmpt_mu_Dcr.pl' on 2020/02/04.) ##
unless ($SIGMA_BE == 2) {
	print STDOUT "[ERROR '$program_name'] The \$SIGMA_BE (=$SIGMA_BE) is NOT 2, which this script requires: $? !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}
	## End of "(Added in creation of 'cmpt_mu_Dcr.pl' on 2020/02/04.)" ##


	### Check the existence & permission mode of the input files. ###
	### (Added on 2020/02/03.) ###

##foreach my $infile ($infile_id_prms, $infile_id_rates, $infile_muP0) {...}
foreach my $infile ($infile_mu_Adel, $infile_muP_case_i) { # Modified in creation of 'cmpt_mu_A2empty.pl' (on 2020/02/07). (CHK) #
    unless ( ( -f $infile ) and ( -r $infile ) ) {
	print STDOUT "[ERROR '$program_name'] CanNOT FIND a readable input file, \"$infile\": $? !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
    }
}
	### End of "(Added on 2020/02/03.)" ###

	##### END of "(IMPORTED w/ modifications from '../../Basic/mu_Dcr/cmpt_mu_Dcr.pl' on 2020/02/07) (HD & CL & CHK)" #####



	### Declare the Beginning of the Script ###
	### (IMPORTED w/ modifications from '../../Basic/mu_Dcr/cmpt_mu_Dcr.pl' on 2020/02/07) (DC) ###

print STDOUT "\n<<< Program name: $program_name (path: $program_path) >>>\n\n";

print STDOUT "\n<< Input Files: >>\n\n";

		# Modified in creation of 'cmpt_mu_A2empty.pl' (on 2020/02/07). (DC1) #
##print STDOUT " (for indel parameters) $infile_id_prms ,\n";
##print STDOUT " (for raw indel rates)  $infile_id_rates ,\n";
##print STDOUT " (for muP0 under \"base\" rate operator)  $infile_muP0 .\n\n"; # Added in creation of 'cmpt_mu_Dcr.pl' (on 2020/02/04). (DC1) #
print STDOUT " (for mutiplication factors for the creation of an \"A\"-region) $infile_mu_Adel ,\n";
print STDOUT " (for muP of case-i segments)  $infile_muP_case_i .\n\n";
		# End of "Modified in creation of 'cmpt_mu_A2empty.pl' (on 2020/02/07). (DC1)" #

print STDOUT "\n<< Output File: >>\n\n"; 

		# Modified in creation of 'cmpt_mu_A2empty.pl' (on 2020/02/07). (DC2) #
##print STDOUT " (for multiplication factors for the creation of a \"D\"-region)  $ofile_mu_Dcr .\n\n";
print STDOUT " (for multiplication factors for 'A -> {emptyset}')  $ofile_mu_A2empty .\n\n";
		# End of "Modified in creation of 'cmpt_mu_A2empty.pl' (on 2020/02/07). (DC2)" #


print STDOUT "\n<< Global Variables specifying the levels of tests: >>\n\n"; # Added on 2020/01/30. (DC2) #

print STDOUT "\n", '$TEST_LEVEL = ', $TEST_LEVEL, " .\n\n";

print STDOUT '$MyLASTPIECE_IO::TEST_LEVEL  = ', $MyLASTPIECE_IO::TEST_LEVEL , " .\n\n"; # Added on 2020/02/04.

=For the predecessor, 'cmpt_mu_Dcr.pl'. (DC)
print STDOUT "\n<< A \"small\" number to be used for checking the consistency: >>\n\n"; #
print STDOUT '$EPS = ', $EPS, " .\n\n";

		# Added in creation of 'cmpt_mu_Dcr.pl' (on 2020/02/04). (DC2) #
print STDOUT "\n<< Option that specifies the type of indel rates: >>\n\n";
print STDOUT '$IF_DRESSED = ', $IF_DRESSED, " .\n\n";
		# End of "Added in creation of 'cmpt_mu_Dcr.pl' (on 2020/02/04). (DC2)" #
=cut # End of "For the predecessor, 'cmpt_mu_Dcr.pl'. (DC)"


print STDOUT "\n<< Parameters necessary for recursive computations of the multiplication factors: >>\n\n";

print STDOUT '($TIME_INIT, $TIME_FIN) = ', "($TIME_INIT, $TIME_FIN) .\n\n";
print STDOUT '#{Sub-time-intervals} = ', $CT_SB_INTVLS, " .\n\n";

print STDOUT '{Upper-bound of #{sites} for computation} = ', $UB_CT_SITES_CMPT, " .\n\n"; 

print STDOUT '$SIGMA_BE = ', $SIGMA_BE, " .\n\n"; # Added in creation of 'cmpt_muP0.pl' (on 2020/02/03). #

	### End of "(IMPORTED w/ modifications from '../../Basic/mu_Dcr/cmpt_mu_Dcr.pl' on 2020/02/07) (DC)" ###


	## Preliminary 00-C: Compute Other Important Parameters ##
	### (IMPORTED w/ modifications from '../../../lastpiece.alpha.pl' on 2020/01/30) (00-C) ###
	### (==> IMPORTED w/ modifications from '../../Basic/mu_Dcr/cmpt_mu_Dcr.pl' on 2020/02/08) (00-C) ###

print STDOUT "\n<<< Preliminary Process 00-C: Computing Other Important Parameters >>>\n\n";

##our $TOTRATE_INDEL = $TOTRATE_INS + $TOTRATE_DEL; ## Disabled in creation of 'cmpt_mu_A2empty.pl' (on 2020/02/08). #
our $SIZE_FULL_INTVL = $TIME_FIN - $TIME_INIT;
our $SIZE_SB_INTVL = $SIZE_FULL_INTVL /$CT_SB_INTVLS;

		# Modified on 2020/01/31. (00-C-1) #
our @TIME_LAPSES_CMPT = (); # Values of the time lapse (= t_F - t_I) at which the factors will be computed.
for (my $i = 0; $i <= $CT_SB_INTVLS; $i++) {
	$TIME_LAPSES_CMPT[$i] = $i * $SIZE_SB_INTVL;
} 
=OBSOLETE as of 2020/01/31. (A)
our @TIME_LAPSES = (); # Values of the time lapse (= t_F - t_I) to be examined.
for (my $i = 0; $i <= $CT_SB_INTVLS; $i++) {
	$TIME_LAPSES[$i] = $i * $SIZE_SB_INTVL;
} 
=cut # End of "OBSOLETE as of 2020/01/31. (A)"
		# End of "Modified on 2020/01/31. (00-C-1)" #

=DISABLED on 2020/01/31. (1)
our @E_INDEL_CTS = (); # Expected counts of indels. #
foreach my $tlapse (@TIME_LAPSES) { push @E_INDEL_CTS, $TOTRATE_INDEL * $tlapse; }
=cut # End of "DISABLED on 2020/01/31. (1)"

##print STDOUT '$TOTRATE_INDEL = ', $TOTRATE_INDEL, " \;\n"; # Disabled in creation of 'cmpt_mu_A2empty.pl' (on 2020/02/08). #
print STDOUT '$SIZE_FULL_INTVL = ', $SIZE_FULL_INTVL, " \;\n";
print STDOUT '$SIZE_SB_INTVL = ', $SIZE_SB_INTVL, " .\n\n";

		# Modified on 2020/01/31. (00-C-2) #
print STDOUT '@TIME_LAPSES_CMPT = (', join (', ', @TIME_LAPSES_CMPT), ") .\n\n";
##print STDOUT '@TIME_LAPSES = (', join (', ', @TIME_LAPSES), ") .\n\n";
		# End of "Modified on 2020/01/31. (00-C-2)" #

=DISABLED on 2020/01/31. (2)
print STDOUT '@E_INDEL_CTS = (', join (', ', @E_INDEL_CTS), ") .\n\n";
=cut # End of "DISABLED on 2020/01/31. (2)"

print STDOUT "\n";

	### End of "(IMPORTED w/ modifications from '../../../lastpiece.alpha.pl' on 2020/01/30) (00-C) ###
	### (==> IMPORTED w/ modifications from '../../Basic/mu_Dcr/cmpt_mu_Dcr.pl' on 2020/02/08) (00-C)" ###



	### MODIFIED in creation of 'cmpt_mu_A2empty.pl' (on 2020/02/07). (00-IN) ###

print STDOUT "\n<<< Preliminary Process 00-IN: Reading in the quantities necessary for the computations >>>\n\n";

print STDOUT "\n<< (Pre 00-IN-D): Reading in the multiplication factors for the deletion of an \"A\"-region >>\n\n";

my ($rf_mu_Adel) =  readin_mu_Adel ($infile_mu_Adel, $CT_SB_INTVLS, $UB_CT_SITES_CMPT);

	# $rf_mu_Adel->[$LI][$jt][$dL_fD] = (w/ 0 < $L_I <= $ub_ct_sites, & 0 <= $jt <= $ct_sb_intvls & 0 <= $dL_fD <= {upper-bound specified by the input} ) 
	# is the multiplication factor that an "A"-region (with $LI initial sites) was deleted
	# at the $jt th time-lapse (actually, between $jt-1 and $jt),
	# in conjunction with $dL_fD sites in the flanking "D"-region(s).

unless (defined $rf_mu_Adel) {
	print STDOUT "[ERROR '$program_name'] The subroutine, 'readin_mu_Adel', FAILED on the input file, \"$infile_mu_Adel\" $? !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}
(print STDOUT "\n ... The subroutine, 'readin_mu_Adel', Ended on the input file, \"$infile_mu_Adel\" ...\n\n") if ($TEST_LEVEL > 1);

my $ct_tlapses = @TIME_LAPSES_CMPT;
##my $ct_tlapses = $CT_SB_INTVLS +1;
my $cap_ct_sites = $UB_CT_SITES_CMPT +1;

		# Check the consistency of @{$rf_mu_Adel}. # (Added on 2020/02/08.) #

for (my $LI = 1; $LI < $cap_ct_sites; $LI++) { # Trailing outer for-loop (over the initial site-counts). #

	my $mu_Adel_fxd_LI = $rf_mu_Adel->[$LI];
	my $cap_dL_fD = @{$mu_Adel_fxd_LI->[0]};

	unless ($cap_dL_fD == 1) { # This Condition is Imposed specially for $SIGMA_BE = 2. #
		print STDOUT "[ERROR '$program_name'] The number of dL_fD's (=$cap_dL_fD) in the input ", '@{$rf_mu_Adel->[$LI]} (w/ $LI=', "$LI)  CLASHES with 1 (unity), which is expected for \$SIGMA_BE = 2: $? !\n";
		print STDERR "[ERROR '$program_name'] $!\n";
		die ;
	}

}  # Trailing outer for-loop (over the initial site-counts). #

		# End of "Check the consistency of @{$rf_mu_Adel}. # (Added on 2020/02/08.)" #


if ($TEST_LEVEL >= 5) { # (00-IN-D) #

	print STDOUT "\n\n[[ Multiplication factors, mu_{Adel} [ (L_I, t_I) ; (0, - dL_fD, t) ], retrieved (& stored in ", '@{$rf_mu_Adel}', "): ]]\n\n";

	print STDOUT join ("\t", '$LI:$dL_fD\$jt', 0 .. $CT_SB_INTVLS), "\n\n";


	for (my $LI = 1; $LI < $cap_ct_sites; $LI++) { # Main outer for-loop (over the initial site-counts).

	    my $mu_Adel_fxd_LI = $rf_mu_Adel->[$LI];

	    my $cap_dL_fD = @{$mu_Adel_fxd_LI->[0]};

	    for (my $dL_fD =0; $dL_fD < $cap_dL_fD; $dL_fD++) { # Main middle for-loop (over the counts of deleted flanking-D sites). # (MODIFIED on 2020/02/02). #

		my @out_mu_Adel = ();

		for (my $jt = 0; $jt < $ct_tlapses; $jt++) { # Inner for-loop (over the time-intervals). #
		    $out_mu_Adel[$jt] = $mu_Adel_fxd_LI->[$jt][$dL_fD];
		} # End of the inner for-loop (over the time-intervals). #

		print STDOUT join ("\t", "$LI:$dL_fD", @out_mu_Adel), "\n";

	    } # End of the main middle for-loop (over the counts of deleted flanking-D sites). #

	} # Main outer for-loop (over the initial site-counts).

			## End of "IMPORTED w/ modifications from the subroutine, 'print_muP0s', in "MyLASTPIECE_IO.pm" (on 2020/02/05)." ## 

	print STDOUT "\n\n\n";

} # End of "if ($TEST_LEVEL >= 5) { # (00-IN-D) # ...}"



print STDOUT "\n<< (Pre 00-IN-E): Reading in the multiplication factors of case-i segments >>\n\n";

my ($rf_muP_case_i) =  readin_muP_case_i ($infile_muP_case_i, $CT_SB_INTVLS);

	# $rf_muP_case_i->[$jt] (w/ 0 <= $jt <= $ct_sb_intvls)
	# is the multiplication factor of case-i gapped segments, which does not have any ancestral or descendant sites in between the PASs,
	# at the $jt th time-lapse.

## ($infile_mu_Adel, $infile_muP_case_i, $ofile_mu_A2empty) = @ARGV[5 .. 7];

unless (defined $rf_muP_case_i) {
	print STDOUT "[ERROR '$program_name'] The subroutine, 'readin_muP_case_i', FAILED on the input file, \"$infile_muP_case_i\" $? !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}
(print STDOUT "\n ... The subroutine, 'readin_muP_case_i', Ended on the input file, \"$infile_muP_case_i\" ...\n\n") if ($TEST_LEVEL > 1);


if ($TEST_LEVEL >= 5) { # (00-IN-E) #

	print STDOUT "\n\n[[ Multiplication factors, muP_{case-i} [ (L_F=0, t_F) ; (L_I=0,t_I) ], retrieved (& stored in ", '@{$rf_muP_case_i}', "): ]]\n\n";

	print STDOUT join ("\t", '#{sites}_Init:#{sites}_Fin\$jt', 0 .. $CT_SB_INTVLS), "\n\n";

	print STDOUT join ("\t", '0:0', @{$rf_muP_case_i}), "\n\n\n";

} # End of "if ($TEST_LEVEL >= 5) { # (00-IN-E) # ...}"



=For the predecessor, 'cmpt_mu_Dcr.pl'. (00-IN)
	## (MODIFIED on 2020/01/31) (00-IN) ##

print STDOUT "\n<<< Preliminary Process 00-IN: Reading in the quantities necessary for the computations >>>\n\n";

	## 00-IN-A: Read in the input file for indel parameters. ##

print STDOUT "\n<< (Pre 00-IN-A): Reading in the indel model parameters >>\n\n";

	# [ Omitted Actual Commands ... ] #


	## Preliminary 00-B: Read in the input file for indel rates. ##

		# Modified in creation of 'cmpt_mu_Dcr.pl' (on 2020/02/04). (00-B-1) #
if ($IF_DRESSED) {
	print STDOUT "\n<< (Pre 00-IN-B): Reading in the DRESSED indel rates >>\n\n";
} else {
	print STDOUT "\n<< (Pre 00-IN-B): Reading in the RAW indel rates >>\n\n";
}
	# [ Omitted Actual Commands ... ] #


	## 00-IN-C: Read in the input file for muP0s (w/ $SIGMA_BE = 2). ##
	## (Added in creation of 'cmpt_mu_Dcr.pl' on 2020/02/04&05). (00-IN-C) ##

print STDOUT "\n<< (Pre 00-IN-C): Reading in the muP0s (w/ \$SIGMA_BE = 2) >>\n\n";

	# [ Omitted Actual Commands ... ] #

	## End of "(MODIFIED on 2020/01/31) (00-IN)" ##

=cut # END of "For the predecessor, 'cmpt_mu_Dcr.pl'. (00-IN)"

	### END of "MODIFIED in creation of 'cmpt_mu_A2empty.pl' (on 2020/02/07). (00-IN)" ###




		### IMPORTED w/ modifications from '../../Basic/mu_Adel/cmpt_mu_Adel.pl' (on 2020/02/08). (1st-half) ###

	###  Main Process: Compute & Output \"extended\" multiplication factors, 'mu_{A->emptyset} [ (L_I, t_I); (0, 0) ; t_2](sigmaBE=2)'. ###

print STDOUT "\n<<<<< Main Process: Compute & Output \"extended\" multiplication factors, 'mu_{A->emptyset} [ (L_I, t_I); (0, 0) ; t_2](sigmaBE=2)' >>>>>\n\n\n";

#	mu_{A->emptyset} [ (L_I, t_I); (0, 0) ; t_2](sigmaBE=2),
#	where L_I is the number of sites in the "A"-region at the 'initial' time t_I,
#	and t_2 is some time-point after the deletion of the "A"-region.


		# (Preliminary) Open a file-handle for outputting to the output file. #
my $fh_out = open_output_file ($ofile_mu_A2empty, 0);
unless (defined $fh_out) {
	print STDOUT "[ERROR '$program_name'] The subroutine 'open_output_file' FAILED to open a file-handle for the file \"$ofile_mu_A2empty\": $? !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}
				## $ofile_mu_A2empty

print $fh_out join ("\t", '#{sites}_Init\Time_Lapse', @TIME_LAPSES_CMPT), "\n\n";

##my $ct_tlapses = @TIME_LAPSES_CMPT;
##my $cap_ct_sites = $UB_CT_SITES_CMPT +1;

for (my $LI = 1; $LI < $cap_ct_sites; $LI++) { # Main Outermost for-loop (over initial site-counts). #

    my $mu_Adel_fxd_LI = $rf_mu_Adel->[$LI];

	# $rf_mu_Adel->[$LI][$jt][$dL_fD] = (w/ 0 < $L_I <= $ub_ct_sites, & 0 <= $jt <= $ct_sb_intvls & 0 <= $dL_fD <= {upper-bound specified by the input (= 0 here)} ) 
	# is the multiplication factor that an "A"-region (with $LI initial sites) was deleted
	# at the $jt th time-lapse (actually, between $jt-1 and $jt),
	# in conjunction with $dL_fD sites in the flanking "D"-region(s).

		### END of "IMPORTED w/ modifications from '../../Basic/mu_Adel/cmpt_mu_Adel.pl' (on 2020/02/08). (1st-half)" ###

		### Newly Added in creation of 'cmpt_mu_A2empty.pl' (on 2020/02/08). ###

    my @vec_mu_A2empty = ();

    for (my $jt=0; $jt < $ct_tlapses; $jt++) { # Main middle for-loop (over output time-lapses). #

	my $mu_A2empty = 0;

	for (my $jt2=0; $jt2 <= $jt; $jt2++) { # Main inner for-loop (over intermediate time-lapses). #

	    $mu_A2empty += $mu_Adel_fxd_LI->[$jt2][0] * $rf_muP_case_i->[$jt - $jt2] ;

		# $rf_muP_case_i->[$jt] (w/ 0 <= $jt <= $ct_sb_intvls)
		# is the multiplication factor of case-i gapped segments, which does not have any ancestral or descendant sites in between the PASs,
		# at the $jt th time-lapse.

	} # Main inner for-loop (over intermediate time-lapses). #

	$vec_mu_A2empty[$jt] = $mu_A2empty;

    } # End of the Main middle for-loop (over output time-lapses). #

		### End of "Newly Added in creation of 'cmpt_mu_A2empty.pl' (on 2020/02/08)." ###

		### IMPORTED w/ modifications from '../../Basic/mu_Adel/cmpt_mu_Adel.pl' (on 2020/02/08). (2nd-half) ###

    print $fh_out join ("\t", $LI, @vec_mu_A2empty), "\n";

} # End of the Main Outermost for-loop (over initial site-counts). #

close ($fh_out);


print STDOUT "\n\n === All The Processes of '$program_name' FINISHED !! ===\n\n\n";

		### END of "IMPORTED w/ modifications from '../../Basic/mu_Adel/cmpt_mu_Adel.pl' (on 2020/02/08). (2nd-half)" ###


