#!/usr/bin/perl
#
###################################################################################
#	
#	LASTPIECE_P (Local Alignment-STate Probabilities that Insertion-type and dEletion-type gaps Co-Exist, type P):
#	    A package of programs to compute the probabilities of gapped segments in each of which an insertion-type gap and a deletion-type gap co-exist, (which were referred to as case-(iv) gapped segments by Ezawa (2016a?)),
#	    under a stochastic model of sequence evolution with biologically realistic insertions/deletions.
#	   (Written almost exclusively in Perl.)
#
#	Version 0.3: Copyright (C) 2020 Kiyoshi Ezawa
#
#    This package is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This package is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License, "GNU_GPL.txt",
#    along with this package.  If not, see <http://www.gnu.org/licenses/>.
#
#    The author can be contacted by e-mailing to <kezawa dot ezawa3 at gmail dot com>
#    (replace " dot " and " at " with "." and "@", respectively).
#
###################################################################################
#
# cmpt_mu_dotA2dotAD2dotAdel.pl:
#	This servant script computes the multiplication factors for '.A -> .AD -> .Adel',
#	mu_{.A -> .AD -> .Adel} [ (LA, t_I); x_m; (-dL_DL, LDR, [t_F - dt_F, t_F]) ] (sigmaBE_m=1, sigmaBE_F = 0).
#	Here, LA is the number of sites in the "A"-region at t_I,
#	LDR is the number of sites in the "D"-region on the right at t_F,
#	and dL_DL is the number of sites in the "D"-region on the left that were deleted
#	simultaneously with the "A"-region at some time in [t_F -dt_F, t_F].
#	The script was created by somewhat modifying the predecessor, 'graft_convolution.pl'.
#	[CREATION STARTED on 2020/02/20, ENDED on 2020/02/21.]
#
# graft_convolution.pl:
#	This servant script performs the "graft-convolution" of TWO multiplication factors of the form:
#	muP_{1st or 2nd} [ (L_F, t_F)  | (L_I, t_I) ],
#	that the number of sites is L_I at the initial time (=t_I) and L_F at the final time (=t_F),
#	with 1 <= L_I, L_F <= $UB_CT_SITES.
#	Here, the "graft-convolution" means the following operation:
#	muP_{Output} [ (L_F, t_F) | (L_I, t_I) ]
#	     = \sum_{Lm = 1}^{$UB_CT_SITES} \int_{t_I}^{t_F} d{}tm  [
#	    muP_{1st} [ (Lm, tm)  | (L_I, t_I) ] * muP_{2nd} [ (L_F, t_F)  | (Lm, tm) ] 
#		].
#	The script was created by taking advantage of the main framework of another servant script,
#	'cmpt_mu_Dcr2Adel.pl' (in "./2nd_ii/mu_Dcr2Adel/").
#	[CREATION STARTED on 2020/02/14, ENDED on 2020/02/dd.]
#	
#
# cmpt_mu_Dcr2Adel.pl
#	This servant script computes the multiplication factor,
#	mu_{Dcr->Adel} [ (L_A, x(=L_A), [t_I, t_I+dt_I]) ; (L_D, [t_F-dt_F, t_F]) ](sigmaBE=1),
#	for the creation of a "D"-region within [t_I, t_I + dt_I]
#	and the deletion of an "A"-region within [t_F - dt_F, t_F],
#	where L_D is the number of sites belonging to the "D"-region at the "final" time (t_F),
#	and L_A is the number of sites belonging to the "A"-region at the "initial" time (t_I).
#	The script was created by taking advantage of the main framework of another servant script,
#	'cmpt_muP_2nd_i.pl' (in "../../2nd_i/PreFinal/").
#	[CREATION STARTED on 2020/02/11, ENDED on 2020/02/dd.]
#
# cmpt_muP_2nd_i.pl:
#	This servant script computes the multiplication factor,
#	muP_{case (iv)}^{2nd (i)} [(L_DF, t_F) | (L_AI, t_I) ],
#	which is the total contributions from ALL indel histories belonging to the class, "2nd (i)".
#	Here, L_DF is the final number of sites in the "D"-region, 
#	L_AI is the initial number of sites in the "A"-region,
#	and t_I and t_F are the initial and final time-points, respectively.
#	The script was created by taking advantage of the main framework of another servant script,
#	'cmpt_mu_A2empty.pl' (in "../mu_A2empty/").
#	[CREATION STARTED on 2020/02/08, ENDED on 2020/02/11.]
#
# cmpt_mu_A2empty.pl:
#	This servant script computes the "extended" multiplication factors
#	for 'A -> {emptyset}':
#	mu_{A->emptyset} [ (L_I, t_I); (0, 0) ; t_2](sigmaBE=2),
#	where L_I is the number of sites in the "A"-region at the 'initial' time t_I,
#	and t_2 is some time-point after the deletion of the "A"-region.
#	(Actually, a "D"-region will be created somewhere in [t_2, t_2 + Delta{}t_2].)
#	The script was created by taking advantage of the main framework of another servant script,
#	'cmpt_mu_Dcr.pl' (in "../../Basic/mu_Dcr/"),
#	as well as of the main process of 'cmpt_mu_Adel.pl' (in "../../Basic/mu_Adel/").
#	[CREATION STARTED on 2020/02/07, ENDED on 2020/02/08.]
#
# cmpt_mu_Dcr.pl:
#	This servant script computes the multiplication factors
#	for the creation of a "D"-region:
#	mu_{D-cr} [ (x, [t_1, t_1 + Delta{}t]) ; (L, t_2) ],
#	where x is the position at which the "D"-region is created,
#	[t_1, t_1 + Delta{}t] is the time-interval within which the D-creation occurred,
#	and L is the number of sites in the "D"-region at the time-point t_2 ( > t_1).
#	The script was created by taking advantage of the main framework of another servant script,
#	'cmpt_muP0.pl' (in "../muP0/") .
#	[CREATION STARTED on 2020/02/03, ENDED on 2020/02/04; Modified on 2020/02/05]
#


	##### (IMPORTED w/ modifications from '../mu_A2empty/cmpt_mu_A2empty.pl' on 2020/02/08) (HD & CL & CHK) #####

use strict;
use warnings;

use MyFile_manipu qw(
	open_input_file2
	open_output_file
	list_sessions_ext
	);

use MyLASTPIECE_IO qw(
	readin_indel_params
	readin_2B_graft_convoluted_1st
	readin_mu_Adel_2nd
	);
	# readin_2B_graft_convoluted_2nd
=A BUG fixed on 2020/02/16.
	readin_2B_graft_convoluted_2st
	);
=cut # End of "A BUG fixed on 2020/02/16."
	#
	# print_muP_case_i
	# print_muP0s
	# readin_indel_rates
	# print_muP_case_ii_or_iii 
	# readin_muP0s
	# readin_muP_case_i
	# readin_muP_case_ii_or_iii
	# readin_mu_Adel
	#
	# REMOVED 'readin_muP_case_ii_or_iii' & 'readin_mu_Adel', in creation of 'graft_convolution.pl' (on 2020/02/15). #
	# ADDED 'readin_2B_graft_convoluted_1st' & 'readin_2B_graft_convoluted_2st', in creation of 'graft_convolution.pl' (on 2020/02/15). #
	# => The 'readin_2B_graft_convoluted_2st' was CORRECTED to be 'readin_2B_graft_convoluted_2nd' (on 2020/02/16).
	# => The 'readin_2B_graft_convoluted_2nd' was REPLACED with the 'readin_mu_Adel_2nd', in creation of 'cmpt_mu_dotA2dotAD2dotAdel.pl' (on 2020/02/20). #

##use MyLASTPIECE_Sppl qw(
##	cal_delta_exp_del_5p2
##	);


##sub create_dir_rw (\$); # Added on 2020/01/29.


$| = 1; # Flush the buffer for standard output as soon as it receives data.

my $bin_perl = "/usr/bin/perl"; # The Perl interpreter. # (Added on 2020/01/29). #

my $program_path = $0;
my ($program_name) = ($program_path =~ m!([^/]+)$!);

	# Controlling the level of test outputs. #
	# (0: silent; 1: very quiet; 2: quiet; 3: medium; 4: verbose; 5: very verbose) 
our $TEST_LEVEL = 2;
##our $TEST_LEVEL = 5;

$MyLASTPIECE_IO::TEST_LEVEL = $TEST_LEVEL ; # Added on 2020/01/27. # RESURRECTED on 2020/02/04.


		### END of "(IMPORTED w/ modifications from '../../../lastpiece.alpha.pl' on 2020/01/30) (HD)" ###

=For the predecessor, 'cmpt_mu_Dcr.pl'. (GL)
our $EPS = 1E-07; # A "small" number to be used for checking the consistency. # (Added on 2020/02/01.) #

our $IF_DRESSED = 0; # Whether the input indel rates are raw (=0) or dressed (=1). (The default value = 0.)  # Added in creation of 'cmpt_mu_Dcr.pl' (on 2020/02/04). #
=cut # End of "For the predecessor, 'cmpt_mu_Dcr.pl'. (GL)"

	####### Take in the Command-Line Arguments. #######


		# Modified in creation of 'graft_convolution.pl' (on 2020/02/15) (CL1) #
		# => Further Modified in creation of 'cmpt_mu_dotA2dotAD2dotAdel.pl' (on 2020/02/20) (CL1) #


unless (@ARGV > 8) {
	print STDERR "\nUsage:  [perl] $program_name {initial_time (non-negative number)} {final_time (positive number, greater than initial_time)} {number_of_sub_time_intervals (positive integer)} {upper-bound_of_number_of_sites} {final_number_of_sites_in_D_region_on_the_right (i.e., \$LDR)} {input_file_for_indel_parameters} {input_file_for_muP0s (whose SIGMABE MUST BE 1!!)} {input_file_for_muPs_DcrR2dotAdel (w/ fixed \$LDR)}  {output_file} ";
	print STDERR "\n\n";
	die "$!\n\n";
}
=For the predecessor, 'graft_convolution.pl'. (CL1)
unless (@ARGV > 7) {
	print STDERR "\nUsage:  [perl] $program_name {initial_time (non-negative number)} {final_time (positive number, greater than initial_time)} {number_of_sub_time_intervals (positive integer)} {upper-bound_of_number_of_sites} {input_file_for_indel_parameters} {input_file_for_muPs_2B_graft-convoluted_1st} {input_file_for_muPs_2B_graft-convoluted_2nd}  {output_file} ";
	print STDERR "\n\n";
	die "$!\n\n";
}
=cut # End of "For the predecessor, 'graft_convolution.pl'. (CL1)"
=For the predecessor, 'cmpt_mu_Dcr.pl'. (CL1)
unless (@ARGV > 8) {
	print STDERR "\nUsage:  [perl] $program_name {initial_time (non-negative number)} {final_time (positive number, greater than initial_time)} {number_of_sub_time_intervals (positive integer)} {upper-bound_of_number_of_sites} {SIGMA_BE (MUST ALWAYS BE 2!!)} {input_file_for_indel_parameters} {input_file_for_raw_indel_rates} {input_file for muP0 (w/ SIGMA_BE=2)} {output_file} "; # CORRECTED on 2020/02/06. # 
		# End of "Modified in creation of 'cmpt_muP0.pl' (on 2020/02/03) (CL1)" #
	print STDERR " [ options ] ,\n";
	print STDERR "   where the [options] can be either of the following:\n\n";
	print STDERR "            --RAW      when using the RAW indel rates;\n";
	print STDERR "            --DRESSED      when using the DRESSED indel rates.\n";
	print STDERR "		  (NOTE1: If both options are used, the '--RAW' will be prioritized.)\n"; # (The '\n";' was added in debugging on 2020/02/05.)
	print STDERR "		  (NOTE2: If no options are used, the '--RAW' will be assumed.)\n"; # (The '\n";' was added in debugging on 2020/02/05.)
	##print STDERR "           --XLEVEL=integer       for the level of MSA space exploration (from =1 (performing only multiple-\"shift\"s) to =5 (performing all programed moves) (default=$XLEVEL);\n"; # ADDED on 2019/08/01.
	print STDERR "\n";

	die "$!\n\n";
}
=cut # End of "For the predecessor, 'cmpt_mu_Dcr.pl'. (CL1)"
		# End of "Modified in creation of 'graft_convolution.pl' (on 2020/02/15) (CL1) #
		# => Further Modified in creation of 'cmpt_mu_dotA2dotAD2dotAdel.pl' (on 2020/02/20) (CL1)" #


		# Modified in creation of 'graft_convolution.pl' (on 2020/02/15) (CL2) #
		# => Further Modified in creation of 'cmpt_mu_dotA2dotAD2dotAdel.pl' (on 2020/02/20) (CL2) #

####our ($TIME_INIT, $TIME_FIN, $CT_SB_INTVLS, $UB_CT_SITES_CMPT, $SIGMA_BE) = @ARGV[0 .. 3+1];
####my ($infile_id_prms, $infile_mu_Dcr, $infile_mu_Adel, $ofile_mu_Dcr2Adel) = @ARGV[5 .. 8];
##our ($TIME_INIT, $TIME_FIN, $CT_SB_INTVLS, $UB_CT_SITES_CMPT) = @ARGV[0 .. 3];
##my ($infile_id_prms, $infile_muPs_in_1st, $infile_muPs_in_2nd, $ofile_muPs_out) = @ARGV[3+1 .. 7];
our ($TIME_INIT, $TIME_FIN, $CT_SB_INTVLS, $UB_CT_SITES_CMPT) = @ARGV[0 .. 3];
my ($LDR, $infile_id_prms, $infile_muP0s, $infile_mu_DcrR2dotAdel_fxd_LDR, $ofile_muPs_out_fxd_LDR) = @ARGV[3+1 .. 8];
		#
		# End of "Modified in creation of 'graft_convolution.pl' (on 2020/02/15) (CL2) #
		# => Further Modified in creation of 'cmpt_mu_dotA2dotAD2dotAdel.pl' (on 2020/02/20) (CL2)" #

=For the predecessor, 'cmpt_mu_Dcr.pl'. (CL2)
		# Added in creation of 'cmpt_mu_Dcr.pl' (on 2020/02/04). (CL) #
my ($if_dressed, $if_raw) = (0, 0);

for (my $k=9; $k < @ARGV; $k++) { # Preliminary for-loop (over additional command-line arguments). #

	my $clarg = $ARGV[$k];
	unless ($clarg =~ s/^(\-)+//) { next; } # Skip if the argument lacks preceding '-'s. #

	my ($option, $value) = split /\=/, $clarg;
	unless ((defined $value) and ($value ne '')) {
	     print STDOUT "[CAUTION '$program_name'] Found an UNEXPECTED FORMAT of the commmand-line argument, ", $ARGV[$k], ": $! ! ... I will SKIP it ...\n\n";
	     next; # Skip an argument in an unexpected format.
	}
	my $org_option = $option;
	$option = uc $option; # Capitalize the option, in advance.

	if ($option eq 'RAW') {
	    $if_raw = 1;

	} elsif ($option eq 'DRESSED') {
	    $if_dressed = 1;

	} else {
	    print STDOUT "[CAUTION '$program_name'] Found an UNEXPECTED OPTION, '$org_option', within the cmmmand-line argument, ", $
ARGV[$k], ": $! ! ... I will SKIP it ...\n\n";
	    next; # Skip an argument with an unexpected option.
	}

} # End of the "Preliminary for-loop (over additional command-line arguments)." #

if ($if_dressed and !($if_raw)) {
	$IF_DRESSED = 1;

	print STDOUT "\n[INFORMATION '$program_name'] AS YOU SPECIFIED, we set \$IF_DRESSED to be 1 (true) ...\n\n";
}
		# End of "Added in creation of 'cmpt_mu_Dcr.pl' (on 2020/02/04). (CL)" #
=cut # End of "For the predecessor, 'cmpt_mu_Dcr.pl'. (CL2)"


	####### END of "Take in the Command-Line Arguments." #######


=DISABLED in creation of 'graft_convolution.pl'. (CHK)
	## Check the value of $SIGMA_BE ##
	## (Added in creation of 'cmpt_mu_Dcr.pl' on 2020/02/04.) ##
	## (=> Resurrected w/ modifications in creation of 'cmpt_mu_Dcr2Adel.pl' on 2020/02/13.) ##
##unless ($SIGMA_BE == 2) {
unless ($SIGMA_BE == 1) { # A BUG fixed on 2020/02/14.
	##print STDOUT "[ERROR '$program_name'] The \$SIGMA_BE (=$SIGMA_BE) is NOT 2, which this script requires: $? !\n";
	print STDOUT "[ERROR '$program_name'] The \$SIGMA_BE (=$SIGMA_BE) is NOT 1, which this script requires: $? !\n"; # Modified in accompany with the Bug-Fix (on 2020/02/14).
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}
	## End of "(Added in creation of 'cmpt_mu_Dcr.pl' on 2020/02/04.) ##
	## (=> Resurrected w/ modifications in creation of 'cmpt_mu_Dcr2Adel.pl' on 2020/02/13.)" ##
=cut # End of "DISABLED in creation of 'graft_convolution.pl'. (CHK)"


	### Check the existence & permission mode of the input files. ###

####foreach my $infile ($infile_id_prms, $infile_mu_Dcr, $infile_mu_Adel) { # Yet Further Modified in creation of 'cmpt_mu_Dcr2Adel.pl' (on 2020/02/13). (CHK) #
##foreach my $infile ($infile_id_prms, $infile_muPs_in_1st, $infile_muPs_in_2nd) { # Yet-Yet Further Modified in creation of 'graft_convolution.pl' (on 2020/02/15). (CHK) #
foreach my $infile ($infile_id_prms, $infile_muP0s, $infile_mu_DcrR2dotAdel_fxd_LDR) { # Again, Modified in creation of 'cmpt_mu_dotA2dotAD2dotAdel.pl' (on 2020/02/20). (CHK) #
    unless ( ( -f $infile ) and ( -r $infile ) ) {
	print STDOUT "[ERROR '$program_name'] CanNOT FIND a readable input file, \"$infile\": $? !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
    }
}


	### Declare the Beginning of the Script ###

print STDOUT "\n<<< Program name: $program_name (path: $program_path) >>>\n\n";

print STDOUT '$LDR = ', $LDR, "\n\n"; # ADDED in creation of 'cmpt_mu_dotA2dotAD2dotAdel.pl' (on 2020/02/20). (DC) #

print STDOUT "\n<< Input Files: >>\n\n";

		# Modified in creation of 'graft_convolution.pl' (on 2020/02/15). (DC1) #
		# => Further Modified in creation of 'cmpt_mu_dotA2dotAD2dotAdel.pl' (on 2020/02/20). (DC1) #
		#
####print STDOUT " (for indel parameters) $infile_id_prms ,\n";
####print STDOUT " (for multiplication factors for the creation of a \"D\"-region)  $infile_mu_Dcr ,\n";
####print STDOUT " (for mutiplication factors for the creation of an \"A\"-region) $infile_mu_Adel .\n\n";
##print STDOUT " (for indel parameters) $infile_id_prms ,\n";
##print STDOUT " (for 1st set of multiplication factors to be graft-convoluted)  $infile_muPs_in_1st ,\n";
##print STDOUT " (for 2nd set of multiplication factors to be graft-convoluted)  $infile_muPs_in_2nd .\n\n";
print STDOUT " (for indel parameters) $infile_id_prms ,\n";
print STDOUT " (for the 'Base' multiplication factors (w/ SIGMA_BE = 1!!))  $infile_muP0s ,\n";
print STDOUT " (for the multiplication factors for 'Dcr(R) -> .Adel' (w/ fixed \$LDR(=$LDR))  $infile_mu_DcrR2dotAdel_fxd_LDR .\n\n";
		#
		# End of "Modified in creation of 'graft_convolution.pl' (on 2020/02/15). (DC1) #
		# => Further Modified in creation of 'cmpt_mu_dotA2dotAD2dotAdel.pl' (on 2020/02/20). (DC1)" #


print STDOUT "\n<< Output File: >>\n\n"; 

		# Modified in creation of 'graft_convolution.pl' (on 2020/02/15). (DC2) #
		# => Further Modified in creation of 'cmpt_mu_dotA2dotAD2dotAdel.pl' (on 2020/02/20). (DC2) #
		#
##print STDOUT " (for multiplication factors resulting from the graft-convolution)  $ofile_muPs_out .\n\n";
print STDOUT " (for multiplication factors resulting from the graft-convolution (w/ fixed \$LDR(=$LDR))  $ofile_muPs_out_fxd_LDR .\n\n";
		#
		# End of "Modified in creation of 'graft_convolution.pl' (on 2020/02/15). (DC2) #
		# => Further Modified in creation of 'cmpt_mu_dotA2dotAD2dotAdel.pl' (on 2020/02/20). (DC2)" #


print STDOUT "\n<< Global Variables specifying the levels of tests: >>\n\n"; # Added on 2020/01/30. (DC2) #

print STDOUT "\n", '$TEST_LEVEL = ', $TEST_LEVEL, " .\n\n";

print STDOUT '$MyLASTPIECE_IO::TEST_LEVEL  = ', $MyLASTPIECE_IO::TEST_LEVEL , " .\n\n"; # Added on 2020/02/04.

=For the predecessor, 'cmpt_mu_Dcr.pl'. (DC)
print STDOUT "\n<< A \"small\" number to be used for checking the consistency: >>\n\n"; #
print STDOUT '$EPS = ', $EPS, " .\n\n";

		# Added in creation of 'cmpt_mu_Dcr.pl' (on 2020/02/04). (DC2) #
print STDOUT "\n<< Option that specifies the type of indel rates: >>\n\n";
print STDOUT '$IF_DRESSED = ', $IF_DRESSED, " .\n\n";
		# End of "Added in creation of 'cmpt_mu_Dcr.pl' (on 2020/02/04). (DC2)" #
=cut # End of "For the predecessor, 'cmpt_mu_Dcr.pl'. (DC)"


print STDOUT "\n<< Parameters necessary for recursive computations of the multiplication factors: >>\n\n";

print STDOUT '($TIME_INIT, $TIME_FIN) = ', "($TIME_INIT, $TIME_FIN) .\n\n";
print STDOUT '#{Sub-time-intervals} = ', $CT_SB_INTVLS, " .\n\n";

print STDOUT '{Upper-bound of #{sites} for computation} = ', $UB_CT_SITES_CMPT, " .\n\n"; 

=DISABLED in creation of 'graft_convolution.pl'. (GL)
print STDOUT '$SIGMA_BE = ', $SIGMA_BE, " .\n\n"; # Added in creation of 'cmpt_muP0.pl' (on 2020/02/03). # => Resurrected in creation of 'cmpt_mu_Dcr2Adel.pl' (on 2020/02/13). #
=cut # End of "DISABLED in creation of 'graft_convolution.pl'. (GL)"



	## Preliminary 00-C: Compute Other Important Parameters ##

print STDOUT "\n<<< Preliminary Process 00-C: Computing Other Important Parameters >>>\n\n";

##our $TOTRATE_INDEL = $TOTRATE_INS + $TOTRATE_DEL; ## Disabled in creation of 'cmpt_mu_A2empty.pl' (on 2020/02/08). #
our $SIZE_FULL_INTVL = $TIME_FIN - $TIME_INIT;
our $SIZE_SB_INTVL = $SIZE_FULL_INTVL /$CT_SB_INTVLS;

		# Modified on 2020/01/31. (00-C-1) #
our @TIME_LAPSES_CMPT = (); # Values of the time lapse (= t_F - t_I) at which the factors will be computed.
for (my $i = 0; $i <= $CT_SB_INTVLS; $i++) {
	$TIME_LAPSES_CMPT[$i] = $i * $SIZE_SB_INTVL;
} 
=OBSOLETE as of 2020/01/31. (A)
our @TIME_LAPSES = (); # Values of the time lapse (= t_F - t_I) to be examined.
for (my $i = 0; $i <= $CT_SB_INTVLS; $i++) {
	$TIME_LAPSES[$i] = $i * $SIZE_SB_INTVL;
} 
=cut # End of "OBSOLETE as of 2020/01/31. (A)"
		# End of "Modified on 2020/01/31. (00-C-1)" #

=DISABLED on 2020/01/31. (1)
our @E_INDEL_CTS = (); # Expected counts of indels. #
foreach my $tlapse (@TIME_LAPSES) { push @E_INDEL_CTS, $TOTRATE_INDEL * $tlapse; }
=cut # End of "DISABLED on 2020/01/31. (1)"

##print STDOUT '$TOTRATE_INDEL = ', $TOTRATE_INDEL, " \;\n"; # Disabled in creation of 'cmpt_mu_A2empty.pl' (on 2020/02/08). #
print STDOUT '$SIZE_FULL_INTVL = ', $SIZE_FULL_INTVL, " \;\n";
print STDOUT '$SIZE_SB_INTVL = ', $SIZE_SB_INTVL, " .\n\n";

		# Modified on 2020/01/31. (00-C-2) #
print STDOUT '@TIME_LAPSES_CMPT = (', join (', ', @TIME_LAPSES_CMPT), ") .\n\n";
##print STDOUT '@TIME_LAPSES = (', join (', ', @TIME_LAPSES), ") .\n\n";
		# End of "Modified on 2020/01/31. (00-C-2)" #

=DISABLED on 2020/01/31. (2)
print STDOUT '@E_INDEL_CTS = (', join (', ', @E_INDEL_CTS), ") .\n\n";
=cut # End of "DISABLED on 2020/01/31. (2)"

print STDOUT "\n";



print STDOUT "\n<<< Preliminary Process 00-IN: Reading in the quantities necessary for the computations >>>\n\n";

	## 00-IN-A: Read in the input file for indel parameters. ##
	## (IMPORTED from 'cmpt_mu_Dcr.pl', in creation of 'cmpt_muP_2nd_i.pl' (on 2020/02/10). (00-IN)

print STDOUT "\n<< (Pre 00-IN-A): Reading in the indel model parameters >>\n\n";


our ($TOTRATE_INS, $TOTRATE_DEL, $CUTOFF_INS_LEN, $CUTOFF_DEL_LEN)
   = readin_indel_params ($infile_id_prms);

unless (defined $CUTOFF_DEL_LEN) {
	print STDOUT "[ERROR '$program_name'] The subroutine, 'readin_indel_params', FAILED on the input file, \"$infile_id_prms\" $? !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}
(print STDOUT "\n ... The subroutine, 'readin_indel_params', Ended on the input file, \"$infile_id_prms\" ...\n\n") if ($TEST_LEVEL > 1);


print STDOUT "\n[ Indel Model Parameters retrieved: ]\n\n";

print STDOUT '($TOTRATE_INS, $TOTRATE_DEL) = ', "($TOTRATE_INS, $TOTRATE_DEL) .\n";
print STDOUT '($CUTOFF_INS_LEN, $CUTOFF_DEL_LEN) = ', "($CUTOFF_INS_LEN, $CUTOFF_DEL_LEN) .\n\n";

	## End of "(IMPORTED from 'cmpt_mu_Dcr.pl', in creation of 'cmpt_muP_2nd_i.pl' (on 2020/02/10). (00-IN)"

	## (Newly ADDED in creation of 'cmpt_muP_2nd_i.pl' (on 2020/02/10).) (00-IN) ##


my $ct_tlapses = @TIME_LAPSES_CMPT;
##my $ct_tlapses = $CT_SB_INTVLS +1;
my $cap_ct_sites = $UB_CT_SITES_CMPT +1;

	## End of "(Newly ADDED in creation of 'cmpt_muP_2nd_i.pl' (on 2020/02/10).) (00-IN)" ##


	### Preliminary 00-IN-H1: Reading in the 1st set of multiplication factors to be graft-convoluted.
	### (ADDED in creation of 'graft_convolution.pl' on 2020/02/15). (00-IN-H1) ###

print STDOUT "\n<< (Pre 00-IN-H1): Reading in the 1st set of multiplication factors to be graft-convoluted >>\n\n";

##my ($rf_muPs_1st) =  readin_2B_graft_convoluted_1st ($infile_muPs_in_1st, $CT_SB_INTVLS, $UB_CT_SITES_CMPT);
my ($rf_muPs_1st) =  readin_2B_graft_convoluted_1st ($infile_muP0s, $CT_SB_INTVLS, $UB_CT_SITES_CMPT); # MODIFIED in creation of 'cmpt_mu_dotA2dotAD2dotAdel.pl' (on 2020/02/20). (00-IN-HI) #

	# $rf_muPs_1st->[$LI][$LF][$jt]
	# (w/ 0 < $LI, $LF <= $ub_ct_sites, & 0 <= $jt <= $ct_sb_intvls)
	# is the multiplication factor, in the 1st set to be graft-convoluted,
	# at the $jt th time-lapse with $L_I sites initially and $L sites finally, both in between the PASs.

unless (defined $rf_muPs_1st) {
	print STDOUT "[ERROR '$program_name'] The subroutine, 'readin_2B_graft_convoluted_1st', FAILED on the input file, \"$infile_muP0s\" $? !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}
(print STDOUT "\n ... The subroutine, 'readin_2B_graft_convoluted_1st', Ended on the input file, \"$infile_muP0s\" ...\n\n") if ($TEST_LEVEL > 1);

my $rf_muP0s = $rf_muPs_1st; # ADDED in creation of 'cmpt_mu_dotA2dotAD2dotAdel.pl' (on 2020/02/20). (00-IN-H1). #
	#
	# $rf_muP0s->[$LA][$Lm][$jt]
	# (w/ 0 < $LA, $Lm <= $ub_ct_sites, & 0 <= $jt <= $ct_sb_intvls)
	# is the "Base" multiplication factor (w/ sigmaBE = 1),
	# at the $jt th time-lapse with $L_A sites initially and $Lm sites finally,
	# both in the "A"-region.

=DISABLED in creation of 'cmpt_mu_Dcr2Adel.pl'. (00-IN-D)
my $ct_tlapses = @TIME_LAPSES_CMPT;
##my $ct_tlapses = $CT_SB_INTVLS +1;
my $cap_ct_sites = $UB_CT_SITES_CMPT +1;
=cut # End of "DISABLED in creation of 'cmpt_mu_Dcr2Adel.pl'. (00-IN-D)"

		# Check the consistency of @{$rf_muPs_1st}. # (Added on 2020/02/15.) #

for (my $LI = 1; $LI < $cap_ct_sites; $LI++) { # Preceding outer for-loop (over the initial site-counts). #

	my $muPs_1st_fxd_LI = $rf_muPs_1st->[$LI];

   for (my $LF = 1; $LF < $cap_ct_sites; $LF++) { # Preceding middle for-loop (over the final site-counts). #
	my $ct_tlapses_in = @{$muPs_1st_fxd_LI->[$LF]};

	unless ($ct_tlapses_in == $ct_tlapses) {
		##print STDOUT "[ERROR '$program_name'] The number of time-lapses (=$ct_tlapses_in) in the input ", '@{$rf_muPs_1st->[$LI][$LF]} (w/ ($LI, $LF) =', "($LI, $LF) )  CLASHES with the number of time-lapses for the output (=$ct_tlapses) : $? !\n";
		print STDOUT "[ERROR '$program_name'] The number of time-lapses (=$ct_tlapses_in) in the input ", '@{$rf_muP0s->[$LA][$Lm]} (w/ ($LA, $Lm) =', "($LI, $LF) )  CLASHES with the number of time-lapses for the output (=$ct_tlapses) : $? !\n";
		print STDERR "[ERROR '$program_name'] $!\n";
		die ;
	}
   } # End of the Preceding inner for-loop (over the final site-counts). #

}  # End of the Preceding outer for-loop (over the initial site-counts). #

		# End of "Check the consistency of @{$rf_muPs_1st}. # (Added on 2020/02/15.)" #


if ($TEST_LEVEL >= 5) { # (00-IN-H1) #

			# Modified in creation of 'cmpt_mu_dotA2dotAD2dotAdel.pl' (on 2020/02/20).) #

	print STDOUT "\n\n[[ Multiplication factors, mu_{P0} [ (Lm, tm) | (LA, t_I) ], retrieved (& stored in ", '@{$rf_muP0s}', "): ]]\n\n";

	print STDOUT join ("\t", '$LA:$Lm\$jt', 0 .. $CT_SB_INTVLS), "\n\n";

=For the predecessor, 'graft_convolution.pl'.
	print STDOUT "\n\n[[ Multiplication factors, muP_{1st} [ (L_F, t_F) | (L_I, t_I) ], retrieved (& stored in ", '@{$rf_muPs_1st}', "): ]]\n\n";

	print STDOUT join ("\t", '$LI:$LF\$jt', 0 .. $CT_SB_INTVLS), "\n\n";
=cut # END of "For the predecessor, 'graft_convolution.pl'."
			# End of "Modified in creation of 'cmpt_mu_dotA2dotAD2dotAdel.pl' (on 2020/02/20).)" #


	for (my $LI = 1; $LI < $cap_ct_sites; $LI++) { # Test outer for-loop (over the initial site-counts).

	    for (my $LF = 1; $LF < $cap_ct_sites; $LF++) { # Test middle for-loop (over the final site-counts).
		##print STDOUT join ("\t", "$LI:$$LF", @{$rf_muPs_1st->[$LI][$LF]}) , "\n"; # A BUG fixed on 2020/02/16. #
		print STDOUT join ("\t", "$LI:$LF", @{$rf_muPs_1st->[$LI][$LF]}) , "\n";

	    } # End of the test middle for-loop (over the final site-counts).

	} # End of the Test outer for-loop (over the initial site-counts).

	print STDOUT "\n\n\n";

} # End of "if ($TEST_LEVEL >= 5) { # (00-IN-H1) # ...}"

	### END of "(ADDED in creation of 'graft_convolution.pl' on 2020/02/15). (00-IN-H1)" ###



	### Preliminary 00-IN-I: Reading in the multiplication factors for 'Dcr(R) -> .Adel' (w/ \$LDR = $LDR).
	### (ADDED in creation of 'cmpt_mu_dotA2dotAD2dotAdel.pl' on 2020/02/20). (00-IN-I) ###

print STDOUT "\n<< (Pre 00-IN-I): Reading in the multiplication factors for 'Dcr(R) -> .Adel' (w/ \$LDR = $LDR) >>\n\n";

my ($rf_mu_DcrR2dotAdel_fxd_LDR) =  readin_mu_Adel_2nd ($infile_mu_DcrR2dotAdel_fxd_LDR, $CT_SB_INTVLS, $UB_CT_SITES_CMPT);

	# $rf_mu_DcrR2dotAdel_fxd_LDR->[$dL_DL][$Lm][$jt]
	# (w/ 0 < $Lm <= $ub_ct_sites, 0 <= $dL_DL <= {input-specified upper-bound}, & 0 <= $jt <= $ct_sb_intvls)
	# is the multiplication factors for 'Dcr(R) -> .Adel', 
	# mu_{Dcr(R)-> .Adel} [ (L_A = $Lm, x, [t_I, t_I+dt_I]) ; (-dL_DL = -$dL_DL, L_DR, [t_F-dt_F, t_F]) ](sigmaBE=0),
	# w/ fixed L_DR (=$LDR), and w/ t_F - t_I = $jt * $SIZE_SB_INTVL.
	# 

unless (defined $rf_mu_DcrR2dotAdel_fxd_LDR) {
	print STDOUT "[ERROR '$program_name'] The subroutine, 'readin_mu_Adel_2nd' , FAILED on the input file, \"$infile_mu_DcrR2dotAdel_fxd_LDR\": $? !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}
(print STDOUT "\n ... The subroutine, 'readin_mu_Adel_2nd', Ended on the input file, \"$infile_mu_DcrR2dotAdel_fxd_LDR\" ...\n\n") if ($TEST_LEVEL > 1);


my $cap_dL_DL = @{$rf_mu_DcrR2dotAdel_fxd_LDR};


if ($TEST_LEVEL >= 5) { # (00-IN-I) #

	print STDOUT "\n\n[[ Multiplication factors, mu_{Dcr(R)-> .Adel} [ (L_A = \$Lm, x, [t_I, t_I+dt_I]) ; (-dL_DL = -\$dL_DL, L_DR, [t_F-dt_F,     t_F]) ](sigmaBE=0) , retrieved (& stored in ", '@{$rf_mu_DcrR2dotAdel_fxd_LDR}', "): ]]\n\n";

	##print STDOUT join ("\t", 'L_A:dL_DL\$jt', 0 .. $CT_SB_INTVLS), "\n\n";
	print STDOUT join ("\t", '$Lm:$dL_DL\$jt', 0 .. $CT_SB_INTVLS), "\n\n";


	for (my $LA = 1; $LA < $cap_ct_sites; $LA++) { # Test outer for-loop (over the initial A-site-counts).

	    for (my $dL_DL = 0; $dL_DL < $cap_dL_DL; $dL_DL++) { # Test middle for-loop (over #{sites in "D"(L) deleted simultaneously with "A"). #
		print STDOUT join ("\t", "$LA:$dL_DL", @{$rf_mu_DcrR2dotAdel_fxd_LDR->[$dL_DL][$LA]}) , "\n";

	    }  # End of the test middle for-loop (over #{sites in "D"(L) deleted simultaneously with "A"). #

	} # End of the Test outer for-loop (over the initial A-site-counts).

	print STDOUT "\n\n\n";

} # End of "if ($TEST_LEVEL >= 5) { # (00-IN-I) # ...}"

	### END of "(ADDED in creation of 'cmpt_mu_dotA2dotAD2dotAdel.pl' on 2020/02/20). (00-IN-I)" ###

=For the predecessor, 'graft_convolution.pl'. (00-IN-H2)

	### Preliminary 00-IN-H2: Reading in the 2nd set of multiplication factors to be graft-convoluted.
	### (ADDED in creation of 'graft_convolution.pl' on 2020/02/15). (00-IN-H2) ###

print STDOUT "\n<< (Pre 00-IN-H1): Reading in the 1st set of multiplication factors to be graft-convoluted >>\n\n";

my ($rf_muPs_2nd) =  readin_2B_graft_convoluted_2nd ($infile_muPs_in_2nd, $CT_SB_INTVLS, $UB_CT_SITES_CMPT);

	# $rf_muPs_2nd->[$LF][$LI][$jt]
	# (w/ 0 < $LI, $LF <= $ub_ct_sites, & 0 <= $jt <= $ct_sb_intvls)
	# is the multiplication factor, in the 2nd set to be graft-convoluted,
	# at the $jt th time-lapse with $L_I sites initially and $L sites finally, both in between the PASs.


unless (defined $rf_muPs_2nd) {
	print STDOUT "[ERROR '$program_name'] The subroutine, 'readin_2B_graft_convoluted_2nd', FAILED on the input file, \"$infile_muPs_in_2nd\" $? !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}
(print STDOUT "\n ... The subroutine, 'readin_2B_graft_convoluted_2nd', Ended on the input file, \"$infile_muPs_in_2nd\" ...\n\n") if ($TEST_LEVEL > 1);

		# Check the consistency of @{$rf_muPs_2nd}. # (Added on 2020/02/15.) #

for (my $LF = 1; $LF < $cap_ct_sites; $LF++) { # Preceding outer for-loop (over the final site-counts). #

	my $muPs_2nd_fxd_LF = $rf_muPs_2nd->[$LF];

   for (my $LI = 1; $LI < $cap_ct_sites; $LI++) { # Preceding middle for-loop (over the initial site-counts). #
	my $ct_tlapses_in = @{$muPs_2nd_fxd_LF->[$LI]};

	unless ($ct_tlapses_in == $ct_tlapses) {
		print STDOUT "[ERROR '$program_name'] The number of time-lapses (=$ct_tlapses_in) in the input ", '@{$rf_muPs_2nd->[$LF][$LI]} (w/ ($LI, $LF) =', "($LI, $LF) )  CLASHES with the number of time-lapses for the output (=$ct_tlapses) : $? !\n";
		print STDERR "[ERROR '$program_name'] $!\n";
		die ;
	}
   } # End of the Preceding inner for-loop (over the initial site-counts). #

}  # End of the Preceding outer for-loop (over the finaal site-counts). #

		# End of "Check the consistency of @{$rf_muPs_2nd}. # (Added on 2020/02/15.)" #


if ($TEST_LEVEL >= 5) { # (00-IN-H2) #

	print STDOUT "\n\n[[ Multiplication factors, muP_{2nd} [ (L_F, t_F) | (L_I, t_I) ], retrieved (& stored in ", '@{$rf_muPs_2nd}', "): ]]\n\n";

	print STDOUT join ("\t", '$LI:$LF\$jt', 0 .. $CT_SB_INTVLS), "\n\n";


	for (my $LI = 1; $LI < $cap_ct_sites; $LI++) { # Test outer for-loop (over the initial site-counts).

	    for (my $LF = 1; $LF < $cap_ct_sites; $LF++) { # Test middle for-loop (over the final site-counts).
		##print STDOUT join ("\t", "$LI:$$LF", @{$rf_muPs_2nd->[$LF][$LI]}) , "\n"; # A BUG fixed on 2020/02/16. #
		print STDOUT join ("\t", "$LI:$LF", @{$rf_muPs_2nd->[$LF][$LI]}) , "\n";

	    } # End of the test middle for-loop (over the final site-counts).

	} # End of the Test outer for-loop (over the initial site-counts).

	print STDOUT "\n\n\n";

} # End of "if ($TEST_LEVEL >= 5) { # (00-IN-H2) # ...}"

	### END of "(ADDED in creation of 'graft_convolution.pl' on 2020/02/15). (00-IN-H1)" ###

=cut # END of "For the predecessor, 'graft_convolution.pl'. (00-IN-H2)"


=For the predecessor, 'cmpt_mu_Dcr2Adel.pl'. (Pre 1)
		### ADDED in creation of 'cmpt_mu_Dcr2Adel.pl' (on 2020/02/13). (Pre 1) ###

print STDOUT "\n<<< Preliminary Process 1: Computing Additional Important Parameters >>>\n\n";


our $TOTRATE_INDEL = $TOTRATE_INS + $TOTRATE_DEL;

print STDOUT '$TOTRATE_INDEL = ', $TOTRATE_INDEL, " .\n\n";

my @min_damping_facs = (); # The $l th element = {Minimum damping factor for the sub-sequence of length $l}.
for (my $l=0; $l < $cap_ct_sites; $l++) {
	$min_damping_facs[$l] = exp (- $SIZE_SB_INTVL * $TOTRATE_INDEL * $l);
}

if ($TEST_LEVEL > 3) { # (Pre 1) #

	print STDOUT "\n[ Minimum Damping Factors: ", 
	'$min_damping_facs[$L] = exp (- $SIZE_SB_INTVL * $TOTRATE_INDEL * $L)', " ]\n\n";

	print STDOUT join ("\t", '$L', '$min_damping_facs[$L]'), "\n\n";
	for (my $L=0; $L < $cap_ct_sites; $L++) {
	    print STDOUT join ("\t", $L, $min_damping_facs[$L]), "\n";
	}
	print STDOUT "\n\n";

} # End of "if ($TEST_LEVEL > 3) { # (Pre 1) # ...}"

		### ADDED in creation of 'cmpt_mu_Dcr2Adel.pl' (on 2020/02/13). (Pre 1) ###
=cut # End of "For the predecessor, 'cmpt_mu_Dcr2Adel.pl'. (Pre 1)"



		### ADDED in creation of 'graft_convolution.pl' (on 2020/02/15&16). (MAIN) ###
		### => MODIFIED in creation of 'cmpt_mu_dotA2dotAD2dotAdel.freeze20200220.pl' (on 2020/02/21). (MAIN) ###

	###  Main Process: Compute & Output multiplication factors, 
	### [ muP_{Output} [ (L_F, t_F) | (L_I, t_I) ], ] ... for the predecessor, 'graft_convolution.pl'.
#	### mu_{.A -> .AD -> .Adel} [ (LA, t_I); x_m; (-dL_DL, LDR = $LDR, [t_F - dt_F, t_F]) ] (sigmaBE_m=1, sigmaBE_F = 0),
	### via Graft-Convolution. 



##print STDOUT "\n<<<<<  Main Process: Compute & Output multiplication factors, muP_{Output} [ (L_F, t_F) | (L_I, t_I) ], via Graft-Convolution  >>>>>\n\n\n";
print STDOUT "\n<<<<<  Main Process: Compute & Output multiplication factors, mu_{.A -> .AD -> .Adel} [ (LA, t_I); x_m; (-dL_DL, LDR = $LDR, [t_F - dt_F, t_F]) ] (sigmaBE_m=1, sigmaBE_F = 0) , via Graft-Convolution  >>>>>\n\n\n";

		# (Preliminary) Open a file-handle for outputting to the output file. #

my $fh_out = open_output_file ($ofile_muPs_out_fxd_LDR, 0);
unless (defined $fh_out) {
	print STDOUT "[ERROR '$program_name'] The subroutine 'open_output_file' FAILED to open a file-handle for the file \"$ofile_muPs_out_fxd_LDR\": $? !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}

## my ($LDR, $infile_id_prms, $infile_muP0s, $infile_mu_DcrR2dotAdel_fxd_LDR, $ofile_muPs_out_fxd_LDR) = @ARGV[3+1 .. 8];

####print $fh_out join ("\t", '#{A_sites}_Init:#{D_sites}_Fin\Time_Lapse', @TIME_LAPSES_CMPT), "\n\n";
####print $fh_out join ("\t", '#{sites}_Init(A):#{sites}_Fin(D)\Time_Lapse', @TIME_LAPSES_CMPT), "\n\n";
##print $fh_out join ("\t", '#{sites}_Init:#{sites}_Fin\Time_Lapse', @TIME_LAPSES_CMPT), "\n\n";
print $fh_out join ("\t", '#{A_sites}_Init:#{deleted_sites_in_D(L)}\Time_Lapse', @TIME_LAPSES_CMPT), "\n\n";

			### RESTART FROM HERE ###

for (my $LA = 1; $LA < $cap_ct_sites; $LA++) { # Main Outermost for-loop (over initial A-site-counts). #

		# (Modified in creation of 'cmpt_mu_dotA2dotAD2dotAdel.pl' (on 2020/02/21).) (MAIN A) #
   my $muP0s_fxd_LA = $rf_muP0s->[$LA];

	# $rf_muP0s->[$LA][$Lm][$jt]
	# (w/ 0 < $LA, $Lm <= $ub_ct_sites, & 0 <= $jt <= $ct_sb_intvls)
	# is the "Base" multiplication factor (w/ sigmaBE = 1),
	# at the $jt th time-lapse with $L_A sites initially and $Lm sites finally,
	# both in the "A"-region.

=For the predecessor, 'graft_convolution.pl'. (MAIN 1)
    my $muPs_1st_fxd_LI = $rf_muPs_1st->[$LI];

	# $rf_muPs_1st->[$LI][$LF][$jt]
	# (w/ 0 < $LI, $LF <= $ub_ct_sites, & 0 <= $jt <= $ct_sb_intvls)
	# is the multiplication factor, in the 1st set to be graft-convoluted,
	# at the $jt th time-lapse with $L_I sites initially and $L sites finally, both in between the PASs.
=cut # END of "For the predecessor, 'graft_convolution.pl'. (MAIN 1)"

		# End of "(Modified in creation of 'cmpt_mu_dotA2dotAD2dotAdel.pl' (on 2020/02/21).) (MAIN A)" #

    ####for (my $LF = 1; $LF < $cap_ct_sites; $LF++) { # Main outer for-loop (over final site-counts). #
    ##for (my $dL_DL = 1; $dL_DL < $cap_dL_DL; $dL_DL++) { # Main outer for-loop (over #{sites in "D"(L) deleted simultaneously with "A"). # Modified in creation of 'cmpt_mu_dotA2dotAD2dotAdel.pl' (on 2020/02/21). (MAIN B) #
    for (my $dL_DL = 0; $dL_DL < $cap_dL_DL; $dL_DL++) { # Main outer for-loop (over #{sites in "D"(L) deleted simultaneously with "A"). # => Fixed a BUG (on 2020/02/21). (MAIN B-CORR) #

		# (Modified in creation of 'cmpt_mu_dotA2dotAD2dotAdel.pl' (on 2020/02/21).) (MAIN B) #
	my $mu_DcrR2dotAdel_fxd_LDR_dL_DL = $rf_mu_DcrR2dotAdel_fxd_LDR->[$dL_DL];

	# $rf_mu_DcrR2dotAdel_fxd_LDR->[$dL_DL][$Lm][$jt]
	# (w/ 0 < $Lm <= $ub_ct_sites, 0 <= $dL_DL <= {input-specified upper-bound}, & 0 <= $jt <= $ct_sb_intvls)
	# is the multiplication factors for 'Dcr(R) -> .Adel', 
	# mu_{Dcr(R)-> .Adel} [ (L_A = $Lm, x, [t_I, t_I+dt_I]) ; (-dL_DL = -$dL_DL, L_DR, [t_F-dt_F, t_F]) ](sigmaBE=0),
	# w/ fixed L_DR (=$LDR), and w/ t_F - t_I = $jt * $SIZE_SB_INTVL.

=For the predecessor, 'graft_convolution.pl'. (MAIN 2)
	my $muPs_2nd_fxd_LF = $rf_muPs_2nd->[$LF];

	# $rf_muPs_2nd->[$LF][$LI][$jt]
	# (w/ 0 < $LI, $LF <= $ub_ct_sites, & 0 <= $jt <= $ct_sb_intvls)
	# is the multiplication factor, in the 2nd set to be graft-convoluted,
	# at the $jt th time-lapse with $L_I sites initially and $L sites finally, both in between the PASs.
=cut # END of "For the predecessor, 'graft_convolution.pl'. (MAIN 2)"

		# End of "(Modified in creation of 'cmpt_mu_dotA2dotAD2dotAdel.pl' (on 2020/02/21).) (MAIN B)" #

	my @vec_muPs_out = (); # The $jt th element = the resulting muP at the $jt th time-lapse.

	for (my $jt=0; $jt < $ct_tlapses; $jt++) { # Main middle for-loop (over output time-lapses). #

	    my $muP_out = 0;

	    for (my $lm = 1; $lm < $cap_ct_sites; $lm++) { # Main inner for-loop (over intermediate site-counts). #

		# (Modified in creation of 'cmpt_mu_dotA2dotAD2dotAdel.pl' (on 2020/02/21).) (MAIN C) #
		my $vec_muP0s = $muP0s_fxd_LA->[$lm];
		my $vec_mu_DcrR2dotAdel = $mu_DcrR2dotAdel_fxd_LDR_dL_DL->[$lm];

		for (my $jt2=0; $jt2 <= $jt; $jt2++) { # Main innermost for-loop (over intermediate time-lapses). #
		    $muP_out += $vec_muP0s->[$jt2] * $vec_mu_DcrR2dotAdel->[$jt - $jt2];

		} # End of the Main innermost for-loop (over intermediate time-lapses). #

=For the predecessor, 'graft_convolution.pl'. (MAIN 3)
		my $vec_muPs_1st = $muPs_1st_fxd_LI->[$lm];
		my $vec_muPs_2nd = $muPs_2nd_fxd_LF->[$lm];

		for (my $jt2=0; $jt2 <= $jt; $jt2++) { # Main innermost for-loop (over intermediate time-lapses). #
		    $muP_out += $vec_muPs_1st->[$jt2] * $vec_muPs_2nd->[$jt - $jt2];

		} # End of the Main innermost for-loop (over intermediate time-lapses). #
=cut # END of "For the predecessor, 'graft_convolution.pl'. (MAIN 3)"
		# End of "(Modified in creation of 'cmpt_mu_dotA2dotAD2dotAdel.pl' (on 2020/02/21).) (MAIN C)" #

	    } # End of the Main inner for-loop (over intermediate site-counts). #

	    $vec_muPs_out[$jt] = $muP_out;

	} # End of the Main middle for-loop (over output time-lapses). #

	##print $fh_out join ("\t", "$LI:$LF", @vec_muPs_out), "\n" ;
	print $fh_out join ("\t", "$LA:$dL_DL", @vec_muPs_out), "\n" ; # (Modified in creation of 'cmpt_mu_dotA2dotAD2dotAdel.pl' (on 2020/02/21).) (MAIN D) #

    } # End of the Main outer for-loop (over #{sites in "D"(L) deleted simultaneously with "A"). # Modified in creation of 'cmpt_mu_dotA2dotAD2dotAdel.pl' (on 2020/02/21). (MAIN B) #
    ##} # End of the Main outer for-loop (over final site-counts). #

} # End of the Main Outermost for-loop (over initial A-site-counts). #


close ($fh_out);

print STDOUT "\n\n === All The Processes of '$program_name' FINISHED !! ===\n\n\n";

		### END of "ADDED in creation of 'graft_convolution.pl' (on 2020/02/15&16). (MAIN) ###
		### => MODIFIED in creation of 'cmpt_mu_dotA2dotAD2dotAdel.freeze20200220.pl' (on 2020/02/21). (MAIN)" ###


=For the predecessor, 'cmpt_mu_Dcr2Adel.pl'. (MAIN)

		### ADDED in creation of 'cmpt_mu_Dcr2Adel.pl' (on 2020/02/13&14). (MAIN) ###

	###  Main Process: Compute & Output multiplication factors, 
	### 'mu_{Dcr->Adel} [ (L_A, x(=L_A), [t_I, t_I+dt_I]) ; (L_D, [t_F-dt_F, t_F]) ](sigmaBE=1)'. ###

print STDOUT "\n<<<<< Main Process: Compute & Output multiplication factors, 'mu_{Dcr->Adel} [ (L_A, x(=L_A), [t_I, t_I+dt_I]) ; (L_D, [t_F-dt_F, t_F]) ](sigmaBE=1)' >>>>>\n\n\n";


		# (Preliminary) Open a file-handle for outputting to the output file. #

my $fh_out = open_output_file ($ofile_mu_Dcr2Adel, 0);
unless (defined $fh_out) {
	print STDOUT "[ERROR '$program_name'] The subroutine 'open_output_file' FAILED to open a file-handle for the file \"$ofile_mu_Dcr2Adel\": $? !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}

print $fh_out join ("\t", '#{A_sites}_Init:#{D_sites}_Fin\Time_Lapse', @TIME_LAPSES_CMPT), "\n\n";
##print $fh_out join ("\t", '#{sites}_Init(A):#{sites}_Fin(D)\Time_Lapse', @TIME_LAPSES_CMPT), "\n\n";
##print $fh_out join ("\t", '#{sites}_Init:#{sites}_Fin\Time_Lapse', @TIME_LAPSES_CMPT), "\n\n";

for (my $LA = 1; $LA < $cap_ct_sites; $LA++) { # Main Outermost for-loop (over initial A-site-counts). #

    my $mu_Adel_fxd_LA = $rf_mu_Adel->[$LA];

	# $rf_mu_Adel->[$LA][$jt][$dL_fD] = (w/ 0 < $LA < $cap_ct_sites, & 0 <= $jt < $ct_tlapses & 0 <= $dL_fD <= {upper-bound specified by the input} ) 
	# is the multiplication factor that an "A"-region (with $LI initial sites) was deleted
	# at the $jt th time-lapse (actually, between $jt-1 and $jt),
	# in conjunction with $dL_fD sites in the flanking "D"-region(s).


    for (my $LD = 1; $LD < $cap_ct_sites; $LD++) { # Main outer for-loop (over final D-site-counts). #

	my @vec_mu_Adel2Dcr = (0); # The $jt-th element = mu_{Adel->Dcr} [...] at the $jt-th time-lapse.
				   # (When $jt = 0, the factor is always zero.)

	for (my $jt=1; $jt < $ct_tlapses; $jt++) { # Main middle for-loop (over output time-lapses). #

	    my $mu_Adel2Dcr = 0;

	    my $ub_dL_fD = $UB_CT_SITES_CMPT - $LD;
	    my $cap_dL_fD = ($ub_dL_fD +1 < $cap_dL_fD0) ? $ub_dL_fD + 1 : $cap_dL_fD0 ;

#my $cap_dL_fD0 = ($SIGMA_BE == 2) ? 0 : 
#                (($CUTOFF_DEL_LEN -1 < 2 * ($UB_CT_SITES_CMPT -1)) ?  $CUTOFF_DEL_LEN -1 : 2 * ($UB_CT_SITES_CMPT -1) ) ; 
#$cap_dL_fD0++;

	    for (my $dL_fD = 0; $dL_fD < $cap_dL_fD; $dL_fD++) { # Main inner for-loop (over #{flanking D-sites deleted simultaneously with the "A"-region). #

		$mu_Adel2Dcr += $rf_mu_Dcr->[$LD + $dL_fD][$jt-1] * $min_damping_facs[$LD + $dL_fD] * $mu_Adel_fxd_LA->[$jt][$dL_fD];

		# $rf_mu_Dcr->[$LF][$jt] (w/ 0< $LF <= $ub_ct_sites  & 0 <= $jt <= $ct_sb_intvls)
		# is the multiplication factor for the creation of a "D"-region
		# at the $jt th time-interval with $LF final sites in between the PASs.


	    } # Main inner for-loop (over #{flanking D-sites deleted simultaneously with the "A"-region). #

	    $vec_mu_Adel2Dcr[$jt] = $mu_Adel2Dcr;

	} # End of the Main middle for-loop (over output time-lapses). #

	##print $fh_out join ("\t", "$LI:$LF", @vec_muP_2nd_i), "\n";

	print $fh_out join ("\t", "$LA:$LD", @vec_mu_Adel2Dcr), "\n";

    }  # Main outer for-loop (over final D-site-counts). #

}  # Main Outermost for-loop (over initial A-site-counts). #

close ($fh_out);

print STDOUT "\n\n === All The Processes of '$program_name' FINISHED !! ===\n\n\n";

		### END of "ADDED in creation of 'cmpt_mu_Dcr2Adel.pl' (on 2020/02/13&14). (MAIN)" ###

=cut # END of "For the predecessor, 'cmpt_mu_Dcr2Adel.pl'. (MAIN)"


=For the predecessor, 'cmpt_mu_A2empty.pl'. (MAIN)

		### IMPORTED w/ modifications from '../../Basic/mu_Adel/cmpt_mu_Adel.pl' (on 2020/02/08). (1st-half) ###

	###  Main Process: Compute & Output \"extended\" multiplication factors, 'mu_{A->emptyset} [ (L_I, t_I); (0, 0) ; t_2](sigmaBE=2)'. ###

print STDOUT "\n<<<<< Main Process: Compute & Output \"extended\" multiplication factors, 'mu_{A->emptyset} [ (L_I, t_I); (0, 0) ; t_2](sigmaBE=2)' >>>>>\n\n\n";

#	mu_{A->emptyset} [ (L_I, t_I); (0, 0) ; t_2](sigmaBE=2),
#	where L_I is the number of sites in the "A"-region at the 'initial' time t_I,
#	and t_2 is some time-point after the deletion of the "A"-region.


		# (Preliminary) Open a file-handle for outputting to the output file. #
my $fh_out = open_output_file ($ofile_mu_A2empty, 0);
unless (defined $fh_out) {
	print STDOUT "[ERROR '$program_name'] The subroutine 'open_output_file' FAILED to open a file-handle for the file \"$ofile_mu_A2empty\": $? !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}
				## $ofile_mu_A2empty

print $fh_out join ("\t", '#{sites}_Init\Time_Lapse', @TIME_LAPSES_CMPT), "\n\n";

##my $ct_tlapses = @TIME_LAPSES_CMPT;
##my $cap_ct_sites = $UB_CT_SITES_CMPT +1;

for (my $LI = 1; $LI < $cap_ct_sites; $LI++) { # Main Outermost for-loop (over initial site-counts). #

    my $mu_Adel_fxd_LI = $rf_mu_Adel->[$LI];

	# $rf_mu_Adel->[$LI][$jt][$dL_fD] = (w/ 0 < $L_I <= $ub_ct_sites, & 0 <= $jt <= $ct_sb_intvls & 0 <= $dL_fD <= {upper-bound specified by the input (= 0 here)} ) 
	# is the multiplication factor that an "A"-region (with $LI initial sites) was deleted
	# at the $jt th time-lapse (actually, between $jt-1 and $jt),
	# in conjunction with $dL_fD sites in the flanking "D"-region(s).

		### END of "IMPORTED w/ modifications from '../../Basic/mu_Adel/cmpt_mu_Adel.pl' (on 2020/02/08). (1st-half)" ###

		### Newly Added in creation of 'cmpt_mu_A2empty.pl' (on 2020/02/08). ###

    my @vec_mu_A2empty = ();

    for (my $jt=0; $jt < $ct_tlapses; $jt++) { # Main middle for-loop (over output time-lapses). #

	my $mu_A2empty = 0;

	for (my $jt2=0; $jt2 <= $jt; $jt2++) { # Main inner for-loop (over intermediate time-lapses). #

	    $mu_A2empty += $mu_Adel_fxd_LI->[$jt2][0] * $rf_muP_case_i->[$jt - $jt2] ;

		# $rf_muP_case_i->[$jt] (w/ 0 <= $jt <= $ct_sb_intvls)
		# is the multiplication factor of case-i gapped segments, which does not have any ancestral or descendant sites in between the PASs,
		# at the $jt th time-lapse.

	} # Main inner for-loop (over intermediate time-lapses). #

	$vec_mu_A2empty[$jt] = $mu_A2empty;

    } # End of the Main middle for-loop (over output time-lapses). #

		### End of "Newly Added in creation of 'cmpt_mu_A2empty.pl' (on 2020/02/08)." ###

		### IMPORTED w/ modifications from '../../Basic/mu_Adel/cmpt_mu_Adel.pl' (on 2020/02/08). (2nd-half) ###

    print $fh_out join ("\t", $LI, @vec_mu_A2empty), "\n";

} # End of the Main Outermost for-loop (over initial site-counts). #

close ($fh_out);

print STDOUT "\n\n === All The Processes of '$program_name' FINISHED !! ===\n\n\n";

		### END of "IMPORTED w/ modifications from '../../Basic/mu_Adel/cmpt_mu_Adel.pl' (on 2020/02/08). (2nd-half)" ###

=cut # END of "For the predecessor, 'cmpt_mu_A2empty.pl'. (MAIN)"



