#!/usr/bin/perl
#
###################################################################################
#	
#	LASTPIECE_P (Local Alignment-STate Probabilities that Insertion-type and dEletion-type gaps Co-Exist, type P):
#	    A package of programs to compute the probabilities of gapped segments in each of which an insertion-type gap and a deletion-type gap co-exist, (which were referred to as case-(iv) gapped segments by Ezawa (2016a?)),
#	    under a stochastic model of sequence evolution with biologically realistic insertions/deletions.
#	   (Written almost exclusively in Perl.)
#
#	Version 0.3: Copyright (C) 2020 Kiyoshi Ezawa
#
#    This package is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This package is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License, "GNU_GPL.txt",
#    along with this package.  If not, see <http://www.gnu.org/licenses/>.
#
#    The author can be contacted by e-mailing to <kezawa dot ezawa3 at gmail dot com>
#    (replace " dot " and " at " with "." and "@", respectively).
#
###################################################################################
#
# cmpt_mu_Dcr2Berdot.pl:
#	This servant script computes the multiplication factors for 'D-cr -> B-er.':
#	mu_{Dcr->Ber.} [(x=1, [t_I, t_I + dt_I]) ;  (LDF, -delta{}LAF, [t_F - dt_F, t_F]) ].
#	Here, LDF is the number of sites in the "D"-region at t_F,
#	delta{}LAF is the number of sites in the "A"-region deleted
#	via 'boundary-erosion' at some time in [t_F - dt_F, t_F].
#	The script was created by considerably modifying the predecessor servant script,
#	'cmpt_mu_Adel.pl'.
#	[CREATION STARTED on 2020/02/24, ENDED on 2020/02/25.]
#
# cmpt_mu_Adel.pl:
#	This servant script computes the multiplication factors
#	for the complete deletion of an "A"-region:
#	mu_{A-del} [ (L_I, t_1) ; (0, -delta{}L_{fD}, [t_2 - Delta{}t, t_2]) ] (sigmaBE),
#	where L_I is the number of sites in the "A"-region at t_2,
#	-delta{}L_{fD} is the number of sites in flanking "D"-region(s)
#	that were deleted in conjunction with the "A"-region,
#	and [t_2- Delta{}t, t_2] is the time-interval within which the A-deletion occurred.
#	sigmaBE is the number of boundaries the insertions at which are incorporated into the "A"-region.
#	The script was created by taking advantage of the main framework of another servant script,
#	'cmpt_mu_Dcr.pl' (in "../mu_Dcr/") .
#	[CREATION STARTED on 2020/02/06, ENDED on 2020/02/08.]
#
#	[CORRECTED on 2020/02/24:]
#	Now, the FORGOTTEN damping factor, exp ( - Delta{}t * Delta{}R^{ID} ($L, t) ),
#	is correctly multiplied when computing the mu_{A-del} [...] w/ the "--DRESSED" option.
#

	##### (IMPORTED w/ modifications from '../mu_Dcr/cmpt_mu_Dcr.pl' on 2020/02/06) (HD & CL & CHK) #####

		### (IMPORTED w/ modifications from '../../../lastpiece.alpha.pl' on 2020/01/30) (HD) ###

use strict;
use warnings;

use MyFile_manipu qw(
	open_input_file2
	open_output_file
	list_sessions_ext
	);

use MyLASTPIECE_IO qw(
	readin_indel_params
	readin_indel_rates
	readin_muP_case_ii_or_iii
	);
	# print_muP_case_i
	# print_muP0s
	#
	# Added 'readin_indel_params' & 'readin_indel_rates' on 2020/01/31. #
	# Added 'print_muP_case_i' & 'print_muP_case_ii_or_iii' on 2020/02/01. # => Removed on 2020/02/03. 
	# => Resurrected 'print_muP_case_ii_or_iii', in creation of 'cmpt_mu_Dcr.pl' (on 2020/02/05).
	# => REMOVED again in creation of 'cmpt_mu_Dcr2Berdot.pl' (on 2020/02/24).
	#
	# Added 'print_muP0s' on 2020/02/03. => Removed on 2020/02/04.
	# Added 'readin_muP0s' on 2020/02/04. => REMOVED in creation of 'cmpt_mu_Dcr2Berdot.pl' (on 2020/02/24). #
	#
	# ADDED 'readin_muP_case_ii_or_iii' in creation of 'cmpt_mu_Dcr2Berdot.pl' (on 2020/02/24).

##use MyLASTPIECE_Sppl qw(
##	cal_delta_exp_del_5p2
##	);


##sub create_dir_rw (\$); # Added on 2020/01/29.


$| = 1; # Flush the buffer for standard output as soon as it receives data.

my $bin_perl = "/usr/bin/perl"; # The Perl interpreter. # (Added on 2020/01/29). #

my $program_path = $0;
my ($program_name) = ($program_path =~ m!([^/]+)$!);

	# Controlling the level of test outputs. #
	# (0: silent; 1: very quiet; 2: quiet; 3: medium; 4: verbose; 5: very verbose) 
our $TEST_LEVEL = 2;
#our $TEST_LEVEL = 5;

$MyLASTPIECE_IO::TEST_LEVEL = $TEST_LEVEL ; # Added on 2020/01/27. # RESURRECTED on 2020/02/04.


		### END of "(IMPORTED w/ modifications from '../../../lastpiece.alpha.pl' on 2020/01/30) (HD)" ###

our $EPS = 1E-07; # A "small" number to be used for checking the consistency. # (Added on 2020/02/01.) #


our $IF_DRESSED = 0; # Whether the input indel rates are raw (=0) or dressed (=1). (The default value = 0.)  # Added in creation of 'cmpt_mu_Dcr.pl' (on 2020/02/04). #

	####### Take in the Command-Line Arguments. #######


		# MODIFIED in creation of 'cmpt_mu_Dcr2Berdot.pl' (on 2020/02/24). (CL1) #
unless (@ARGV > 7) {
	print STDERR "\nUsage:  [perl] $program_name {initial_time (non-negative number)} {final_time (positive number, greater than initial_time)} {number_of_sub_time_intervals (positive integer)} {upper-bound_of_number_of_sites} {input_file_for_indel_parameters} {input_file_for_raw_indel_rates} {input_file for mu_{Dcr}} {output_file} ";

=For the predecessor, 'cmpt_mu_Adel.pl'. (CL1)
unless (@ARGV > 8) {
	print STDERR "\nUsage:  [perl] $program_name {initial_time (non-negative number)} {final_time (positive number, greater than initial_time)} {number_of_sub_time_intervals (positive integer)} {upper-bound_of_number_of_sites} {SIGMA_BE (=0,1,or 2)} {input_file_for_indel_parameters} {input_file_for_raw_indel_rates} {input_file for muP0 (w/ SIGMA_BE=2)} {output_file} ";
=cut # End of "For the predecessor, 'cmpt_mu_Adel.pl'. (CL1)"
		# End of "MODIFIED in creation of 'cmpt_mu_Dcr2Berdot.pl' (on 2020/02/24). (CL1)" #

	print STDERR " [ options ] ,\n";
	print STDERR "   where the [options] can be either of the following:\n\n";
	print STDERR "            --RAW      when using the RAW indel rates;\n";
	print STDERR "            --DRESSED      when using the DRESSED indel rates.\n";
	print STDERR "		  (NOTE1: If both options are used, the '--RAW' will be prioritized.)\n"; # (The '\n";' was added in debugging on 2020/02/05.)
	print STDERR "		  (NOTE2: If no options are used, the '--RAW' will be assumed.)\n"; # (The '\n";' was added in debugging on 2020/02/05.)
	##print STDERR "           --XLEVEL=integer       for the level of MSA space exploration (from =1 (performing only multiple-\"shift\"s) to =5 (performing all programed moves) (default=$XLEVEL);\n"; # ADDED on 2019/08/01.
	print STDERR "\n";

	die "$!\n\n";
}

		# MODIFIED in creation of 'cmpt_mu_Dcr2Berdot.pl' (on 2020/02/24). (CL2) #
##our ($TIME_INIT, $TIME_FIN, $CT_SB_INTVLS, $UB_CT_SITES_CMPT, $SIGMA_BE) = @ARGV[0 .. 3+1];
##my ($infile_id_prms, $infile_id_rates, $infile_muP0, $ofile_mu_Adel) = @ARGV[5 .. 8];
our ($TIME_INIT, $TIME_FIN, $CT_SB_INTVLS, $UB_CT_SITES_CMPT) = @ARGV[0 .. 3];
my ($infile_id_prms, $infile_id_rates, $infile_mu_Dcr, $ofile_mu_Dcr2Berdot) = @ARGV[3+1 .. 7];
		# End of "MODIFIED in creation of 'cmpt_mu_Dcr2Berdot.pl' (on 2020/02/24). (CL2)" #



		# Added in creation of 'cmpt_mu_Dcr.pl' (on 2020/02/04). (CL) #
my ($if_dressed, $if_raw) = (0, 0);

##for (my $k=9; $k < @ARGV; $k++) { # Preliminary for-loop (over additional command-line arguments). #
for (my $k=8; $k < @ARGV; $k++) { # Preliminary for-loop (over additional command-line arguments). # => Modified in creation of 'cmpt_mu_Dcr2Berdot.pl' (on 2020/02/25). #

	my $clarg = $ARGV[$k];
	unless ($clarg =~ s/^(\-)+//) { next; } # Skip if the argument lacks preceding '-'s. #

	my ($option, $value) = split /\=/, $clarg;

=DISABLED in DEBUGGING (on 2020/02/25). (CL)
	unless ((defined $value) and ($value ne '')) {
	     print STDOUT "[CAUTION '$program_name'] Found an UNEXPECTED FORMAT of the commmand-line argument, ", $ARGV[$k], ": $! ! ... I will SKIP it ...\n\n";
	     next; # Skip an argument in an unexpected format.
	}
=cut # ENd of "DISABLED in DEBUGGING (on 2020/02/25). (CL)"

	my $org_option = $option;
	$option = uc $option; # Capitalize the option, in advance.

	if ($option eq 'RAW') {
	    $if_raw = 1;

	} elsif ($option eq 'DRESSED') {
	    $if_dressed = 1;

	} else {
	    print STDOUT "[CAUTION '$program_name'] Found an UNEXPECTED OPTION, '$org_option', within the cmmmand-line argument, ", $
ARGV[$k], ": $! ! ... I will SKIP it ...\n\n";
	    next; # Skip an argument with an unexpected option.
	}

} # End of the "Preliminary for-loop (over additional command-line arguments)." #

if ($if_dressed and !($if_raw)) {
	$IF_DRESSED = 1;

	print STDOUT "\n[INFORMATION '$program_name'] AS YOU SPECIFIED, we set \$IF_DRESSED to be 1 (true) ...\n\n";
}
		# End of "Added in creation of 'cmpt_mu_Dcr.pl' (on 2020/02/04). (CL)" #

	####### END of "Take in the Command-Line Arguments." #######

=For the predecessor, 'cmpt_mu_Adel.pl'. (CHK)
	## Check the value of $SIGMA_BE ##
	## (Added in creation of 'cmpt_mu_Dcr.pl' on 2020/02/04.) ##
	## (=> Modified in creation of 'cmpt_mu_Adel.pl' on 2020/02/06.) ##
##unless ($SIGMA_BE == 2) {
unless ($SIGMA_BE =~ m/^[012]$/) {
	##print STDOUT "[ERROR '$program_name'] The \$SIGMA_BE (=$SIGMA_BE) is NOT 2, which this script requires: $? !\n";
	print STDOUT "[ERROR '$program_name'] The \$SIGMA_BE (=$SIGMA_BE) is NEITHER 0, 1, NOR 2, which this script requires: $? !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}
	## End of "(Added in creation of 'cmpt_mu_Dcr.pl' on 2020/02/04.) ##
	## (=> Modified in creation of 'cmpt_mu_Adel.pl' on 2020/02/06.)" ##
=cut # End of "For the predecessor, 'cmpt_mu_Adel.pl'. (CHK)"


	### Check the existence & permission mode of the input files. ###
	### (Added on 2020/02/03.) ###

##foreach my $infile ($infile_id_prms, $infile_id_rates, $infile_muP0) {
foreach my $infile ($infile_id_prms, $infile_id_rates, $infile_mu_Dcr) { # Modified in creation of 'cmpt_mu_Dcr2Berdot.pl' (on 2020/02/24). #
    unless ( ( -f $infile ) and ( -r $infile ) ) {
	print STDOUT "[ERROR '$program_name'] CanNOT FIND a readable input file, \"$infile\": $? !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
    }
}
	### End of "(Added on 2020/02/03.)" ###

	##### END of "(IMPORTED w/ modifications from '../mu_Dcr/cmpt_mu_Dcr.pl' on 2020/02/06) (HD & CL & CHK)" #####



	### Declare the Beginning of the Script ###
	### (IMPORTED w/ modifications from '../mu_Dcr/cmpt_mu_Dcr.pl' on 2020/02/06) (DC) ###

print STDOUT "\n<<< Program name: $program_name (path: $program_path) >>>\n\n";

print STDOUT "\n<< Input Files: >>\n\n";

print STDOUT " (for indel parameters) $infile_id_prms ,\n";
print STDOUT " (for raw indel rates)  $infile_id_rates ,\n";
##print STDOUT " (for muP0 under \"base\" rate operator)  $infile_muP0 .\n\n"; # Added in creation of 'cmpt_mu_Dcr.pl' (on 2020/02/04). (DC1) #
print STDOUT " (for multiplication factors for the creation of a \"D\"-region)  $infile_mu_Dcr .\n\n"; # Modified in creation of 'cmpt_mu_Dcr.pl' (on 2020/02/04). (DC1) #

print STDOUT "\n<< Output File: >>\n\n"; 

##print STDOUT " (for multiplication factors for the deletion of an \"A\"-region)  $ofile_mu_Adel .\n\n";
print STDOUT " (for multiplication factors for 'D-cr -> B-er.')  $ofile_mu_Dcr2Berdot .\n\n";


print STDOUT "\n<< Global Variables specifying the levels of tests: >>\n\n"; # Added on 2020/01/30. (DC2) #

print STDOUT "\n", '$TEST_LEVEL = ', $TEST_LEVEL, " .\n\n";

print STDOUT '$MyLASTPIECE_IO::TEST_LEVEL  = ', $MyLASTPIECE_IO::TEST_LEVEL , " .\n\n"; # Added on 2020/02/04.

print STDOUT "\n<< A \"small\" number to be used for checking the consistency: >>\n\n"; #
print STDOUT '$EPS = ', $EPS, " .\n\n";

		# Added in creation of 'cmpt_mu_Dcr.pl' (on 2020/02/04). (DC2) #
print STDOUT "\n<< Option that specifies the type of indel rates: >>\n\n";
print STDOUT '$IF_DRESSED = ', $IF_DRESSED, " .\n\n";
		# End of "Added in creation of 'cmpt_mu_Dcr.pl' (on 2020/02/04). (DC2)" #


print STDOUT "\n<< Parameters necessary for recursive computations of the multiplication factors: >>\n\n";

print STDOUT '($TIME_INIT, $TIME_FIN) = ', "($TIME_INIT, $TIME_FIN) .\n\n";
print STDOUT '#{Sub-time-intervals} = ', $CT_SB_INTVLS, " .\n\n";

print STDOUT '{Upper-bound of #{sites} for computation} = ', $UB_CT_SITES_CMPT, " .\n\n"; 

=For the predecessor, 'cmpt_mu_Adel.pl'. (DC)
print STDOUT '$SIGMA_BE = ', $SIGMA_BE, " .\n\n"; # Added in creation of 'cmpt_muP0.pl' (on 2020/02/03). #
=cut # End of "For the predecessor, 'cmpt_mu_Adel.pl'. (DC)"

	### End of "(IMPORTED w/ modifications from '../mu_Dcr/cmpt_mu_Dcr.pl' on 2020/02/06) (DC)" ###


	##### (IMPORTED from '../muP_case_i2iii/cmpt_muP_case_i2iii.pl' on 2020/02/03) (Pre 00) #####
	##### (==> IMPORTED w/ modifications from '../muP0/cmpt_muP0.pl' on 2020/02/04&05) (Pre 00) #####
	##### (==> IMPORTED from '../mu_Dcr/cmpt_mu_Dcr.pl' on 2020/02/06) (Pre 00) #####


	## (MODIFIED on 2020/01/31) (00-IN) ##

print STDOUT "\n<<< Preliminary Process 00-IN: Reading in the quantities necessary for the computations >>>\n\n";

	## 00-IN-A: Read in the input file for indel parameters. ##

print STDOUT "\n<< (Pre 00-IN-A): Reading in the indel model parameters >>\n\n";

our ($TOTRATE_INS, $TOTRATE_DEL, $CUTOFF_INS_LEN, $CUTOFF_DEL_LEN)
   = readin_indel_params ($infile_id_prms);

unless (defined $CUTOFF_DEL_LEN) {
	print STDOUT "[ERROR '$program_name'] The subroutine, 'readin_indel_params', FAILED on the input file, \"$infile_id_prms\" $? !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}
(print STDOUT "\n ... The subroutine, 'readin_indel_params', Ended on the input file, \"$infile_id_prms\" ...\n\n") if ($TEST_LEVEL > 1);

##my ($infile_id_prms, $infile_id_rates, $ofile_muP_case_i, $ofile_muP_case_ii, $ofile_muP_case_iii) = @ARGV[4 .. 8];

print STDOUT "\n[ Indel Model Parameters retrieved: ]\n\n";

print STDOUT '($TOTRATE_INS, $TOTRATE_DEL) = ', "($TOTRATE_INS, $TOTRATE_DEL) .\n";
print STDOUT '($CUTOFF_INS_LEN, $CUTOFF_DEL_LEN) = ', "($CUTOFF_INS_LEN, $CUTOFF_DEL_LEN) .\n\n";


	## Preliminary 00-B: Read in the input file for indel rates. ##

		# Modified in creation of 'cmpt_mu_Dcr.pl' (on 2020/02/04). (00-B-1) #
if ($IF_DRESSED) {
	print STDOUT "\n<< (Pre 00-IN-B): Reading in the DRESSED indel rates >>\n\n";
} else {
	print STDOUT "\n<< (Pre 00-IN-B): Reading in the RAW indel rates >>\n\n";
}
##print STDOUT "\n<< (Pre 00-IN-B): Reading in the raw indel rates >>\n\n";
		# End of "Modified in creation of 'cmpt_mu_Dcr.pl' (on 2020/02/04). (00-B-1)" #

our ($rf_RATES_INS, $rf_RATES_DEL, $rf_RATES_INS_BD, $rf_RATES_DEL_BD)
    = readin_indel_rates ($infile_id_rates);

unless (defined $rf_RATES_DEL_BD) {
	print STDOUT "[ERROR '$program_name'] The subroutine, 'readin_indel_rates', FAILED on the input file, \"$infile_id_rates\" $? !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}
(print STDOUT "\n ... The subroutine, 'readin_indel_rates', Ended on the input file, \"$infile_id_rates\" ...\n\n") if ($TEST_LEVEL > 1);


if ($TEST_LEVEL > 2) {

		# Modified in creation of 'cmpt_mu_Dcr.pl' (on 2020/02/04). (00-B-2) #
  if ($IF_DRESSED) {
    print STDOUT "\n[ DRESSED Indel Rates (x \$SIZE_SB_INTVL) retrieved: ]\n\n";
  } else {
    print STDOUT "\n[ RAW Indel Rates retrieved: ]\n\n";
  }
    ##print STDOUT "\n[ Raw Indel Rates retrieved: ]\n\n";
		# End of "Modified in creation of 'cmpt_mu_Dcr.pl' (on 2020/02/04). (00-B-2)" #

    print STDOUT join ("\t", 'Length', 'Ins', 'Del', 'Ins(bd)', 'Del(bd)'), "\n\n";

    my $maxlen = 0;
    foreach my $rf_rates ($rf_RATES_INS, $rf_RATES_DEL, $rf_RATES_INS_BD, $rf_RATES_DEL_BD) {
	my $len = @{$rf_rates} - 1;
	if ($maxlen < $len) { $maxlen = $len; }
    }

    for (my $l=1; $l <= $maxlen; $l++) {
	my @out_rates = ();
	foreach my $rf_rates ($rf_RATES_INS, $rf_RATES_DEL, $rf_RATES_INS_BD, $rf_RATES_DEL_BD) {
	    my $rate = $rf_rates->[$l];
	    unless (defined $rate) { $rate = 'undef'; }
	    push @out_rates, $rate;
	}
	print STDOUT join ("\t", $l, @out_rates), "\n";
    }
    print STDOUT "\n\n";

} # End of "if ($TEST_LEVEL > 2) {...}"



		# IMPORTED from 'cmpt_mu_Dcr2Adel.pl', in creation of 'cmpt_mu_Dcr2Berdot.pl' (on 2020/02/24). (00-IN-G) #

print STDOUT "\n<< (Pre 00-IN-G): Reading in the multiplication factors for the creation of a \"D\"-region >>\n\n";

my ($rf_mu_Dcr) = readin_muP_case_ii_or_iii ($infile_mu_Dcr, $CT_SB_INTVLS, $UB_CT_SITES_CMPT);

	# $rf_mu_Dcr->[$LF][$jt] (w/ 0< $LF <= $ub_ct_sites  & 0 <= $jt <= $ct_sb_intvls)
	# is the multiplication factor for the creation of a "D"-region
	# at the $jt th time-interval with $LF final sites in between the PASs.

unless (defined $rf_mu_Dcr) {
	print STDOUT "[ERROR '$program_name'] The subroutine, 'readin_muP_case_ii_or_iii', FAILED on the input file, \"$infile_mu_Dcr\" $? !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}
(print STDOUT "\n ... The subroutine, 'readin_muP_case_ii_or_iii', Ended on the input file, \"$infile_mu_Dcr\" ...\n\n") if ($TEST_LEVEL > 1);

		# End of "IMPORTED from 'cmpt_mu_Dcr2Adel.pl', in creation of 'cmpt_mu_Dcr2Berdot.pl' (on 2020/02/24). (00-IN-G)" #


			# (Added in creation of 'cmpt_mu_Dcr2Berdot.pl' on 2020/02/24) (00-IN-G) #

if ($TEST_LEVEL >= 5) { # (00-IN-G) #

	print STDOUT "\n\n[[ Multiplication factors, mu_{Dcr} [ (x, [t_I, t_I + dt_I]) ; (LDF, t_F) ], retrieved (& stored in ", '@{$rf_mu_Dcr}', "): ]]\n\n";

	print STDOUT join ("\t", '$LDF\$jt', 0 .. $CT_SB_INTVLS), "\n\n";

	my $ct_tintvls = $CT_SB_INTVLS +1;
	my $cap_ct_sites = $UB_CT_SITES_CMPT +1;

	for (my $LDF = 1; $LDF < $cap_ct_sites; $LDF++) { # Main outer for-loop (over the final D-site-counts).

	    my $vec_mu_Dcr = $rf_mu_Dcr->[$LDF];


	    print STDOUT join ("\t", $LDF, @{$vec_mu_Dcr}), "\n";


	} # Main outer for-loop (over the final site-count).


	print STDOUT "\n\n\n";

} # End of "if ($TEST_LEVEL >= 5) { # (00-IN-G) # ...}"

			# End of "(Added in creation of 'cmpt_mu_Dcr2Berdot.pl' on 2020/02/24) (00-IN-G)" #

=For the predecessor, 'cmpt_mu_Adel.pl'. (00-IN-C)

	## 00-IN-C: Read in the input file for muP0s (w/ $SIGMA_BE = 2). ##
	## (Added in creation of 'cmpt_mu_Dcr.pl' on 2020/02/04&05). (00-IN-C) ##

print STDOUT "\n<< (Pre 00-IN-C): Reading in the muP0s (w/ \$SIGMA_BE = 2) >>\n\n";

my ($rf_muP0s) =  readin_muP0s ($infile_muP0, $CT_SB_INTVLS, $UB_CT_SITES_CMPT);

	# $rf_muP0s->[$LI][$jt][$l] = (w/ 0 < $L_I, l <= $ub_ct_sites, & 0 <= $jt <= $ct_sb_intvls) 
	# is the multiplication factor computed under the "base" rate operator
	# at the $jt th time-interval with $L_I sites initially and $l sites finally, both in between the PASs.


unless (defined $rf_muP0s) {
	print STDOUT "[ERROR '$program_name'] The subroutine, 'readin_muP0s', FAILED on the input file, \"$infile_muP0\" $? !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}
(print STDOUT "\n ... The subroutine, 'readin_muP0s', Ended on the input file, \"$infile_muP0\" ...\n\n") if ($TEST_LEVEL > 1);


if ($TEST_LEVEL >= 5) {

	print STDOUT "\n\n[[ \"Base\" multiplication factors, mu_{P0} [ (L, t) | (L_I, t_I) ], retrieved (& stored in ", '@{$rf_muP0s}', "): ]]\n\n";

	print STDOUT join ("\t", '$L_I:$l\$jt', 0 .. $CT_SB_INTVLS), "\n\n";

			## IMPORTED w/ modifications from the subroutine, 'print_muP0s', in "MyLASTPIECE_IO.pm" (on 2020/02/05). ## 

	my $ct_tintvls = $CT_SB_INTVLS +1;
	my $cap_ct_sites = $UB_CT_SITES_CMPT +1;

	for (my $L_I = 1; $L_I < $cap_ct_sites; $L_I++) { # Main outer for-loop (over the initial site-count).

	    my $muP0s_fxd_L_I = $rf_muP0s->[$L_I];

	    for (my $l =1; $l < $cap_ct_sites; $l++) { # Main middle for-loop (over the final site-counts). # (MODIFIED on 2020/02/02). #

		my @out_muP0s = ();

		for (my $jt = 0; $jt < $ct_tintvls; $jt++) { # Inner for-loop (over the time-intervals). #
		    $out_muP0s[$jt] = $muP0s_fxd_L_I->[$jt][$l];
		} # End of the inner for-loop (over the time-intervals). #

		print STDOUT join ("\t", "$L_I:$l", @out_muP0s), "\n";

	    } # End of the main middle for-loop (over the final site-counts). #

	} # Main outer for-loop (over the initial site-count).

			## End of "IMPORTED w/ modifications from the subroutine, 'print_muP0s', in "MyLASTPIECE_IO.pm" (on 2020/02/05)." ## 

	print STDOUT "\n\n\n";

} # End of "if ($TEST_LEVEL >= 5) {...}"

	## End of "(Added in creation of 'cmpt_mu_Dcr.pl' on 2020/02/04&05). (00-IN-C)" ##
=cut # END of "For the predecessor, 'cmpt_mu_Adel.pl'. (00-IN-C)"


	## End of "(MODIFIED on 2020/01/31) (00-IN)" ##


	## Preliminary 00-C: Compute Other Important Parameters ##
	### (IMPORTED w/ modifications from '../../../lastpiece.alpha.pl' on 2020/01/30) (00-C) ###

print STDOUT "\n<<< Preliminary Process 00-C: Computing Other Important Parameters >>>\n\n";

our $TOTRATE_INDEL = $TOTRATE_INS + $TOTRATE_DEL;
our $SIZE_FULL_INTVL = $TIME_FIN - $TIME_INIT;
our $SIZE_SB_INTVL = $SIZE_FULL_INTVL /$CT_SB_INTVLS;

our @TIME_LAPSES_CMPT = (); # Values of the time lapse (= t_F - t_I) at which the factors will be computed.
for (my $i = 0; $i <= $CT_SB_INTVLS; $i++) {
	$TIME_LAPSES_CMPT[$i] = $i * $SIZE_SB_INTVL;
} 

print STDOUT '$TOTRATE_INDEL = ', $TOTRATE_INDEL, " \;\n";
print STDOUT '$SIZE_FULL_INTVL = ', $SIZE_FULL_INTVL, " \;\n";
print STDOUT '$SIZE_SB_INTVL = ', $SIZE_SB_INTVL, " .\n\n";

print STDOUT '@TIME_LAPSES_CMPT = (', join (', ', @TIME_LAPSES_CMPT), ") .\n\n";


print STDOUT "\n";

	### End of "(IMPORTED w/ modifications from '../../../lastpiece.alpha.pl' on 2020/01/30) (00-C)" ###

	##### END of "(IMPORTED from '../muP_case_i2iii/cmpt_muP_case_i2iii.pl' on 2020/02/03) (Pre 00) #####
	##### (==> IMPORTED w/ modifications from '../muP0/cmpt_muP0.pl' on 2020/02/04&05) (Pre 00) #####
	##### (==> IMPORTED from '../mu_Dcr/cmpt_mu_Dcr.pl' on 2020/02/06) (Pre 00)" #####


	##### (IMPORTED w/ modifications from '../mu_Dcr/cmpt_mu_Dcr.pl' on 2020/02/06) (Pre 1) #####

print STDOUT "\n<<< Preliminary Process 1: Computing ", '@delrates_x_dt', " >>>\n\n";

my @delrates_x_dt  = (0); # $delrates_x_dt[$l] = {the rate of deletion of length $l} x dt.


if ($IF_DRESSED) { # The input rates are "dressed".

   for (my $l=1; $l <= $CUTOFF_INS_LEN; $l++) {
	$delrates_x_dt[$l] = $rf_RATES_DEL->[$l];
   }

} else { # The input rates are "raw".

   for (my $l=1; $l <= $CUTOFF_INS_LEN; $l++) {
	$delrates_x_dt[$l] = $SIZE_SB_INTVL * $rf_RATES_DEL->[$l];
   }

} # End of "if ($IF_DRESSED) {...} else {...}". #

=For the predecessor, '../mu_Dcr/cmpt_mu_Dcr.pl'.
print STDOUT "\n<<< Preliminary Process 1: Computing ", '@insrates_x_dt', " >>>\n\n";

my @insrates_x_dt  = (0); # $insrates_x_dt[$l] = {the rate of insertion of length $l} x dt.


if ($IF_DRESSED) { # The input rates are "dressed".

   for (my $l=1; $l <= $CUTOFF_INS_LEN; $l++) {
	$insrates_x_dt[$l] = $rf_RATES_INS->[$l];
   }

} else { # The input rates are "raw".

   for (my $l=1; $l <= $CUTOFF_INS_LEN; $l++) {
	$insrates_x_dt[$l] = $SIZE_SB_INTVL * $rf_RATES_INS->[$l];
   }
} # End of "if ($IF_DRESSED) {...} else {...}". #
=cut # End of "For the predecessor, '../mu_Dcr/cmpt_mu_Dcr.pl'."

	##### End of "(IMPORTED w/ modifications from '../mu_Dcr/cmpt_mu_Dcr.pl' on 2020/02/06) (Pre 1)" #####


		### IMPORTED w/ modifications from 'cmpt_mu_Dcr2Adel.pl' (on 2020/02/24). (Pre 2) ###

print STDOUT "\n<<< Preliminary Process 2: Computing Minimum Damping Factors >>>\n\n";

my $cap_ct_sites = $UB_CT_SITES_CMPT + 1; # MOVED from near the top of the Main Process.

my @min_damping_facs = (); # The $l th element = {Minimum damping factor for the sub-sequence of length $l}.
for (my $l=0; $l < $cap_ct_sites; $l++) {
	$min_damping_facs[$l] = ($IF_DRESSED) ? exp (- $SIZE_SB_INTVL * $TOTRATE_INDEL * $l) : 1;
}

if ($TEST_LEVEL > 3) { # (Pre 2) #

	print STDOUT "\n[ Minimum Damping Factors: ", 
	'$min_damping_facs[$L] = ($IF_DRESSED) ? exp (- $SIZE_SB_INTVL * $TOTRATE_INDEL * $L) : 1', " ]\n\n";

	print STDOUT join ("\t", '$L', '$min_damping_facs[$L]'), "\n\n";
	for (my $L=0; $L < $cap_ct_sites; $L++) {
	    print STDOUT join ("\t", $L, $min_damping_facs[$L]), "\n";
	}
	print STDOUT "\n\n";

} # End of "if ($TEST_LEVEL > 3) { # (Pre 2) # ...}"

		### END of "IMPORTED w/ modifications from 'cmpt_mu_Dcr2Adel.pl' (on 2020/02/24). (Pre 2)" ###



	###  Main Process: Compute & Output multiplication factors, 'mu_{Dcr->Ber.} [(x=1, [t_I, t_I + dt_I]) ;  (LDF, -delta{}LAF, [t_F - dt_F, t_F]) ]'.

print STDOUT "\n<<<<< Main Process: Compute & Output multiplication factors,  'mu_{Dcr->Ber.} [(x=1, [t_I, t_I + dt_I]) ;  (LDF, -delta{}LAF, [t_F - dt_F, t_F]) ]' >>>>>\n\n";

		# (Preliminary) Open a file-handle for outputting to the output file. #
my $fh_out = open_output_file ($ofile_mu_Dcr2Berdot , 0);
unless (defined $fh_out) {
	print STDOUT "[ERROR '$program_name'] The subroutine 'open_output_file' FAILED to open a file-handle for the file \"$ofile_mu_Dcr2Berdot\": $? !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}

## my ($infile_id_prms, $infile_id_rates, $infile_mu_Dcr, $ofile_mu_Dcr2Berdot) = @ARGV[3+1 .. 7];

print $fh_out join ("\t", '#{D-sites}_Fin:#{A-sites_deleted_via_B-er}\Time_Lapse', @TIME_LAPSES_CMPT), "\n\n";

my $ct_tlapses = @TIME_LAPSES_CMPT;
##my $cap_ct_sites = $UB_CT_SITES_CMPT +1; # MOVED to the top of "Pre 2".

my @vec_null_probs = ();
for (my $jt =0; $jt < $ct_tlapses; $jt++) { $vec_null_probs[$jt] = 0; }


		# Then, perform the main computation. #

my $cap_delta_LA = ($CUTOFF_DEL_LEN -1 < $UB_CT_SITES_CMPT -1) ?  $CUTOFF_DEL_LEN -1 : $UB_CT_SITES_CMPT -1  ; 
$cap_delta_LA++;

for (my $LDF = 1; $LDF < $cap_ct_sites; $LDF++) { # Main Outermost for-loop (over final D-site-counts (i.e., $LDF)). #

	##print $fh_out join ("$LDF:0", @vec_null_probs), "\n"; # By definition of 'boundary-eroding' deletions, mu_{Dcr->Ber.} w/ $dLA = 0 should always be ZERO. # A BUG fixed on 2020/02/25. #
	print $fh_out join ("\t", "$LDF:0", @vec_null_probs), "\n"; # By definition of 'boundary-eroding' deletions, mu_{Dcr->Ber.} w/ $dLA = 0 should always be ZERO.

    for (my $dLA = 1; $dLA < $cap_delta_LA; $dLA++) { # Main Outer for-loop (over #{A-sites deleted via 'boundary-erosion'} (i.e., $dLA)) #

	my $cap_delta_LD = ($CUTOFF_DEL_LEN - $dLA < $UB_CT_SITES_CMPT - $LDF) ?
				$CUTOFF_DEL_LEN - $dLA : $UB_CT_SITES_CMPT - $LDF ; 
	$cap_delta_LD++;

		# Initialize the output multiplication factors (at various time-lapses). #
	my @vec_mu_out = @vec_null_probs;

	for (my $dLD = 1; $dLD < $cap_delta_LD; $dLD++) { # Main middle for-loop (over #{D-sites deleted via 'boundary-erosion'} (i.e., $dLD)) #

	    my $delta_L = $dLA + $dLD;
	    my $LD_bf = $LDF + $dLD;

	    my $vec_mu_Dcr = $rf_mu_Dcr->[$LD_bf];

		# $rf_mu_Dcr->[$LF][$jt] (w/ 0< $LF <= $ub_ct_sites  & 0 <= $jt <= $ct_sb_intvls)
		# is the multiplication factor for the creation of a "D"-region
		# at the $jt th time-interval with $LF final sites in between the PASs.

	    my $mdfac = $min_damping_facs[$LD_bf];

	    my $delrate_x_dt = $delrates_x_dt[$delta_L];

	    for (my $jt=1; $jt < $ct_tlapses; $jt++) { # Main inner for-loop (over time-lapses). #

		$vec_mu_out[$jt] += $vec_mu_Dcr->[$jt-1] * $mdfac * $delrate_x_dt ;

	    } # End of the main middle for-loop (over time-lapses). #

	} # End of the main middle for-loop (over #{D-sites deleted via 'boundary-erosion'} (i.e., $dLD)) #

	print $fh_out join ("\t", "$LDF:$dLA", @vec_mu_out), "\n";

			### RESTART FROM HERE ###

    } # End of the Main Outer for-loop (over #{A-sites deleted via 'boundary-erosion'} (i.e., $dLA)) #

} # End of the Main Outermost for-loop (over final D-site-counts (i.e., $LDF)). #


close ($fh_out);


print STDOUT "\n\n === All The Processes of '$program_name' FINISHED !! ===\n\n\n";


=For the predecessor, 'cmpt_mu_Adel.pl'. (MAIN)

	###  Main Process: Compute & Output multiplication factors, 'mu_{A-del} [ (L_I, t_1) ; (0, -delta{}L_{fD}, [t_2 - Delta{}t, t_2]) ] (sigmaBE = $SIGMA_BE)',  for the deletion of an "A"-region. ###

print STDOUT "\n<<<<< Main Process: Compute & Output multiplication factors,  'mu_{A-del} [ (L_I, t_1) ; (0, -delta{}L_{fD}, [t_2 - Delta{}t, t_2]) ] (sigmaBE = $SIGMA_BE)',  for the deletion of an \"A\"-region >>>>>\n\n\n";

		# (Preliminary) Open a file-handle for outputting to the output file. #
my $fh_out = open_output_file ($ofile_mu_Adel, 0);
unless (defined $fh_out) {
	print STDOUT "[ERROR '$program_name'] The subroutine 'open_output_file' FAILED to open a file-handle for the file \"$ofile_mu_Adel\": $? !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}

print $fh_out join ("\t", '#{sites}_Init:#{deleted_sites_in_fD}\Time_Lapse', @TIME_LAPSES_CMPT), "\n\n";

my $ct_tlapses = @TIME_LAPSES_CMPT;
##my $cap_ct_sites = $UB_CT_SITES_CMPT +1; # MOVED to the top of "Pre 2".



# our ($rf_RATES_INS, $rf_RATES_DEL, $rf_RATES_INS_BD, $rf_RATES_DEL_BD)
# print STDOUT '($CUTOFF_INS_LEN, $CUTOFF_DEL_LEN) = ', "($CUTOFF_INS_LEN, $CUTOFF_DEL_LEN) .\n\n";

		# Then, perform the main computation. #

my $cap_delta_L_fD = ($SIGMA_BE == 2) ? 0 : 
		(($CUTOFF_DEL_LEN -1 < 2 * ($UB_CT_SITES_CMPT -1)) ?  $CUTOFF_DEL_LEN -1 : 2 * ($UB_CT_SITES_CMPT -1) ) ; 
$cap_delta_L_fD++;

for (my $LI = 1; $LI < $cap_ct_sites; $LI++) { # Main Outermost for-loop (over initial site-counts). #

##=For the predecessor, '../mu_Dcr/cmpt_mu_Dcr.pl'. (Main 3a) 
##	my $insrate_x_dt = $insrates_x_dt[$LI];
##=cut # End of "For the predecessor, '../mu_Dcr/cmpt_mu_Dcr.pl'. (Main 3a)"

	my $muP0s_fxd_LI = $rf_muP0s->[$LI];

	# $rf_muP0s->[$LI][$jt][$L] = (w/ 0 < $LI, L <= $ub_ct_sites, & 0 <= $jt <= $ct_sb_intvls) 
	# is the multiplication factor computed under the "base" rate operator
	# at the $jt th time-interval with $L_I sites initially and $L sites finally, both in between the PASs.

    for (my $dL_fD = 0; $dL_fD < $cap_delta_L_fD; $dL_fD++) { # Main Outer for-loop (over #{deleted sites in flanking "D"-region(s)) #

	my $cap_l_F = ($CUTOFF_DEL_LEN - $dL_fD < $UB_CT_SITES_CMPT) ? 
			    $CUTOFF_DEL_LEN - $dL_fD : $UB_CT_SITES_CMPT ;
	$cap_l_F++;

	my @vec_mu_Adel = (0); # Always set the factor for the zero-time-lapse to be zero. #

	for (my $jt=1; $jt < $ct_tlapses; $jt++) { # Main middle for-loop (over time-lapses). #

	    my $vec_muP0s = $muP0s_fxd_LI->[$jt-1]; # The time-lapse of mu_{P0} MUST BE smaller than that of mu_Adel by 1 sub-interval (i.e., Delta{}t). #

	    my $mu_Adel = 0;

	    for (my $L = 1; $L < $cap_l_F; $L++) { # Main inner for-loop (over final site-counts).

		##$mu_Adel += $delrates_x_dt[$L + $dL_fD] * $vec_muP0s->[$L];
		$mu_Adel += $delrates_x_dt[$L + $dL_fD] * $min_damping_facs[$L] * $vec_muP0s->[$L]; ## CORRECTED on 2020/02/24. ##

	    } # Main inner for-loop (over final site-counts).

	    $vec_mu_Adel[$jt] = $mu_Adel;

##=For the predecessor, '../mu_Dcr/cmpt_mu_Dcr.pl'. (Main 3b) 
##	    my $vec_mu_Dcr = $mu_Dcr[$jt];
##
##	    for (my $L=1; $L < $cap_ct_sites; $L++) {
##		$vec_mu_Dcr->[$L] += $insrate_x_dt * $vec_muP0s->[$L];
##	    }
##=cut # End of "For the predecessor, '../mu_Dcr/cmpt_mu_Dcr.pl'. (Main 3b)" 

	} # End of the main middle for-loop (over time-lapses). #

	print $fh_out join ("\t", "$LI:$dL_fD", @vec_mu_Adel), "\n";

    } # End of the Main Outer for-loop (over #{deleted sites in flanking "D"-region(s)) #

} # End of the Main Outermost for-loop (over initial site-counts). #


close ($fh_out);


print STDOUT "\n\n === All The Processes of '$program_name' FINISHED !! ===\n\n\n";

=cut # END of "For the predecessor, 'cmpt_mu_Adel.pl'. (MAIN)"


