package MyLinearAlgebra;
#
###################################################################################
#	DENSERM_P: A package to DEtect Negative SElection on Recurrent Mutations
#			(almost exclusively via Perl modules and scripts).
#	Version 0.3: Copyright (C) 2012 Kiyoshi Ezawa
#
#    This package is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This package is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License, "GNU_GPL.txt",
#    along with this package.  If not, see <http://www.gnu.org/licenses/>.
#
#    The author can be contacted by e-mailing to <kezawa dot ezawa3 at gmail dot com>
#    (replace " dot " and " at " with "." and "@", respectively).
###################################################################################
#
# MyLinearAlgebra.pm: This module contains subroutines
#		which solve various linear algebraic problems.
#
#	SOURCES:
#	Some subroutines in this module are Perl-translated versions of the programs in:
#	"Handbook of Numerical Methods executable on UNIX Engineering Workstation, basic C version." (in Japanese ), written by Tagawa H (Saiensu-sha, Co., Ltd.; Tokyo, Japan; 1998). 
#

use strict;
use warnings;

use Exporter;

our @ISA  = qw(Exporter);

our @EXPORT_OK = qw(
	multiply_matrices lu_decomposition lu_fw_bk_sbst
	solve_lin_eq_lu invert_matrix_lu );

our $MIN_PIVOT = 1.0e-13;	# if the pivot element is less than this value,
				# then give up the calculation.
our $TINY = 1.0e-20;

# multiply_matrices: This subroutine receives 2 matrices (in the form of arrays of arrays)
#		@a and @b and returns their product @a * @b.
#
sub multiply_matrices (\@\@) {

	my ($a, $b) = @_;

	my $ct_rows = @{$a};
	($ct_rows and @{$b}) or (return []);
	my $ct_terms = @{$a->[0]};
	unless (@{$b} == $ct_terms) {
		print STDERR "Illegal input for \'mutiply_matrices\' .\n";
		return undef;
	}

	my $ct_clms = @{$b->[0]};
	my @prod = ();
	for (my $i=0; $i < $ct_rows; $i++) {
		my @col = ();
		for (my $j=0; $j < $ct_clms; $j++) {
			my $sum = 0;
			for (my $k=0; $k <$ct_terms; $k++) {
				$sum += $a->[$i][$k] * $b->[$k][$j];
			}
			push @col, $sum;
		}
		push @prod, \@col;
	}

	return \@prod;

}		# end of the subroutine multiply_matrices #


# lu_decomposition: This subroutine receives a square matrix,
#		  replaces it with the LU-decomposition of its rowwise permutation,
#		  and returns a record of the row permutations effected by the partial pivoting,
#		along with a sign +1/-1 which tells whether the number of row interchanges
#		was even or odd.
#
sub lu_decomposition (\@) {

	my ($a) = @_;

	my $n = @{$a}; # matrix size #
	my @indx = (); # a record of the row permutation effected by the partial pivoting
	my $d = 1; # =  +1 if the number of row interchange was even, = -1 if it was odd.

	my @vv = (); # stores scaling factors for rows 
	for (my $i=0; $i< $n; $i++) {
	    my $big = 0;
	    for (my $j=0; $j< $n; $j++) {
		my $abs = ($a->[$i][$j] > 0) ? $a->[$i][$j] : - $a->[$i][$j] ;
		($abs > $big) and ($big = $abs);
	    }

	    if ($big == 0) {
		print STDERR "Singular matrix in \'lu_decomposition\'.\n";
		return (undef, undef);
	    }
	    push @vv, 1/$big;
	}

	for (my $j=0; $j < $n; $j++) {  # The loop over columns of Crout's method.

	   for (my $i=0; $i < $j; $i++) {  #  Equation (2.3.12) except for $i = $j.
		my $sum = $a->[$i][$j];
		for (my $k=0; $k<$i; $k++) { $sum -= $a->[$i][$k] * $a->[$k][$j]; }
		$a->[$i][$j] = $sum;
	   }

	   # Initialize for the search for the largest pivot element.
	   my $big = 0;
	   my $imax = -1;
	   for (my $i = $j; $i < $n; $i++) { # $i = $j of equation (2.3.12) and
					    # $i = $j+1, ..., $n of equation (2.3.13).
		my $sum = $a->[$i][$j];
		for (my $k=0; $k < $j; $k++) { $sum -= $a->[$i][$k] * $a->[$k][$j]; }
		$a->[$i][$j] = $sum;
		my $dum = ($sum > 0) ? $vv[$i] * $sum : -$vv[$i] * $sum ;
		if ($dum > $big) { # Is the figure of merit for the pivot better than the best so far ?
		   $big = $dum;
		   $imax  = $i;
		}
	   }

		# This block has no counterpart in the textbook.
	   if ($big < $MIN_PIVOT) {
		print STDERR "Almost singular matrix for \'lu_decomposition\'.\n";
		return (undef, undef);
	   }

	   if ($j != $imax) {	# Do we need to interchange rows?
		for (my $k=0; $k <$n; $k++) { # Yes, do so...
			my $dum = $a->[$imax][$k];
			$a->[$imax][$k] = $a->[$j][$k];
			$a->[$j][$k] = $dum;
		}
		$d = - $d;	# ... and  change the parity of $d.
		$vv[$imax] = $vv[$j]; # Also interchange the scale factor.
	   }
	   push @indx, $imax;

	   ($a->[$j][$j] == 0) and ($a->[$j][$j] = $TINY);
	   # if the pivot element is zero the matrix is singular
	   # (at least to the precision of the algorithm).
	   # For some applications on singular matrices,
	   # it is desirable to substitute $TINY for zero.

	   if ($j != $n -1) {	# Now, finally, divide by the pivot element.
		my $dum = 1/$a->[$j][$j];
		for (my $i = $j+1; $i< $n; $i++) { $a->[$i][$j] *= $dum; }
	   }
	}		# Go back for the next column in the reduction.

	return (\@indx, $d);

}		# end of the subroutine lu_decomposition # 


# lu_fw_bk_sbst: This subroutine receives the LU-decomposition (@{$lu}) of a square matrix,
#		along with the record (@{$indx}) of row permutation,
#		and a numerical vector (@{$num} = the r.h.s of the linear equation AX = B).
#		Then it conducts forward substitution followed by backsubstitution
#		to solve the linear equation, and returns the solution vector X.
#		@{$lu} and @{$indx} are NOT modified by this subroutine and can be
#		left in place for successive calls with different right-hand sides @{$num}.
#		The subroutine takes into account the possibility that @{$num} will begin with
#		many zero elements, so it is efficient for use in matrix inversion.
#
sub lu_fw_bk_sbst (\@\@\@) {

	my ($lu, $indx, $num) = @_;

	my $n = @{$lu};
	unless (@{$num} == $n) {
		print STDERR "Illegal input for \'lu_fw_bk_sbst\'.\n";
		return undef;
	}

	my @x = @{$num};
	# When $ii is set to a positive value,
	# it will become the index of the 1st nonvanishing element of @{$num}.
	my $ii= -1;

	# We now do the forward substitution, equation (2.3.6). The only new wrinkle is
	# to unscramble the permutation as we go.
	for (my $i=0; $i <$n; $i++) {
	   my $ip = $indx->[$i];
	   my $sum = $x[$ip];
	   $x[$ip] = $x[$i];
	   if ($ii >= 0) {
		for (my $j =$ii; $j < $i; $j++) { $sum -= $lu->[$i][$j] * $x[$j]; }
	   } elsif ($sum) { # A nonzer element was encountered, so from now on we will
		$ii = $i;   # have to do the sums in the loop above. 
	   }
	   $x[$i] = $sum;
	}

	# Now we do the backsubstitution, equation (2.3.7). 
	for (my $i=$n-1; $i >= 0; $i--) {
	   my $sum = $x[$i];
	   for (my $j= $i+1; $j < $n;  $j++) { $sum -= $lu->[$i][$j] * $x[$j]; }
	   $x[$i] = $sum/$lu->[$i][$i];	# Store a component of the solution vector X.
	}	# All done !

	return \@x;

}		# end of the subroutine lu_fw_bk_sbst #

# solve_lin_eq_lu: This subroutine solves a set of n linear equations AX = B
#		by means of the LU-decomposition.
#		Along the way, the subroutine replaces the input coefficient matrix A (=@{$a})
#		into its LU-decomposition.
#
sub solve_lin_eq_lu (\@\@) {

	my ($a, $b) = @_;

	unless (@{$a} and (@{$a->[0]} == @{$a})  and (@{$a} == @{$b}) ) {
		print STDERR "Illegal input for \'solve_lin_eq_lu\'.\n";
		return undef;
	}

	my ($indx, $sgn) = lu_decomposition (@{$a});
	unless (defined $indx) {
		print STDERR "\'lu_decomposition\' failed.\n";
		return undef;
	}

	my $sol = lu_fw_bk_sbst (@{$a}, @{$indx}, @{$b});
	unless (defined $sol) {
		print STDERR "\'lu_fw_bk_sbst\' failed.\n";
		return undef;
	}

	return $sol;

}	# end of the subroutine solve_lin_eq_lu #

# invert_matrix_lu: 
#	This subroutine receives a regular matrix,
#	and returns its inverted matrix produced via the LU decomposition.
#	(CREATED on April 28 (Tue), 2009)
#
# INPUT: \@matrix, where
#	$matrix[$i]->[$j] = a numerical value with 0 <= $i, $j < scalar (@matrix);
#
# OUTPUT: \@inversion, where
#	@matrix * @inversion = diag (1, ..., 1).
#
sub invert_matrix_lu (\@) { 

	my ($matrix) = @_;
	my $size_mtrx = @{$matrix};
	unless (($size_mtrx>0) and (@{$matrix->[0]} == $size_mtrx)) {
	    print STDERR "[ERROR 'invert_matrix_lu'] The INPUT is NOT a SQUARE MATRIX of a POSITIVE SIZE: $! !\n\n";
	    return undef;
	}

	my ($indx, $sgn) = lu_decomposition (@{$matrix});
	unless (defined $indx) {
	    print STDERR "[ERROR 'invert_matrix_lu'] 'lu_decomposition' FAILED: $! !\n";
	    return undef;
	}

	my @inversion = my @zero_vec = ();
	for (1..$size_mtrx) { push @zero_vec, 0; push @inversion, []}

	for (my $i=0; $i<$size_mtrx; $i++) { # The main for-loop #

	    my @unit_i = @zero_vec;
	    $unit_i[$i] = 1;

	    my $inv_i = lu_fw_bk_sbst (@{$matrix}, @{$indx}, @unit_i);
	    unless (defined $inv_i) {
		print STDERR "[ERROR 'invert_matrix_lu'] 'lu_fw_bk_sbst' FAILED for the $i th colum: $! !\n";
		return undef;
	    }

		## A BUG FIXED on April 28 (Tue), 2009 ##
	    for (my $j=0; $j<$size_mtrx; $j++) { push @{$inversion[$j]}, $inv_i->[$j]; }
	    ##push @inversion, $inv_i;
	    
	}		# End of the main for-loop #

	return \@inversion;

}	# end of the subroutine 'invert_matrix_lu' #


1;		# a 'spell which makes the module "use"-able and/or "require"-able #


