package MyMalnManipu;
#
###################################################################################
#	DENSERM_P: A package to DEtect Negative SElection on Recurrent Mutations
#			(almost exclusively via Perl modules and scripts).
#	Version 0.3: Copyright (C) 2012 Kiyoshi Ezawa
#
#    This package is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This package is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License, "GNU_GPL.txt",
#    along with this package.  If not, see <http://www.gnu.org/licenses/>.
#
#    The author can be contacted by e-mailing to <kezawa dot ezawa3 at gmail dot com>
#    (replace " dot " and " at " with "." and "@", respectively).
###################################################################################
#
# MyMalnManipu.pm: This module manipulates multiple or pairwise alignments
#		in a more sophisticated manner than the module "MyMalnIO.pm".
#
# NOTE: This package is almost exclusively for nucleotide multiple-alignments.
#
# REVISED on Nov 19 (Fri), 2004:
#	The subroutine 'compare_columns', which is  a satellite of 'compare_malign',
#	no longer depends on the recursion technique.
#

use strict;
use warnings;
use Exporter;

our @ISA = qw( Exporter);

our @EXPORT_OK = qw(
	count_inv_sites
	pad_5p_end pad_3p_end merge_pad_align rm_gapcolumn rm_gapcolumn2 rm_gapcolumn_codon
	assign_number assign_number_gen assign_number_cnctseqs
	map_resnum2alncoord map_resnum2alncoord_2
	separate_nogap_sites separate_resnum compare_malign diff_malign
	construct_blocks block_starts_ends monosite_picker id_indicator id_indicator_aa
	print_block_maln equiv_columns
	trim_diff_maln maln_identities block_maln_identities block_maln_id_distribution
	id_sliding_window bl_id_sliding_window average_id_proportions pick_poor_windows pick_bl_poor_windows
	mask_frames compl_mask_frames extract_bl_selected_frames select_in_phase
	select_unamb_frames sel_complete_codon_frames sel_codon_frames_in_ref
	extract_codon_alignment extract_codon_aln_with_ref
	prepare_oden_input
	extract_intact_codonaln extract_intact_cnct_codonaln
	remove_stop_codons
	pr_in_oden_dec chop_alignment
	);

use MyArrayManipu_num qw( diff_nmarray );

sub fabricate_pads (\@$);

#sub extract_codon_alignment (\@\@\@) {
#sub sel_complete_codon_frames (\@\@) {
#sub select_unamb_frames (\@\@) {
#sub select_in_phase (\@\@) {

#sub pick_bl_poor_windows (\@\@\@\@$$) {
#sub pick_poor_windows (\@\@\@\@$$) {
sub compare_columns (\@\@$);
sub compare_malign_fw (\@\@\@\@);
sub compare_malign_bw (\@\@\@\@);
sub id_indicator (\@);
sub phase_column (\@);


# count_inv_sites:
#	This subroutine counts the invariant sites in a multiple alignment,
#	as well as the 'effective sites' which do not contain characters to be dismissed.
#
# INPUT: (\@malign, %dismissed), where
#	${$malign[$m]} is the $m th sequence in the alignment;
#	%dismissed = ($char => (some defined value), ...) for the characters to be dismissed.
#
# OUTPUT: ($ct_effsites, $ct_inv_sites).
#
sub count_inv_sites (\@\%) {

	my ($malign, $dismissed) = @_; 

	my $ct_seqs = @{$malign};
	($ct_seqs>0) or (return (0, 0));

	my @cp_maln = ();
	my $min_len;
	foreach my $seq (@{$malign}) { # The 1st outer foreach-loop #
	    my $cp_seq = uc ${$seq};
	    push @cp_maln, \$cp_seq;
	    my $len_seq = length $cp_seq;
	    ((defined $min_len) and ($min_len < $len_seq)) or ($min_len = $len_seq);
	}			# End of the 1st outer foreach-loop #

	my $ct_inv = my $ct_eff = 0;

	for (my $l=0; $l<$min_len; $l++) { # The 2nd outer foreach-loop #
	   my $char_1st = substr (${$cp_maln[0]}, 0, 1);
	   my $ct_eq = my $ct_dismiss = 0;
	   for (my $k=0; $k<$ct_seqs; $k++) { # The inner for-loop #
		my $char = substr (${$cp_maln[$k]}, 0, 1, '');
		(defined $dismissed->{$char}) and ($ct_dismiss++);
		($char_1st eq $char) and ($ct_eq++);
	   }			 # End of the inner for-loop #
	   ($ct_dismiss>0) and next;
	   $ct_eff++;
	   ($ct_eq == $ct_seqs) and ($ct_inv++);
	}			# End of the 2nd outer foreach-loop #

	return ($ct_eff, $ct_inv);

}		# end of the subroutine 'count_inv_sites' #


# fabricate_pads: This subroutine makes pads for merging alignments longitudinally,
# 		from coordinate information
#	 	and by concatenating ambiguous characters 'n' or 'x'
#		and flanking them  with gaps
#
# The subroutine regards overlapping alignments as 2 adjacent  ones,
# and adjusts the pad lengths accordingly.
#
# OUTPUT: [ [\$pad_seq1, \$pad_seq2, ...]_interaln_1, 
#		[\$pad_seq1, \$pad_seq2, ...]_interaln_2, ... ]
#
sub fabricate_pads (\@$) {

	my ($rf_coord, $ambi) =  @_;
	my $ct_seqs = @{$rf_coord->[0]};

	my @out_pads = ();

	for (my $i=1; $i<@{$rf_coord}; $i++) {	# the outer for #

		my @pad_lens = ();
		my $max_pad_len = 0;
		my $neg_pad_len = 0;
		for (my $k = 0; $k<$ct_seqs; $k++) { # the inner for #

			my $pres_init = $rf_coord->[$i][$k][0];
			my $last_end = $rf_coord->[$i-1][$k][1];
			my $pad_len = ($pres_init < $last_end) ?
				$last_end - $pres_init -1 : $pres_init - $last_end -1;
			push @pad_lens , $pad_len;
			($pad_len > $max_pad_len) and ($max_pad_len = $pad_len);
			($pad_len < $neg_pad_len) and ($neg_pad_len = $pad_len);

		}			# end of the inner for #

		if ($neg_pad_len < 0) {
			for (my $k=0; $k<@pad_lens; $k++) { 
				$pad_lens[$k] -= $neg_pad_len;
			}
			$max_pad_len -= $neg_pad_len;
		}

		my @pads = ();

		foreach my $pad_len (@pad_lens) { # the inner foreach #

			my $ct_gaps = $max_pad_len -  $pad_len;
			my $mid_ct = int($ct_gaps/2);

			my $pad = '-' x $mid_ct . $ambi x $pad_len 
				 	. '-' x ($ct_gaps - $mid_ct);

			push @pads, \$pad;
		}				# end of the inner foreach #

		push @out_pads, \@pads;
	}				# end of the outer for #

	return \@out_pads;

}			# end of the subroutine fabricate_pads #


# pad_5p_end: This subroutine receives an incomplete multiple-alignment
#		and pads its 5'-end with ambiguous characters ('x' for peptides, 'n' for nucleoties)
#		flanked with gaps.
#
# INPUT FORMAT:
#    $rf_align = [ [\$seq1, \$seq2, ...]_align1, [\$seq1, \$seq2, ...]_align2, ... ],
#    $rf_coord = [ [ [init, end]_seq1, [init, end]_seq2, ... ]_align1,
#		   [ [init, end]_seq1, [init, end]_seq2, ... ]_aling2,
#		   ... ],
#    $rf_len = [ len_seq1, len_seq2, ... ],
#    $fl_pep = 1 for a peptide alignment, =0 for a nucleotide alignment.
#
sub pad_5p_end (\@\@\@$) {

	my ($rf_align, $rf_coord, $rf_len, $fl_pep ) = @_;


	my $ambi = $fl_pep ? 'x' : 'n'; 
		# $fl_pep =1 for peptides, =0 for nucleotides #

	my $ct_align = @{$rf_align};

	unless ($ct_align ) {
		print STDERR "No alignments available.\n";
		 return 0;
	}

	unless ($ct_align == @{$rf_coord}) {
		printf STDERR "Discrepancy between the number of alignments (%d) and that of coordinate information (%d).\n",
		 $ct_align, scalar(@{$rf_coord});
		return 0;
	}

		# initialize #
	my @init_coord = @{$rf_coord->[0]};
	my $ct_seqs = @{$rf_align->[0]};

	unless (@init_coord == $ct_seqs) {
		printf STDERR "Discrepancy bewteen the number of coordinates (%d) and that of sequences ($ct_seqs).\n", scalar (@init_coord) ;
		return 0;

	}

	unless ($ct_seqs == @{$rf_len}) { 
		printf STDERR "Discrepancy between the number of sequences ($ct_seqs) and that of sequence lengths (%d).\n", scalar(@{$rf_len}) ;
		return 0;
	}

	my @flank_coord = ();
	for (my $k = 0; $k <$ct_seqs; $k++) {
	    my ($init, $end) = @{$init_coord[$k]};
	    my $len = $rf_len->[$k]; 
	    push @flank_coord, ( ($init < $end ) ? [-1, 0] : [$len + 2, $len + 1] ) ;
	}

	my @dummy_coord = (\@flank_coord, \@init_coord);
	my $rf_pads = fabricate_pads (@dummy_coord, $ambi);

	for (my $k=0; $k< $ct_seqs; $k++) {# the inner for #

		${$rf_align->[0][$k]} =
		    ${$rf_pads->[0][$k]} . ${$rf_align->[0][$k]} ;
		my $fl_coord = $flank_coord[$k]->[1];
		$rf_coord->[0][$k][0] = $fl_coord ? ($fl_coord -1) : 1;

	}			# end of the inner for #

	return 1;

}		# end of the subroutine pad_5p_end #


# pad_3p_end: This subroutine receives an incomplete multiple-alignment
#		and pads its 3'-end with ambiguous characters ('x' for peptides, 'n' for nucleoties)
#		flanked with gaps.
#
# INPUT FORMAT:
#    $rf_align = [ [\$seq1, \$seq2, ...]_align1, [\$seq1, \$seq2, ...]_align2, ... ],
#    $rf_coord = [ [ [init, end]_seq1, [init, end]_seq2, ... ]_align1,
#		   [ [init, end]_seq1, [init, end]_seq2, ... ]_aling2,
#		   ... ],
#    $rf_len = [ len_seq1, len_seq2, ... ],
#    $fl_pep = 1 for a peptide alignment, =0 for a nucleotide alignment.
#
sub pad_3p_end (\@\@\@$) {

	my ($rf_align, $rf_coord, $rf_len, $fl_pep ) = @_;


	my $ambi = $fl_pep ? 'x' : 'n'; 
		# $fl_pep =1 for peptides, =0 for nucleotides #

	my $ct_align = @{$rf_align};

	unless ($ct_align ) {
		print STDERR "No alignments available.\n";
		 return 0;
	}

	unless ($ct_align == @{$rf_coord}) {
		printf STDERR "Discrepancy between the number of alignments (%d) and that of coordinate information (%d).\n",
		 $ct_align, scalar(@{$rf_coord});
		return 0;
	}

		# initialize #
	my @fin_coord = @{$rf_coord->[$ct_align -1]};
	my $ct_seqs = @{$rf_align->[$ct_align -1]};

	unless (@fin_coord == $ct_seqs) {
		printf STDERR "Discrepancy bewteen the number of coordinates (%d) and that of sequences ($ct_seqs).\n", scalar (@fin_coord) ;
		return 0;

	}

	unless ($ct_seqs == @{$rf_len}) { 
		printf STDERR "Discrepancy between the number of sequences ($ct_seqs) and the number of sequence lengths (%d).\n", scalar(@{$rf_len}) ;
		return 0;
	}

	my @flank_coord = ();
	for (my $k = 0; $k <$ct_seqs; $k++) {
	    my ($init, $end) = @{$fin_coord[$k]};
	    my $len = $rf_len->[$k]; 
	    push @flank_coord, ( ($init < $end ) ? [$len + 1, $len + 2] : [0, -1] ) ;
	}

	my @dummy_coord = (\@fin_coord, \@flank_coord);
	my $rf_pads = fabricate_pads (@dummy_coord, $ambi);

	for (my $k=0; $k< $ct_seqs; $k++) {# the inner for #

		${$rf_align->[0][$k]} .= ${$rf_pads->[0][$k]};
		my $fl_coord = $flank_coord[$k]->[0];
		$rf_coord->[$ct_align-1][$k][1] = $fl_coord ? ($fl_coord -1) : 1;

	}			# end of the inner for #

	return 1;

}		# end of the subroutine pad_3p_end #


# merge_pad_align : This subroutine receives more than 1 alignment constructed from
#		   the same sequence set,
#	            and, using their coordinate information,
#		    merges the 2 contiguous alignments
#		    while padding the 'opening's with 'n' for the nucleotides,
#		    and with 'x' for the peptides, along with gaps '-'.
#		  The subroutine also returns the start and end coordinates
#		 of sequences in the composite alignment.
#
# PREREQUISITE: @{$rf_coord} = 
#		([ [init, end]_seq1, ..]_align1, [ [init, end]_seq1, ..]_align2, ...)
#		 are supposed to be arranged in a regular order.
#
sub merge_pad_align (\@\@$) {

	my ($rf_align, $rf_coord, $fl_pep) = @_;

	my $ambi = $fl_pep ? 'x' : 'n'; 
		# $fl_pep =1 for peptides, =0 for nucleotides #

	my $ct_align = @{$rf_align};
	unless ($ct_align == @{$rf_coord}) {
		printf STDERR "Discrepancy between the number of alignments %d and the number of coordinate information %d.\n",
		 $ct_align, scalar(@{$rf_coord});
		return (undef, undef);
	}

	## ($ct_align == 1) and (return ($rf_align, $rf_coord) ); # a bug-fix on Nuv 18 (Tue), 2003 #


	my $rf_pads = fabricate_pads (@{$rf_coord}, $ambi);

		# initialize #
	my @merged = ();
	my $ct_seqs = @{$rf_align->[0]};
	for (my $k=0; $k<$ct_seqs; $k++) {
		my $tmp = ${$rf_align->[0][$k]};
		push @merged , \$tmp;
	}

	for (my $i=1; $i<$ct_align; $i++) {	# the outer for #

		for (my $k=0; $k< $ct_seqs; $k++) {# the inner for #

			${$merged[$k]} .=
			    ${$rf_pads->[$i-1][$k]} . ${$rf_align->[$i][$k]} ;

		}			# end of the inner for #
	}				# end of the outer for #

	my @end_coords = ();
	for (my $k = 0; $k<$ct_seqs; $k++) {
		my $init = $rf_coord->[0][$k][0];
		my $end = $rf_coord->[$ct_align -1][$k][1];

		unless ( (defined $init) and (defined $end) ) {
			print STDERR "Failed to retrieve end coordinates\n";
			return (\@merged, undef);
		}

		push @end_coords, [$init, $end];
	}

	return (\@merged, \@end_coords);

}			# end of the subroutine merge_pad_align #


# rm_gapcolumn: This subroutine receives the sequence part of a multiple-alignment,
#		removes the columns consisting exclusively of the gap '-',
#		and returns (the reference to ) the resulting sequences.
#
# CAUTION: This subroutine 'chews up' the input sequences. So, watch out !!
#
sub rm_gapcolumn (\@) {

	my ($rf_align) = @_;

	my $ct_seqs = @{$rf_align};
	$ct_seqs or (return []);

	my @refined  = ();
	for (my $i=0; $i<$ct_seqs; $i++) {
		my $tmp = '';
		push @refined , \$tmp;
	}

	my $seq_len = length ${$rf_align->[0]};

	foreach my $l (1 .. $seq_len) { 	# outer foreach #

		my @reservoir = ();
		my $ct_gap = 0;
		foreach my $rf_seq (@{$rf_align}) {
			my $char = substr(${$rf_seq}, 0, 1, '');	
			push @reservoir, $char;
			($char eq '-') and ($ct_gap++); 
		}

		($ct_gap == $ct_seqs) and  next;

		for (my $i=0; $i<$ct_seqs; $i++) { ${$refined[$i]} .= $reservoir[$i]; }

	}					# end of the outer foreach #

	return \@refined;

}			# end of the subroutine rm_gapcolumn #


# rm_gapcolumn2:
#	This revised version of the subroutine 'rm_gapcolumn'
#	receives the sequence part of a multiple-alignment,
#	removes the columns consisting exclusively of the gap, 
#	which is specified by an input character,
#	and returns (the reference to ) the resulting sequences.
#
# CAUTION: This subroutine 'chews up' the input sequences. So, watch out !!
#
sub rm_gapcolumn2 (\@$) {

	my ($rf_align, $char_gap) = @_;

	my $ct_seqs = @{$rf_align};
	$ct_seqs or (return []);

	my @refined  = ();
	for (my $i=0; $i<$ct_seqs; $i++) {
		my $tmp = '';
		push @refined , \$tmp;
	}

	my $seq_len = length ${$rf_align->[0]};

	foreach my $l (1 .. $seq_len) { 	# outer foreach #

		my @reservoir = ();
		my $ct_gap = 0;
		foreach my $rf_seq (@{$rf_align}) {
			my $char = substr(${$rf_seq}, 0, 1, '');	
			push @reservoir, $char;
			($char eq $char_gap) and ($ct_gap++); 
		}

		($ct_gap == $ct_seqs) and  next;

		for (my $i=0; $i<$ct_seqs; $i++) { ${$refined[$i]} .= $reservoir[$i]; }

	}					# end of the outer foreach #

	return \@refined;

}			# end of the subroutine rm_gapcolumn2 #


# rm_gapcolumn_codon: This subroutine receives the sequence part of a multiple-alignment of codon sequences, 
#		 removes the columns consisting exclusively of the "gap codon" '---',
#		and returns (the reference to) the resulting sequences.
#
sub rm_gapcolumn_codon (\@) {

	my ($rf_align) = @_;

	my $ct_seqs = @{$rf_align};
	$ct_seqs or (return ([], 1) );

	my @refined = ();
	for (my $i=0; $i<$ct_seqs; $i++) {
		my $tmp = '';
		push @refined , \$tmp;
	}

	my $seq_len = length ${$rf_align->[0]};
	my $codon_ct = int ($seq_len / 3);
	my $fl_caution = ($seq_len % 3)? 0 : 1;	# =0 if error occurs, =1 if normal #

	foreach my $l (1 .. $codon_ct) { 	# outer foreach #

		my @reservoir = ();
		my $ct_gap = 0;
		foreach my $rf_seq (@{$rf_align}) {
			my $codon = substr(${$rf_seq}, 0, 3, '');	
			push @reservoir, $codon;
			($codon eq '---') and ($ct_gap++); 
		}

		($ct_gap == $ct_seqs) and  next;

		for (my $i=0; $i<$ct_seqs; $i++) { ${$refined[$i]} .= $reservoir[$i]; }

	}					# end of the outer foreach #


	return (\@refined, $fl_caution);
}			# end of the subroutine rm_gapcolumn_codon #


# assign_number: This subroutine recieves the sequence part of a multiple-alignment,
#		constructs a list of residue numbers residing in  each alignment column,
#		and returns the reference to an array of the lists.
#
# NOTE: This subroutine adopts the (START=1) convention.
#
sub assign_number (\@) {

	my ($rf_align) = @_;

	my $ct_seqs = @{$rf_align};
	$ct_seqs or (return []);

	my $seq_len = length (${$rf_align->[0]} );
	my @res_nums = ();
	my @copy = ();
	my @last_resnums;
	foreach my $seq (@{$rf_align}) {
			push @copy, ${$seq};
			push @last_resnums, 0; 
	}

	for (my $l=0; $l<$seq_len; $l++) {

		my @column = ();
		for (my $i=0; $i<$ct_seqs; $i++) {
			my $char = substr($copy[$i], 0, 1, '');
			if ($char eq '-')  {
				push @column, '-';
			} else {
				$last_resnums[$i]++;
				push @column, $last_resnums[$i];
			}
		}

		push @res_nums, \@column;
	}

	return \@res_nums;

}			# end of the subroutine assign_number #


# assign_number_gen:
#	This revised version of 'assign_number' can cope with
#	'fragmental' sequences begining at the sites different than
#	coordinate=1.
#	
# assign_number:
#	This subroutine recieves the sequence part of a multiple-alignment,
#	constructs a list of residue numbers residing in  each alignment column,
#	and returns the reference to an array of the lists.
#
# INPUT: (\@align, \@beg_coords, $gap), where
#	${$align[$m]} is the $m th sequence embedded in the alignment;
#	$beg_coords[$m] is the beginning coordinate of the $m th sequence.
#
# OUTPUT: \@relv_coords, where
#	$relv_coords[$l][$m] is the coordinate of the $l th site of the $m th sequence,
#	  or '-' if it is occupied with a gap ($gap).
#
sub assign_number_gen (\@\@$) {

	my ($rf_align, $beg_coords, $gap) = @_;

	my $ct_seqs = @{$rf_align};
	$ct_seqs or (return []);

	my $seq_len = length (${$rf_align->[0]});
	my @res_nums = ();
	my @copy = ();
	my @last_resnums;
	for(my $m=0; $m<$ct_seqs; $m++) {
	    push @copy, ${$rf_align->[$m]};
	    my $beg = $beg_coords->[$m]; 
	    push @last_resnums, ( ($beg eq int ($beg)) ? $beg -1 : int ($beg) );
	}

	for (my $l=0; $l<$seq_len; $l++) {
	    my @column = ();
	    for (my $i=0; $i<$ct_seqs; $i++) {
		my $char = substr($copy[$i], 0, 1, '');
		if ($char eq $gap)  {
		    push @column, '-';
		} else {
		    $last_resnums[$i]++;
		    push @column, $last_resnums[$i];
		}
	    }
	    push @res_nums, \@column;
	}

	return \@res_nums;

}		# end of the subroutine 'assign_number_gen' #


# assign_number_cnctseqs:
#	This revised version of 'assign_number_gen' can
#	handle concatenated sequences as well.
#
# assign_number_gen:
#	This revised version of 'assign_number' can cope with
#	'fragmental' sequences begining at the sites different than
#	coordinate=1.
#	
# assign_number:
#	This subroutine recieves the sequence part of a multiple-alignment,
#	constructs a list of residue numbers residing in  each alignment column,
#	and returns the reference to an array of the lists.
#
# INPUT: (\@align, \@beg_coords, \@end_coords, $gap), where
#	${$align[$m]} is the $m th sequence embedded in the alignment;
#	$beg_coords[$m] and $end_coords[$m] are the beginning and ending
#	coordinates, respectively, of the $m th sequence.
#
# OUTPUT: \@relv_coords, where
#	$relv_coords[$l][$m] is the coordinate of the $l th site of the $m th sequence,
#	  or '-' if it is occupied with a gap ($gap).
#
sub assign_number_cnctseqs (\@\@\@$) {

	my ($rf_align, $beg_coords, $end_coords, $gap) = @_;

	my $ct_seqs = @{$rf_align};
	$ct_seqs or (return []);

	unless ((@{$beg_coords} == $ct_seqs) and (@{$end_coords} == $ct_seqs)) {
	    print STDERR "[ERROR 'assign_number_cnctseqs'] Either the number of the beginning coordinates (", scalar (@{$beg_coords}), ") or that of the ending coordinates (", scalar (@{$end_coords}), ") CLASHES with that of the sequences $ct_seqs !\n";
	    return undef;
	}

	my @intervals = ();

	my $seq_len = length (${$rf_align->[0]});
	my @res_nums = ();
	my @copy = ();
	my @last_resnums = my @intvls = ();
	for(my $m=0; $m<$ct_seqs; $m++) {
	    my @leftbds = split /:/, $beg_coords->[$m];
	    my @rightbds = split /:/, $end_coords->[$m];
	    my $ct_sbalns = @leftbds;
	    unless (@rightbds == $ct_sbalns) {
		print STDERR "[ERROR 'assign_number_cnctseqs'] The number of left boundaries ($ct_sbalns) CLASHES with that of right boundaries (", scalar (@rightbds), ") for the $m th sequence !\n";
		return undef;
	    }
	    my @intvls_per_seq = ();
	    for (my $r=0; $r<$ct_sbalns; $r++) {
		my $lb = $leftbds[$r];
		($lb == int $lb) or ($lb = 1 + int $lb);
		my $rb = int $rightbds[$r];
		push @intvls_per_seq, [$lb, $rb];
	    }
	    push @intvls, \@intvls_per_seq;
	    push @copy, ${$rf_align->[$m]};
	    ##my $beg = $beg_coords->[$m]; 
	    ##push @last_resnums, ( ($beg eq int ($beg)) ? $beg -1 : int ($beg) );
	    push @last_resnums, $intvls[$m]->[0][0] -1;
	}

	for (my $l=0; $l<$seq_len; $l++) {
	    my @column = ();
	    for (my $m=0; $m<$ct_seqs; $m++) {
		my $char = substr($copy[$m], 0, 1, '');
		if ($char eq $gap)  {
		    push @column, '-';
		} else {
		    ##$last_resnums[$m]++;
		    my $curr_resnum = $last_resnums[$m]+1;
		    if ($curr_resnum > $intvls[$m]->[0][1]) {
			# If the coordinate exceeds the right boundary of
			# the current interval, jump to the next interval.
			shift @{$intvls[$m]};

			unless (@{$intvls[$m]}>0) {
			    print STDERR "[CAUTION 'assign_number_cnctseqs'] The $m th sequence exceeded the rightmost boundary, so I can't go any further !!\n\n";
			    return \@res_nums;
			}
			$curr_resnum = $intvls[$m]->[0][0];
		    }
		    push @column, $curr_resnum;
		    $last_resnums[$m] = $curr_resnum;
		}
	    }
	    push @res_nums, \@column;
	}

	return \@res_nums;

}		# end of the subroutine 'assign_number_cnctseqs' #


# map_resnum2alncoord:
#	This subroutine receives a sequence embedded in an alignment,
#	then constructs a map from residue numbers of the sequence
#	to positions in the alignment.
#
# INPUT: (\$embedded_sequence, \%gaps), where
#	%gaps = (a gap character => 1, ...).
#	We regard characters other than gap characters as representing residues.
#
# OUTPUT: \@map, where
#    $map[$res_num -1] = $pos_aln,
#	with $res_num and $pos_aln being 1 based.
#
sub map_resnum2alncoord (\$\%) {

	my ($emb_seq, $gaps) = @_;

	my $cp_emb_seq = ${$emb_seq};
	my $aln_len = length $cp_emb_seq;
	my @map = ();

	for (my $l=0; $l < $aln_len; $l++) {
	    my $char = substr ($cp_emb_seq, 0, 1, ''); 
	    (defined $gaps->{$char}) and next;
	    push @map, $l+1;
	}

	return \@map;

}		# end of the subroutine map_resnum2alncoord #


# map_resnum2alncoord_2:
# 	This revised version of the subroutine 'map_resnum2alncoord'
#	takes residue numbers of start and end residues.
#	And it assigns the alignment coordinate of -1 for the upstream franking region,
#	and aln_len + 1 for the downstream franking region.
#
# map_resnum2alncoord:
#	This subroutine receives a sequence embedded in an alignment,
#	then constructs a map from residue numbers of the sequence
#	to positions in the alignment.
#
# INPUT: (\$embedded_sequence, \%gaps, \@startend, $seqlen), where
#	%gaps = (a gap character => 1, ...).
#	We regard characters other than gap characters as representing residues.
#	@startend = ($beg, $end) is the residue numbers of the start and end residues
#	(with 1 representing the N-terminal end of the peptide,
#		or the 5'-end of the polynucleotide).
#
# OUTPUT: \@map, where
#    $map[$res_num -1] = $pos_aln if the $res_num-th residue is included in the alignment,
#			with $res_num and $pos_aln being 1 based;
#			=  -1 if the residue is upstream of the alignment;
#			= $aln_len + 1 if the residue is downstream of the alignment.
#
sub map_resnum2alncoord_2 (\$\%\@$) {

	my ($emb_seq, $gaps, $startend, $seqlen) = @_;

	my $cp_emb_seq = ${$emb_seq};
	my $aln_len = length $cp_emb_seq;
	my ($beg, $end) = @{$startend};
	my @map = ();

	$beg--; $end--; # convert from 1-based to 0-based #

	for (my $i=0; $i<$beg; $i++) { push @map, -1; }

	for (my $l=0; $l < $aln_len; $l++) {
	    my $char = substr ($cp_emb_seq, 0, 1, ''); 
	    (defined $gaps->{$char}) and next;
	    push @map, $l+1;
	}

	for (my $i=$end+1; $i<$seqlen; $i++) { $map[$i] = $aln_len + 1; }

	return \@map;

}		# end of the subroutine map_resnum2alncoord_2 #


# separate_nogap_sites: This subroutine receives the sequence part of a multiple alignment,
#			  separates columns into two groups according to whether it contains
#			  gaps '-' (or N) or not,
#			  and returns the lists of columns thus separated and of their positions
#			  (under the "start = 0" convention). 
#
sub separate_nogap_sites (\@\@) {

	my ($rf_align, $rf_positions) = @_;

	my $ct_seqs = @{$rf_align};
	$ct_seqs  or (return ([],[],[],[]) );

	my @copy = ();
	foreach my $rf_seq (@{$rf_align}) { push @copy, ${$rf_seq}; }

	my $aln_len = length $copy[0];
	my $fl_pos = (@{$rf_positions} == $aln_len);

	my @nogap_sites = my @gapped_sites = my @nogap_pos = my @gapped_pos = ();

	for (my $l=0; $l <$aln_len; $l++) {	# the outer for #

		my @column = ();
		my $ct_gaps = 0;
		for (my $i=0; $i < $ct_seqs; $i++) {	# the inner for #

			my $char = substr ( $copy[$i], 0, 1, '');
			( ($char eq '-') or ($char eq 'N') ) and ($ct_gaps++);
			push @column, $char;
		}				# end of the inner for #

		my $pos = $fl_pos ? $rf_positions->[$l] : $l;
		if ($ct_gaps) {

			push @gapped_sites , \@column;
			push @gapped_pos, $pos;

		} else {

			push @nogap_sites, \@column;
			push @nogap_pos, $pos;
		}
	}					# end of the outer for #

	return (\@nogap_sites, \@nogap_pos, \@gapped_sites, \@gapped_pos);

}			# end of the subroutine separate_nogap_sites #

# separate_resnum: This subroutine separates arrays of residue numbers,
#		   each for each column,
#		   into those of  gapped part and and nogap part.
#
sub separate_resnum (\@\@\@) {

	my ($rf_resnums, $rf_nogap_pos, $rf_gapped_pos) = @_;

	my $ct_clms = @{$rf_resnums};
	$ct_clms or (return ([],[]) );

	my @copy = @{$rf_resnums};
	@{$rf_nogap_pos} or (return ([], \@copy) );
	@{$rf_gapped_pos} or (return (\@copy, []) );

	if ( ($rf_nogap_pos->[$#{$rf_nogap_pos}] > $ct_clms -1) 
		or ($rf_gapped_pos->[$#{$rf_gapped_pos}] > $ct_clms -1) ) {

		print STDERR "Illegal input for \"separate_resnum\"!\n";
		return (undef, undef);
	}

	my @ng_resnums = my @gpd_resnums = ();

	foreach my $ng_pos (@{$rf_nogap_pos}) { push @ng_resnums, $rf_resnums->[$ng_pos]; }

	foreach my $gpd_pos (@{$rf_gapped_pos}) { push @gpd_resnums, $rf_resnums->[$gpd_pos]; }

	return (\@ng_resnums, \@gpd_resnums);

}			# end of the subroutine separate_resnum #



# compare_columns: This subroutine compares 2 lists of residue numbers,
#		each corresponding to a column in a multiple-alignment,
#		and return -1, 0, or +1 if it judged the 1st column being 'earlier' (or 'less')
#		than the 2nd one.
#
# REVISED on Nov 19 (Fri), 2004:
#	The subroutine no longer uses the recursion technique,
#	which is unnecessary for doing the task.
#	The 2nd argument is kept just for the backward compatibility.
#	You should set it to 0 without any special reasons.
#
sub compare_columns (\@\@$) {

	my ($column1, $column2, $init_indx) = @_;

		# If one of the columns is empty, we regard them as 'equal'.
	( (@{$column1} > 0) and (@{$column2} > 0) ) or (return 0) ; 

	##my $last_indx = (@{$column1} < @{$column2}) ? $#{$column1} : $#{$column2} ;
 	my $min_size = (@{$column1} < @{$column2}) ? @{$column1} : @{$column2} ;

	for (my $i= $init_indx; $i< $min_size; $i++) { # the for-loop #

	    my $resnum1 = $column1->[$i];
	    my $resnum2 = $column2->[$i];

	    if ( $resnum1 eq '-') {

		($resnum2 eq '-') ? next : ( return -1 ) ;

	    } elsif ($resnum2 eq '-') {
			return +1;

	    } elsif ($resnum1 < $resnum2) {
			return -1;

	    } elsif ($resnum1 > $resnum2) {
			return +1;
	    }
	} 				# end of the for-loop #

		# Now the colums are judged as 'equal' to each other. #
	return 0;

}			# end of the subroutine compare_columns #


=the original version
# compare_columns: This subroutine compares 2 lists of residue numbers,
#		each corresponding to a column in a multiple-alignment,
#		and return -1, 0, or +1 if it judged the 1st column being 'earlier' (or 'less')
#		than the 2nd one.
#
sub compare_columns (\@\@$) {

	my ($column1, $column2, $indx) = @_;

	( (@{$column1} > 0) and (@{$column2} > 0) ) or (return 0) ; 
	my $last_indx = (@{$column1} < @{$column2}) ? $#{$column1} : $#{$column2} ;
	my $resnum1 = $column1->[$indx];
	my $resnum2 = $column2->[$indx];

	if ( $resnum1 eq '-') {

		if ($resnum2 eq '-') {
			return ($indx >= $last_indx) ? 0 : compare_columns(@{$column1},@{$column2},$indx+1);
		} else {
			return -1;
		}

	} elsif ($resnum2 eq '-') {
			return +1;

	} elsif ($resnum1 < $resnum2) {
			return -1;

	} elsif ($resnum1 > $resnum2) {
			return +1;

	} else {
		return ($indx >= $last_indx) ? 0 : compare_columns(@{$column1},@{$column2},$indx+1);
	}

}			# end of the subroutine compare_columns #
=cut


# compare_malign_fw: This subroutine recieves 2 arrays of lists of residue numbers in the alignment columns,
#		looks for columns which are different between these arrays via a FORWARD SWEEP,
#		and returns the differences and their positions in the frame of each alignments.
#
# PREREQUISITE: The sequence sets used in the 2 alignments have to be identical.
#
sub compare_malign_fw (\@\@\@\@) {

	my ($rf_resnums1, $rf_resnums2, $rf_frame1, $rf_frame2) = @_;

	my $ct_column1 = @{$rf_resnums1};
	my $ct_column2 = @{$rf_resnums2};
	my @frames1 = (@{$rf_frame1}) ? @{$rf_frame1} : (0 .. $ct_column1 -1);
	my @frames2 = (@{$rf_frame2}) ? @{$rf_frame2} : (0 .. $ct_column2 -1);

	unless ($ct_column1) {
		return ([], $rf_resnums2, [], \@frames2);
	}

	unless ($ct_column2) {
		my @frame_nums = (0 .. $ct_column1 -1);
		return ($rf_resnums1, [], \@frames1, []);
	}

	my $frame1 = my $frame2 = 0;
	my @ext1 = my @ext2 = ();
	my @ext_frames1  = my @ext_frames2 = ();
	#my $last_sign = 0;		# assume that the beginnings coincide for the 2 alignments #

		# processing the bulk part #
	while ( ($frame1 < $ct_column1) and ($frame2 < $ct_column2) ) {

		my $column1 = $rf_resnums1->[$frame1];
		my $column2 = $rf_resnums2->[$frame2];
		my $sign = compare_columns (@{$column1}, @{$column2}, 0);

		if ($sign == 0) {

			$frame1++;
			$frame2++;
			
		} elsif ($sign < 0) { # column1 is 'less than' column2 #

			push @ext1, $column1;
			push @ext_frames1, $frames1[$frame1];
			$frame1++;

		} else {	# column2 is 'less than' column1 #

			push @ext2, $column2;
			push @ext_frames2, $frames2[$frame2];
			$frame2++;

		}
	}				# end of while #

		# take care of the marginal part #
	if ( ($frame1 >= $ct_column1) and ($frame2 < $ct_column2) ) {

		push @ext2, @{$rf_resnums2}[$frame2 .. $ct_column2 -1];
		push @ext_frames2, @frames2[$frame2 .. $ct_column2 -1];

	} elsif ( ($frame1 < $ct_column1) and ($frame2 >= $ct_column2) ) {

		push @ext1, @{$rf_resnums1}[$frame1 .. $ct_column1 -1];
		push @ext_frames1, @frames1[$frame1 .. $ct_column1 -1];
	}

	return (\@ext1, \@ext2, \@ext_frames1, \@ext_frames2);

}			# end of the subroutine compare_malign_fw #


# compare_malign_bw: This subroutine recieves 2 arrays of lists of residue numbers in the alignment columns,
#		looks for columns which are different between these arrays via a BACKWARD SWEEP,
#		and returns the differences and their positions in the frame of each alignments.
#
# PREREQUISITE: The sequence sets used in the 2 alignments have to be identical.
#
sub compare_malign_bw (\@\@\@\@) {

	my ($rf_resnums1, $rf_resnums2, $rf_frame1, $rf_frame2) = @_;

	my $ct_column1 = @{$rf_resnums1};
	my $ct_column2 = @{$rf_resnums2};
	my @frames1 = (@{$rf_frame1}) ? @{$rf_frame1} : (0 .. $ct_column1 -1);
	my @frames2 = (@{$rf_frame2}) ? @{$rf_frame2} : (0 .. $ct_column2 -1);

	unless ($ct_column1) {
		return ([], $rf_resnums2, [], \@frames2);
	}

	unless ($ct_column2) {
		return ($rf_resnums1, [], \@frames1, []);
	}

	my $frame1 = $ct_column1 -1;
	my $frame2 = $ct_column2 -1;
	my @ext1 = my @ext2 = ();
	my @ext_frames1  = my @ext_frames2 = ();
	#my $last_sign = 0;		# assume that the beginnings coincide for the 2 alignments #

		# processing the bulk part #
	while ( ($frame1 >= 0) and ($frame2 >= 0) ) {

		my $column1 = $rf_resnums1->[$frame1];
		my $column2 = $rf_resnums2->[$frame2];
		my $sign = compare_columns (@{$column1}, @{$column2}, 0);

		if ($sign == 0) {

			$frame1--;
			$frame2--;
			
		} elsif ($sign > 0) { # column1 is 'greater than' column2 #

			unshift @ext1, $column1;
			unshift @ext_frames1, $frames1[$frame1];
			$frame1--;

		} else {	# column2 is 'greater than' column1 #

			unshift @ext2, $column2;
			unshift @ext_frames2, $frames2[$frame2];
			$frame2--;

		}
	}				# end of while #

		# take care of the marginal part #
	if ( ($frame1 < 0) and ($frame2 >= 0) ) {

		unshift @ext2, @{$rf_resnums2}[0 .. $frame2];
		unshift @ext_frames2, @frames2[0 .. $frame2];

	} elsif ( ($frame1 >= 0) and ($frame2 < 0) ) {

		unshift @ext1, @{$rf_resnums1}[0 .. $frame1];
		unshift @ext_frames1, @frames1[0 .. $frame1];
	}

	return (\@ext1, \@ext2, \@ext_frames1, \@ext_frames2);

}			# end of the subroutine compare_malign_bw #


# compare_malign: This subroutine recieves 2 arrays of lists of residue numbers in the alignment columns,
#		looks for columns which are different between these arrays,
#		and returns the differences and their positions in the frame of each alignments.
#
# PREREQUISITE: The sequence sets used in the 2 alignments have to be identical.
#
sub compare_malign (\@\@\@\@) {

	my ($rf_resnums1, $rf_resnums2, $rf_frame1, $rf_frame2) = @_;

	my $ct_column1 = @{$rf_resnums1};
	my $ct_column2 = @{$rf_resnums2};
	my @frames1 = (@{$rf_frame1}) ? @{$rf_frame1} : (0 .. $ct_column1 -1);
	my @frames2 = (@{$rf_frame2}) ? @{$rf_frame2} : (0 .. $ct_column2 -1);

	unless ($ct_column1) {
		return ([], $rf_resnums2, [], \@frames2);
	}

	unless ($ct_column2) {
		return ($rf_resnums1, [], \@frames1, []);
	}

	my ($rf_ext1, $rf_ext2, $extframes1, $extframes2) = compare_malign_fw (@{$rf_resnums1}, @{$rf_resnums2}, @frames1, @frames2);

	($rf_ext1, $rf_ext2, $extframes1, $extframes2) = compare_malign_bw (@{$rf_ext1}, @{$rf_ext2}, @{$extframes1}, @{$extframes2});

	return ($rf_ext1, $rf_ext2, $extframes1, $extframes2);

}				# end of the subroutine compare_malign #


# diff_malign: This subroutine extracts portins of a multiple alignment
#		which are different from another alignment,
#		referring to the output of "compare_malign".
# 
sub diff_malign (\@\@) {

	my ($rf_align, $rf_exframe) = @_;

	(@{$rf_exframe}) or ( return []); 

	my @copy = my @diffs = ();
	foreach my $rf_seq (@{$rf_align}) {
		push @copy, ${$rf_seq};
	}

	my $exp_frame = 0;
	foreach my $frame (@{$rf_exframe}) {

		if ( (my $diff = $frame - $exp_frame) > 0)  {

			for (my $i=0; $i<@copy; $i++) { substr($copy[$i], 0, $diff, ''); }

		} elsif ($diff < 0) {

			print STDERR "Error in diff_malign:\n";
			print STDERR "expected frame ($exp_frame) exceeds the real one ($frame).\n";
			return undef;
		}

		my @column = ();
		for (my $i=0; $i<@copy; $i++) {

			my $res = substr($copy[$i], 0, 1, '');
			push @column, $res;
		}

		push @diffs, \@column;
		$exp_frame = $frame + 1;
	}

	return \@diffs;
}			# end of the subroutine diff_malign #

# construct_blocks: This suboutine puts together consecutive columns into a block,
#			amd returns a set of blocks.
#
sub construct_blocks (\@\@\@$) {

	my ($rf_align, $rf_resnums, $rf_frames, $thrsh_distcolumns) = @_;
	
	@{$rf_frames} or (return ([], [], []) );
	my @block_columns = my @block_resnums = my @block_frames = ();

	my @columns = ( $rf_align->[0]);
	my @resnums = ( $rf_resnums->[0]);
	my @frames = ( $rf_frames->[0]);

	for (my $i=1; $i<@{$rf_frames}; $i++) {		# a for-loop #

		my $last_frame = $frames[$#frames];
	 	if ($rf_frames->[$i] <=  $last_frame + $thrsh_distcolumns) {

			push @columns,  $rf_align->[$i];
			push @resnums,  $rf_resnums->[$i];
			push @frames, $rf_frames->[$i];

		} elsif ($rf_frames->[$i] > $last_frame) {

			my @tmp = @columns;
			push @block_columns, \@tmp;
			my @tmp2 = @resnums;
			push @block_resnums, \@tmp2;
			my @tmp3 = @frames;
			push @block_frames, \@tmp3;
			@columns = ( $rf_align->[$i]);
			@resnums = ( $rf_resnums->[$i]);
			@frames = ( $rf_frames->[$i]);
		} else {

			print STDERR "The present frame ($rf_frames->[$i]) is less than the last frame $last_frame.\n";
			return (undef, undef, undef);
		}

	}					# end of the for-loop #

	push @block_columns, \@columns;
	push @block_resnums, \@resnums;
	push @block_frames, \@frames;

	return (\@block_columns, \@block_resnums, \@block_frames);

}				# end of the subroutine construct_blocks #


# extract_bl_selected_frames: This subroutine extracts a subset of columns and residue numbers
#				specified by a list of selected frames,
#				and returns them in a block-segregated form.
#
sub extract_bl_selected_frames (\@\@\@$) {

	my ($rf_align, $rf_resnums, $rf_sel_frames, $thrsh_distcolumn) = @_;

	(@{$rf_align} and @{$rf_resnums} ) or (return (undef, undef, undef) );
	@{$rf_sel_frames} or (return ([],[],[]) );
	my $last_sel_frame = $rf_sel_frames->[$#{$rf_sel_frames}];
	 (@{$rf_resnums} -1 < $last_sel_frame ) and (return (undef, undef, undef) );
	( length(${$rf_align->[0]})  -1 < $last_sel_frame) and (return (undef, undef, undef) );

	my @sel_resnums = ();
	foreach my $sel_frame (@{$rf_sel_frames}) { push @sel_resnums , $rf_resnums->[$sel_frame]; }

	my $rf_sel_align = diff_malign (@{$rf_align}, @{$rf_sel_frames});

	my ($bl_sel_aln, $bl_sel_resnums, $bl_sel_frames)
		= construct_blocks (@{$rf_sel_align}, @sel_resnums, @{$rf_sel_frames}, $thrsh_distcolumn);

	return ($bl_sel_aln, $bl_sel_resnums, $bl_sel_frames);

}			# end of the subroutine extract_bl_selected_frames #


# start_search: This subroutine looks for the start coordinate of a block specified,
#		in terms of the coordinate of a specified sequence.
#
sub start_search (\@$$$) {

	my ($rf_resnums, $seq_number, $frame_start, $frame_end) = @_;

	my $ct_columns = @{$rf_resnums};
	unless ($frame_end < $ct_columns)  {

		print STDERR "Illegal block end coordinate ($frame_end),\n";
		print STDERR "which is greater than the last columns index (" . $ct_columns -1 . "\n";
		return undef;
	}

	my $start = '-';
	for (my $i =$frame_start; $i<=$frame_end; $i++) {

		$start = $rf_resnums->[$i][$seq_number];
		(defined $start) or next;
		($start eq '-') or (return $start);
	}

		# if the sequence in the block consists only of gaps #
	foreach (my $i=$frame_start-1; $i > -1 ; $i--) {

		$start = $rf_resnums->[$i][$seq_number];
		(defined $start) or next;
		($start eq '-') or (return $start);
	}

		# if the sequences preceeding the block also consist only of gaps #
	foreach (my $i=$frame_end +1; $i < $ct_columns; $i++) {

		$start = $rf_resnums->[$i][$seq_number];
		(defined $start) or next;
		unless ($start eq '-') {

			$start--;
			last;
		}
	}

	return $start;
}					# end of the subroutine start_search #


# end_search: This subroutine looks for the end coordinate of a block specified,
#		in terms of the coordinate of a specified sequence.
#
sub end_search (\@$$$) {

	my ($rf_resnums, $seq_number, $frame_start, $frame_end) = @_;

	my $ct_columns = @{$rf_resnums};
	unless ($frame_end < $ct_columns)  {

		print STDERR "Illegal block end coordinate ($frame_end),\n";
		print STDERR "which is greater than the last columns index (" . $ct_columns -1 . "\n";
		return undef;
	}

	my $end = '-';
	for (my $i =$frame_end; $i>= $frame_start; $i--) {

		$end = $rf_resnums->[$i][$seq_number];
		(defined $end) or next;
		($end eq '-') or (return $end);
	}

		# if the sequence in the block consists only of gaps #
	foreach (my $i=$frame_start-1; $i > -1 ; $i--) {

		$end = $rf_resnums->[$i][$seq_number];
		(defined $end) or next;
		($end eq '-') or (return $end);
	}

		# if the sequences preceeding the block also consist only of gaps #
	foreach (my $i=$frame_end +1; $i < $ct_columns; $i++) {

		$end = $rf_resnums->[$i][$seq_number];
		(defined $end) or next;
		unless ($end eq '-') {

			$end--;
			last;
		}
	}

	return $end;
}					# end of the subroutine end_search #


# block_starts_ends: This subroutine constructs 2 arrays,
#		     one consists of positions where each block starts in terms of each sequence coordinates,
#		     and the other of those where each block ends.
#
# CAUTION: The 0-th argument "$rf_resnums" refers to a list of residue numbers
#		FOR THE FULL ALIGNMENT (not the different part only)!!
# 
sub block_starts_ends (\@\@) {

	my ($rf_resnums, $rf_block_frames) = @_;

	(@{$rf_resnums} and (@{$rf_block_frames}) )  or (return ([], []) );
	my @starts = my @ends = ();
	my $seq_ct = @{$rf_resnums->[0]};

	foreach my $block_frames (@{$rf_block_frames}) { # the outer foreach #

		my $frame_start = $block_frames->[0];
		my $frame_end = $block_frames->[$#{$block_frames}];
		my @block_starts = my @block_ends = ();

		for (my $i = 0; $i<$seq_ct; $i++) {	# the inner for #

			my $start = $rf_resnums->[$frame_start]->[$i];
			my $end = $rf_resnums->[$frame_end]->[$i];

			($start eq '-') and ($start = start_search (@{$rf_resnums}, $i, $frame_start, $frame_end) );
			(defined $start) or (return (undef, undef) );
			($end eq '-') and ($end = end_search (@{$rf_resnums}, $i, $frame_start, $frame_end) );
			(defined $end) or (return (undef, undef) );

			push @block_starts, $start;
			push @block_ends , $end;
		}				# end of the inner for #

		push @starts, \@block_starts;
		push @ends, \@block_ends;
	}					# end of the outer foreach #

	return (\@starts, \@ends);

}				# end of the subroutine block_starts_ends #


# monosite_picker: This subroutine makes a list of
#		positions in the alignment where the column
#		consists only of residues of one kind.
#		This subroutine adopts the (START = 0) convention.
#
sub monosite_picker (\@) {

	my ($rf_seqs) = @_;

	my @list_mono = ();
	my $ct_seqs = @{$rf_seqs};
	($ct_seqs) or (return \@list_mono ) ;

	my @copy = ();
	foreach my $seq (@{$rf_seqs}) { push @copy, ${$seq}; }
	my $seq_len = length (${$rf_seqs->[0]});
	my $first_seq = shift @copy;

	OUTER:
	#foreach my $l  (1 .. $seq_len) {
	foreach my $l  (0 .. $seq_len -1) {

		my $ct_id = 0;
		my $first_char = substr($first_seq, 0, 1, '');
		(defined $first_char) or last;
		$first_char = uc $first_char;
		($first_char =~ m/[ACGTU]/ ) and ($ct_id++);

		for (my $i=0; $i<@copy; $i++) {
			my $char = substr ($copy[$i], 0, 1, '');
			(defined $char) or (last OUTER);  # I regard the sequence has already ended. #
			(uc($char) eq $first_char) and $ct_id++;
		}

		($ct_id == $ct_seqs) and (push @list_mono, $l);
	}

	return \@list_mono;
}				# end of the subroutine monosite_picker #


# id_indicator: This subroutine makes a line which indicates
#		the positions of columns in the alignment
#		consisting only of one kind of nucleotides.
#
sub id_indicator (\@) {

	my ($rf_seqs) = @_;

	my $id_ind = '';
	my $ct_seqs = @{$rf_seqs};
	($ct_seqs) or (return \$id_ind ) ;

	my @copy = ();
	foreach my $seq (@{$rf_seqs}) { push @copy, uc ${$seq}; }
	for (my $i=0; $i < $ct_seqs; $i++) { $copy[$i] =~ tr/U/T/ ; }
	my $seq_len = length (${$rf_seqs->[0]});
	my $first_seq = shift @copy;

	foreach my $l  (1 .. $seq_len) {

		my $first_char = substr($first_seq, 0, 1, '');
		#my $ct_id = ( ($first_char eq '-') or ($first_char eq 'N') ) ? 0 : 1;
		my $ct_id = ($first_char =~ tr/ATGC// );

		for (my $i=0; $i<@copy; $i++) {
			my $char = substr ($copy[$i], 0, 1, '');
			($char eq $first_char) and $ct_id++;
		}

		$id_ind .= ($ct_id == $ct_seqs) ? '*' : ' ';
	}

	return \$id_ind;
}				# end of the subroutine id_indicator #


# id_indicator_aa: This subroutine makes a line which indicates
#		the positions of columns in the alignment
#		consisting only of one kind of amino acids.
#
sub id_indicator_aa (\@) {

	my ($rf_seqs) = @_;

	my $id_ind = '';
	my $ct_seqs = @{$rf_seqs};
	($ct_seqs) or (return \$id_ind ) ;

	my @copy = ();
	foreach my $seq (@{$rf_seqs}) { push @copy, uc ${$seq}; }
	my $seq_len = length (${$rf_seqs->[0]});
	my $first_seq = shift @copy;

	foreach my $l  (1 .. $seq_len) {

		my $first_char = substr($first_seq, 0, 1, '');
		#my $ct_id = ( ($first_char eq '-') or ($first_char eq 'N') ) ? 0 : 1;
		my $ct_id = ($first_char =~ tr/ACDEFGHIKLMNPQRSTUVWY// );
			# NOTE: U = 'selenocystein'

		for (my $i=0; $i<@copy; $i++) {
			my $char = substr ($copy[$i], 0, 1, '');
			($char eq $first_char) and $ct_id++;
		}

		$id_ind .= ($ct_id == $ct_seqs) ? '*' : ' ';
	}

	return \$id_ind;
}				# end of the subroutine id_indicator_aa #


# print_block_maln: This subroutine prints out a  block-separated multiple- alignment.
#			It returns the count of completely identical columns in the alignment. 
#
sub print_block_maln (\@\@\@\@\*$) {

	my ($rf_bl_columns, $rf_bl_frames, $rf_resnums, $rf_seqIDs, $fh, $offset) = @_;

	my $ct_blocks = @{$rf_bl_columns};
	my $ct_seqs = @{$rf_seqIDs};
	unless ($ct_blocks) { print $fh "\n"; return 0; }
 	unless (@{$rf_bl_frames} == $ct_blocks) {

		print STDERR "Illegal input for \"print_block_maln\".\n";
		return undef;
	}

	my ($rf_bl_starts, $rf_bl_ends) = block_starts_ends (@{$rf_resnums}, @{$rf_bl_frames});

	unless (defined $rf_bl_starts) {

		print STDERR "Subroutine \"block_starts_ends\" failed.\n";
		return undef;
	}

	my @labels = ();
	foreach my $seqID (@{$rf_seqIDs}) {

		my $label = $seqID . ' ' x (15 - length($seqID) );
		push @labels, $label;
	}

	my $id_ct = 0;

	for (my $i=0; $i < $ct_blocks; $i++) { # the outer for #

	    my $frame = $rf_bl_frames->[$i];
	    my $ct_columns = @{$frame};
	    my $frame_start = $frame->[0] + 1;
	    my $frame_end = $frame->[$#{$frame}] + 1;
	    printf $fh "Block_%d (frame:%7d  --%7d, %7d columns)\n", $i+1+ $offset, $frame_start, $frame_end, $ct_columns;

	    my @reservoir = ();
	    for (my $j=0; $j < $ct_seqs; $j++) { # the inner for #
		my $start = $rf_bl_starts->[$i][$j];
		my $end = $rf_bl_ends->[$i][$j];
		unless (defined $start) { 
			print STDERR "The starting position of block $i, sequence $rf_seqIDs->[$j] is NOT defined.\n";
			return undef;
		}
		unless (defined $end) { 
			print STDERR "The ending position of block $i, sequence $rf_seqIDs->[$j] is NOT defined.\n";
			return undef;
		}
		$start = ($start eq '-') ? (' ' x 4 . '-' x 3) : sprintf ("%7d", $start);
		$end = ($end eq '-') ? (' ' x 4 . '-' x 3) : sprintf ("%7d", $end);
		my $seq = '';
		##foreach my $column (@{$rf_bl_columns->[$i]}) { $seq .= $column->[$j]; }
		for (my $k=0; $k<@{$rf_bl_columns->[$i]}; $k++) {

			unless (defined $rf_bl_columns->[$i][$k]) {

				print STDERR "The column $k of block $i is NOT defined.\n";
				return undef;
			}
			unless (defined $rf_bl_columns->[$i][$k][$j]) {

				print STDERR "The sequence $rf_seqIDs->[$j] is NOT defined at the column $k of block $i.\n";
				return undef;
			}
			$seq .= $rf_bl_columns->[$i][$k][$j];
		}

		my $line = $labels[$j] . $start . '  ' . $seq . $end . "\n";
		print $fh $line;
		push @reservoir, \$seq;

	    }				# end of the inner for #

	    my $id_ind = id_indicator(@reservoir);
	    $id_ct +=  (${$id_ind} =~ tr/\*//);

	    print $fh ' ' x 24 . ${$id_ind} . "\n\n";
	}				# end of the outer for #

	return $id_ct;
}				# end of the subroutine print_block_maln #


# select_equiv_cand: This subroutine looks for candidates for equivalent alignments,
#			and returns 1 if it finds out candidates.
#
sub select_equiv_cand (\@\@\@\@) {

	my ($bl_fore, $bl_hind, $starts_hind, $ends_fore) = @_;

	my $ct_columns = @{$bl_fore};
	($ct_columns == @{$bl_hind}) or (return 0);	# when the 2 blocks are of different sizes #
	($ct_columns) and (return 1);			# when the 2 blocks are both null #

	my $ct_seqs = @{$bl_fore->[0]};
	($ct_seqs == @{$bl_hind->[0]}) or (return 0);
	my $fl_cand =1;

	for (my $i = 0; $i<$ct_seqs; $i++) {

		my $ct_equiv = my $ct_foregaps = my $ct_hindgaps = 0;
		for (my $l = 0; $l<$ct_columns; $l++) {

			($bl_fore->[$l][$i] eq $bl_hind->[$l][$i]) and ($ct_equiv++);
			($bl_fore->[$l][$i] eq '-') and ($ct_foregaps++);
			($bl_hind->[$l][$i] eq '-') and ($ct_hindgaps++);
		}

		($ct_equiv == $ct_columns) and next;
		($ct_foregaps == $ct_columns) and ($ends_fore->[$i] == $starts_hind->[$i] -1) and next;
		($ct_hindgaps == $ct_columns) and ($ends_fore->[$i] == $starts_hind->[$i]) and next;

		$fl_cand = 0;
		last;
	}

	return $fl_cand;
}				# end of the subroutine select_equiv_cand #


# equiv_columns: This subroutine compares 2 blocks of columns,
#		and returns 1 if they are equivalnet and 0 otherwise.
#
sub equiv_columns (\@\@) {

	my ($bl_columns_1, $bl_columns_2) = @_;

	my $ct_columns = @{$bl_columns_1};
	(@{$bl_columns_2} == $ct_columns) or (return 0);
	($ct_columns) or (return 1);			# the two null blocks are regarded as equivalent #
	my $ct_seqs = @{$bl_columns_1->[0]};
	(@{$bl_columns_2->[0]} == $ct_seqs) or (return 0);

	my $fl_equiv = 1;

	OUTER:
	for (my $i=0; $i<$ct_columns; $i++) { # the outer for #
		for (my $l=0; $l<$ct_seqs; $l++) { # the inner for #

			($bl_columns_1->[$i][$l] eq $bl_columns_2->[$i][$l]) and next;

			$fl_equiv = 0;
			last OUTER;
		}			# end of the inner for #
	}				# end of the outer for #

	return $fl_equiv;
}				# end of the subroutine equiv_columns #


# trim_diff_maln: This subroutine recieves differences in 2 multiple alignments,
#		removes those which give different interpretations of the equivalent alignment,
#		and returns the resulting tidied-up differences and the removed parts.
#
sub trim_diff_maln (\@\@\@\@\@\@\@\@) {

	my ($bl_aln_1, $bl_resnums_1, $bl_frames_1, $resnums_1, $bl_aln_2, $bl_resnums_2, $bl_frames_2, $resnums_2) = @_;

	my $ct_bl_1 = @{$bl_aln_1};
	my $ct_bl_2 = @{$bl_aln_2};
	($ct_bl_1 and $ct_bl_2 ) or ( return ([],[],[],[],[],[],[],[],[],[],[],[]) );

	my ($bl_starts_1, $bl_ends_1) = block_starts_ends (@{$resnums_1}, @{$bl_frames_1});
	my ($bl_starts_2, $bl_ends_2) = block_starts_ends (@{$resnums_2}, @{$bl_frames_2});

	my @trbl_aln_1 = my @trbl_resnums_1 = my @trbl_frames_1 = ();
	my @trbl_aln_2 = my @trbl_resnums_2 = my @trbl_frames_2 = ();
	my @rmd_aln_1 = my @rmd_resnums_1 = my @rmd_frames_1 = ();
	my @rmd_aln_2 = my @rmd_resnums_2 = my @rmd_frames_2 = ();

	my $last_indx_1 = my $last_indx_2 = 0;

	OUTER:
	while ( ($last_indx_1 < $ct_bl_1 -1 ) and ($last_indx_2 < $ct_bl_2 -1) ) {

		until ( select_equiv_cand (@{$bl_aln_1->[$last_indx_1]}, @{$bl_aln_1->[$last_indx_1 + 1]}, @{$bl_starts_1->[$last_indx_1 +1]}, @{$bl_ends_1->[$last_indx_1]} )  ) {

			push @trbl_aln_1, $bl_aln_1->[$last_indx_1];
			push @trbl_resnums_1, $bl_resnums_1->[$last_indx_1];
			push @trbl_frames_1, $bl_frames_1->[$last_indx_1];
			$last_indx_1++;
			($last_indx_1 < $ct_bl_1 -1) or  (last OUTER);
		}

		until (select_equiv_cand (@{$bl_aln_2->[$last_indx_2]}, @{$bl_aln_2->[$last_indx_2 +1]}, @{$bl_starts_2->[$last_indx_2 +1]}, @{$bl_ends_2->[$last_indx_2]})  ) {

			push @trbl_aln_2, $bl_aln_2->[$last_indx_2];
			push @trbl_resnums_2, $bl_resnums_2->[$last_indx_2];
			push @trbl_frames_2, $bl_frames_2->[$last_indx_2];
			$last_indx_2++;
			($last_indx_2 < $ct_bl_2 -1) or  (last OUTER);
		}					# end of the until-loop #

		if ( equiv_columns (@{$bl_aln_2->[$last_indx_2]}, @{$bl_aln_1->[$last_indx_1 + 1]}) and equiv_columns (@{$bl_aln_1->[$last_indx_1]}, @{$bl_aln_2->[$last_indx_2 + 1]})  ){

			push @rmd_aln_1, ($bl_aln_1->[$last_indx_1], $bl_aln_1->[$last_indx_1 +1]);
			push @rmd_resnums_1, ($bl_resnums_1->[$last_indx_1], $bl_resnums_1->[$last_indx_1 +1]);
			push @rmd_frames_1, ($bl_frames_1->[$last_indx_1], $bl_frames_1->[$last_indx_1 +1]);
			
			push @rmd_aln_2, ($bl_aln_2->[$last_indx_2], $bl_aln_2->[$last_indx_2 +1]);
			push @rmd_resnums_2, ($bl_resnums_2->[$last_indx_2], $bl_resnums_2->[$last_indx_2 +1]);
			push @rmd_frames_2, ($bl_frames_2->[$last_indx_2], $bl_frames_2->[$last_indx_2 +1]);

		} else {

			push @trbl_aln_1, ($bl_aln_1->[$last_indx_1], $bl_aln_1->[$last_indx_1 +1]);
			push @trbl_resnums_1, ($bl_resnums_1->[$last_indx_1], $bl_resnums_1->[$last_indx_1 +1]);
			push @trbl_frames_1, ($bl_frames_1->[$last_indx_1], $bl_frames_1->[$last_indx_1 +1]);
			
			push @trbl_aln_2, ($bl_aln_2->[$last_indx_2], $bl_aln_2->[$last_indx_2 +1]);
			push @trbl_resnums_2, ($bl_resnums_2->[$last_indx_2], $bl_resnums_2->[$last_indx_2 +1]);
			push @trbl_frames_2, ($bl_frames_2->[$last_indx_2], $bl_frames_2->[$last_indx_2 +1]);

		}

		$last_indx_1 += 2;
		$last_indx_2 += 2;
	}					# end of the outer while #

	if ($last_indx_1 < $ct_bl_1) {

		push @trbl_aln_1, @{$bl_aln_1}[$last_indx_1 .. $ct_bl_1 -1 ];
		push @trbl_resnums_1, @{$bl_resnums_1}[$last_indx_1 .. $ct_bl_1 -1];
		push @trbl_frames_1, @{$bl_frames_1}[$last_indx_1 .. $ct_bl_1 - 1]; 
	}

	if ($last_indx_2 < $ct_bl_2) {

		push @trbl_aln_2, @{$bl_aln_2}[$last_indx_2 .. $ct_bl_2 -1 ];
		push @trbl_resnums_2, @{$bl_resnums_2}[$last_indx_2 .. $ct_bl_2 -1];
		push @trbl_frames_2, @{$bl_frames_2}[$last_indx_2 .. $ct_bl_2 - 1]; 
	}

	return (\@trbl_aln_1, \@trbl_resnums_1, \@trbl_frames_1, \@trbl_aln_2, \@trbl_resnums_2, \@trbl_frames_2, \@rmd_aln_1, \@rmd_resnums_1, \@rmd_frames_1, \@rmd_aln_2, \@rmd_resnums_2, \@rmd_frames_2);

}			# end of the subroutine trim_diff_maln #


# maln_identities: This subroutine receives a list of columns in a multiple alignment,
#		  counts the number of matched pairs for each column,
#		  and returns a list of numbers thus counted.
#
sub maln_identities (\@) {

	my ($rf_columns) = @_;

	my $ct_columns = @{$rf_columns};
	($ct_columns) or ( return ([], []) );

	my @column_id_cts = ();
	my $tot_id_ct = 0;
	my $mono_ct = 0;	# the count of monomorphic sites (columns) #	
	my $tot_ct_pairs = 0;

	my $ct_seqs = @{$rf_columns->[0]};
	my $max_ct_pairs = $ct_seqs * ($ct_seqs -1) / 2;
	my $ct_nogap_clms = 0;

	foreach my $column (@{$rf_columns}) {	# the outer foreach #

		my %ct_residues;	# key = residue identity, value = its occurrence in a column #
		my $ct_nogaps = 0;
		foreach my $char (@{$column}) {

			( ($char eq '-') or ($char eq 'N') ) and next;
			(defined $ct_residues{$char}) ? ($ct_residues{$char}++) : ($ct_residues{$char} = 1);
			$ct_nogaps++;
		}

		($ct_nogaps == $ct_seqs) and  ($ct_nogap_clms++);
		my $ct_pairs = $ct_nogaps * ($ct_nogaps -1) /2;
		$tot_ct_pairs += $ct_pairs;

		my $id_ct = 0;
		foreach my $char (keys %ct_residues) {

			my $ct = $ct_residues{$char};
			($ct == 1) or ($id_ct += $ct* ($ct -1) /2 );
		}

		$tot_id_ct += $id_ct;
		push @column_id_cts, $id_ct;
		($id_ct  == $max_ct_pairs) and ($mono_ct++);
	}					# end of the outer foreach #

	#my $av_id_ct = $tot_id_ct/($ct_pairs*$ct_columns);
	my $av_id_ct = ($tot_ct_pairs) ? $tot_id_ct/$tot_ct_pairs : 0;
	my $av_mono_ct = ($ct_nogap_clms) ? $mono_ct/ $ct_nogap_clms : 0;
	my @column_id_stat = ($tot_id_ct, $mono_ct, $tot_ct_pairs, $ct_nogap_clms, $av_id_ct, $av_mono_ct);
	return (\@column_id_cts, \@column_id_stat);

}			# end of the subroutine maln_identities #

# block_maln_identities: This subroutine receives a list of blocks of columns,
#			do similar things with the subroutine 'maln_identities',
#			and returns a list of numbers distributed to the blocks.
#
sub block_maln_identities (\@) {

	my ($rf_blocks) = @_;

	(@{$rf_blocks}) or ( return ([],[]) );

	my @block_id_cts = my @bl_id_stats = ();
	foreach my $block (@{$rf_blocks}) { 

		my ($column_id_cts, $column_id_stat)  = maln_identities (@{$block});
		push @block_id_cts, $column_id_cts;
		push @bl_id_stats, $column_id_stat;
	}

	return (\@block_id_cts, \@bl_id_stats) ;

}			# end of the subroutine block_maln_identities #


# maln_id_distributon This subroutine receives a list of columns in a multiple alignment,
#		  counts the number of matched pairs for each column,
#		  and returns a list of numbers thus counted, PLUS a list of numbers of
#		  residue pairs each for a column.
#
sub maln_id_distribution (\@) {

	my ($rf_columns) = @_;

	my $ct_columns = @{$rf_columns};
	($ct_columns) or ( return ([], []) );

	my @column_id_cts = ();
	my @column_pair_cts = ();
	#my @column_mono_distri = ();

	my $ct_seqs = @{$rf_columns->[0]};
	my $max_ct_pairs = $ct_seqs * ($ct_seqs -1) / 2;

	foreach my $column (@{$rf_columns}) {	# the outer foreach #

		my %ct_residues;	# key = residue identity, value = its occurrence in a column #
		my $ct_nogaps = 0;
		foreach my $char (@{$column}) {

			( ($char eq '-') or ($char eq 'N') ) and next;
			(defined $ct_residues{$char}) ? ($ct_residues{$char}++) : ($ct_residues{$char} = 1);
			$ct_nogaps++;
		}

		my $id_ct = 0;
		foreach my $char (keys %ct_residues) {

			my $ct = $ct_residues{$char};
			($ct == 1) or ($id_ct += $ct* ($ct -1) /2 );
		}

		push @column_id_cts, $id_ct;
		my $ct_pairs = $ct_nogaps * ($ct_nogaps -1) /2;
		push @column_pair_cts, $ct_pairs;
		#push @column_mono_distri, ( ($id_ct == $max_ct_pairs) ? 1 : 0 );
	}					# end of the outer foreach #

	return (\@column_id_cts, \@column_pair_cts);

}			# end of the subroutine maln_id_distribution #


# block_maln_id_distribution: This subroutine receives a list of blocks of columns,
#			do similar things with the subroutine 'maln_id_distribution',
#			and returns a list of numbers each for each block.
#
sub block_maln_id_distribution (\@) {

	my ($rf_bl_columns) = @_;

	(@{$rf_bl_columns}) or ( return ([],[]) );

	my @block_id_distri = my @bl_pair_distri = ();
	foreach my $block (@{$rf_bl_columns}) { 

		my ($column_id_distri, $column_pair_distri)  = maln_id_distribution (@{$block});
		push @block_id_distri, $column_id_distri;
		push @bl_pair_distri, $column_pair_distri;
		#push @bl_mono_distri, $column_mono_distri;
	}

	return (\@block_id_distri, \@bl_pair_distri) ;

}			# end of the subroutine block_maln_id_distribution #

# id_sliding_window: This subroutine calculates distribution of identities
#			within a sliding-window set at each position.
#
sub id_sliding_window (\@\@\@$$$) {

	my ($pair_cts, $matched_cts, $frames, $ct_seqs, $window_size, $step_size) = @_;

	unless ( ($step_size > 0) and ($window_size > 0) ) {

		print STDERR "Illegal parameters for \'id_sliding_window\' \!\n";
		return (undef, undef);
	}
	unless ($ct_seqs) {

		print STDERR "There are NO SEQUENCES to be processed with \'id_sliding_window\' \!\n";
		return (undef, undef);
	}

	my $column_ct = @{$frames};
	$column_ct or (return ([], []) );

	my @matched_propts = my @mono_column_propts = my @wd_starts = my @wd_ends  = ();
	my $max_pair_ct = $ct_seqs * ($ct_seqs -1) /2;

	my $start_frame = $frames->[0];
	my $end_frame = $frames->[$column_ct -1];
	my $upper_bound = ($end_frame >=  $start_frame + $window_size) ? $end_frame - $window_size +2 : $start_frame +1 ;
	my $base_point = 0;

	while ( ($base_point < $column_ct) and ($frames->[$base_point] < $upper_bound) ) {
					 # the outermost while #

		my $frame_ct = 0;
		#my $av_frame = 0;
		my $mono_ct = 0;
		my $pair_sum = 0;
		my $matched_sum = 0;
		my $wd_start = $frames->[$base_point];
		push @wd_starts, $wd_start;
		my $wd_end = $wd_start;

		my $wd_bound = $wd_start + $window_size;
		for (my $i= $base_point; ($i<$column_ct) and ($frames->[$i] < $wd_bound); $i++) {
							 # the middle for #

			$frame_ct++;
			#$av_frame += $frames->[$i];
			$pair_sum += $pair_cts->[$i];
			$matched_sum += $matched_cts->[$i];
			($matched_cts->[$i] == $max_pair_ct) and ($mono_ct++);
			$wd_end = $frames->[$i];

		}					# end of the middle for #

		push @wd_ends, $wd_end;
		#$av_frame /= $frame_ct;
		#push @av_frames, $av_frame;
		push @matched_propts, ($pair_sum ? $matched_sum/$pair_sum : 0);
		push @mono_column_propts, $mono_ct/$frame_ct;

		my $wd_next_begin = $frames->[$base_point] + $step_size;
		while ( ($base_point < $column_ct) and ($frames->[$base_point] < $wd_next_begin) ) {
			 $base_point++;
		}

	}		# end of the outermost while #

	return (\@matched_propts, \@mono_column_propts, \@wd_starts, \@wd_ends);

}			# end of the subroutine id_sliding_window #


# bl_id_sliding_window: This subroutine conducts sliding-window analysis
#			on the identity distribution for each of blocks.
#
sub bl_id_sliding_window (\@\@$$) {

	my ($rf_bl_columns, $rf_bl_frames, $window_size, $step_size) = @_;

	my $bl_ct = @{$rf_bl_columns};
	$bl_ct or (return ([],[]) );

	my $ct_seqs = @{$rf_bl_columns->[0][0]};

	my ($bl_id_distri, $bl_pair_distri) = block_maln_id_distribution (@{$rf_bl_columns});
	my @matched_propts = my @mono_column_propts = my @wd_starts = my @wd_ends  = ();

	for (my $i=0; $i< $bl_ct; $i++) {

		my ($bl_mtchd_propts, $bl_monoclm_propts, $bl_wd_starts, $bl_wd_ends)
		= id_sliding_window (@{$bl_pair_distri->[$i]}, @{$bl_id_distri->[$i]}, @{$rf_bl_frames->[$i]}, $ct_seqs, $window_size, $step_size);

		push @matched_propts, $bl_mtchd_propts;
		push @mono_column_propts, $bl_monoclm_propts;
		push @wd_starts, $bl_wd_starts;
		push @wd_ends, $bl_wd_ends;
	}		

	return (\@matched_propts, \@mono_column_propts, \@wd_starts, \@wd_ends, $bl_id_distri, $bl_pair_distri);

} 			# end of the subroutine bl_id_sliding_window #

# average_id_proportions: This subroutine calculates average proportions
#			 of matched pairs and monomorphic sites
#			for the entire alignment and for each blocks contained in it,
#			 out of the distribution of matched pairs in the columns
#			 (collected from each block).
#
sub average_id_proportions (\@\@$) {

	my ($bl_mtchd_distri, $bl_pair_distri, $seq_ct) = @_;

	my $ct_bl = @{$bl_mtchd_distri} ;
	$ct_bl or (return ([],[], 0, 0,0) );
	my $max_pair_ct = $seq_ct * ($seq_ct -1) / 2;

	my @av_matched_pr = my @av_mono_pr = ();
	my $ent_av_matched = my $ent_av_mono = my $tot_pairs = my $tot_clms = 0;
	# $tot_clms: The count of columns containing no gaps. #

	for (my $i = 0; $i<$ct_bl; $i++) {

		my $av_matched = my $av_mono = my $subtot_pairs = my $subtot_clms = 0;
		for (my $l=0; $l<@{$bl_pair_distri->[$i]}; $l++) {

			my $ct_pairs = $bl_pair_distri->[$i][$l];
			$subtot_pairs += $ct_pairs;
			($ct_pairs == $max_pair_ct) and ($subtot_clms++);
			my $ct_matched = $bl_mtchd_distri->[$i][$l];
			$av_matched += $ct_matched;
			($ct_matched == $max_pair_ct) and ($av_mono++);
			
		}

		$ent_av_matched += $av_matched;
		$ent_av_mono += $av_mono;
		$tot_pairs += $subtot_pairs;
		$tot_clms += $subtot_clms;

		$av_matched = $subtot_pairs ? $av_matched/$subtot_pairs : 0;
		$av_mono = $subtot_clms ? $av_mono/$subtot_clms : 0;

		push @av_matched_pr, $av_matched;
		push @av_mono_pr, $av_mono;
	}

	$ent_av_matched = $tot_pairs ? $ent_av_matched/$tot_pairs: 0;
	$ent_av_mono = $tot_clms ? $ent_av_mono/$tot_clms : 0;

	return (\@av_matched_pr, \@av_mono_pr, $ent_av_matched, $ent_av_mono, $tot_clms);

}			# end of the subroutine average_id_proportions #

# pick_poor_windows: This subroutine selects windows whose proportions of matched pairs and/or
#		    that of monomorphic columns are below the specified level.
#
sub pick_poor_windows (\@\@\@\@$$) {

	my ($matched_propts, $mono_clm_propts, $wd_starts, $wd_ends, $thrsh_matched, $thrsh_mono) = @_;

	my $wd_ct = @{$matched_propts};
	$wd_ct or ( return ([],[],[],[]) );

	my @poor_matched_propts = my @poor_mono_propts = my @poor_starts = my @poor_ends = ();

	for (my $i=0; $i<$wd_ct; $i++) {

		my $mt_propt = $matched_propts->[$i];
		my $mono_propt = $mono_clm_propts->[$i];

			# windows which satisfies 2 conditions are regarded as fine. #
		($mt_propt >= $thrsh_matched) and ($mono_propt >= $thrsh_mono) and next;

		push @poor_matched_propts, $mt_propt;
		push @poor_mono_propts, $mono_propt;
		push @poor_starts, $wd_starts->[$i];
		push @poor_ends, $wd_ends->[$i];
	}

	return (\@poor_matched_propts, \@poor_mono_propts, \@poor_starts, \@poor_ends);

}				# end of the subroutine pick_poor_windows #

# pick_bl_poor_windows: This subroutine selects windows which do not satisfy the specified
#			conditions on the matched pair and monomorphic column proportions,
#			block by block.
#
sub pick_bl_poor_windows (\@\@\@\@$$) {

	my ($bl_mtchd_propts, $bl_mono_propts, $bl_wd_starts, $bl_wd_ends, $thrsh_matched, $thrsh_mono) = @_;

	my $bl_ct = @{$bl_mtchd_propts};
	$bl_ct or (return ([],[],[],[],0) );

	my @poor_mtchd_propts = my @poor_mono_propts = my @poor_starts = my @poor_ends = ();
	my $ct_poor_wds = 0;

	for (my $i=0; $i<$bl_ct; $i++) {

		my @output = pick_poor_windows (@{$bl_mtchd_propts->[$i]}, @{$bl_mono_propts->[$i]},
			@{$bl_wd_starts->[$i]}, @{$bl_wd_ends->[$i]}, $thrsh_matched, $thrsh_mono);

		$ct_poor_wds += scalar @{$output[0]};

		push @poor_mtchd_propts, $output[0];
		push @poor_mono_propts, $output[1];
		push @poor_starts, $output[2];
		push @poor_ends, $output[3];
	}

	return (\@poor_mtchd_propts, \@poor_mono_propts, \@poor_starts, \@poor_ends, $ct_poor_wds);

}				# end of the subroutine pick_bl_poor_windows #

# mask_frames : This subroutine receives the sequence part of a multiple-alignment and a list of frames,
#		and returns sequences  with the specified frames masked.
#
sub mask_frames (\@\@) {

	my ($rf_align, $rf_frames) = @_;

	my $ct_seq = @{$rf_align};
	$ct_seq or (return [] );

		# a part of the fatal-bug fixation (on Feb 18 (Wed), 2004 ) #
	#unless (@{$rf_frames}) {
	#	my @tmp =  @{$rf_align};
	#	return \@tmp;
	#}

	my @out_align = my @copy = ();
	foreach my $align (@{$rf_align}) {
		my $tmp = '';
		push @out_align , \$tmp;
		#push @copy, ${$align};
		my $seqcp = ${$align};
		push @copy, \$seqcp;	# a bug fixation on March 20 (Sat), 2004 #
					# accompanied by the change from $copy[$i]  to ${$copy[$i]} below #
	}
		# the rest of the fatal-bug fixation (on Feb 18 (Wed), 2004 ) #
	(@{$rf_frames}) or (return \@copy);
	
	my $len_align = length ${$copy[0]};

	my $expected_frame = 0;
	foreach my $frame (@{$rf_frames}) {

		my $len_unmasked = $frame - $expected_frame;
		if ($len_unmasked > 0) {
			for (my $i=0; $i<$ct_seq; $i++) {
				${$out_align[$i]} .= substr ( ${$copy[$i]}, 0, $len_unmasked, '');
			}
				
		} elsif ($len_unmasked < 0) {
			print STDERR "negative unmasked length (expected_frame = $expected_frame, real_frame = $frame).\n";
			return undef;
		}

		for (my $i=0; $i<$ct_seq; $i++) {
			my $char = substr (${$copy[$i]}, 0, 1, '');
			${$out_align[$i]} .= ($char eq '-') ? '-' : 'N';
		}

		$expected_frame = $frame + 1;

	}				# end of the foreach #

	for (my $i=0; $i<$ct_seq; $i++) { ${$out_align[$i]} .= ${$copy[$i]}; }

	return \@out_align;

}				# end of the subroutine mask_frames #


# compl_mask_frames : This subroutine receives the sequence part of a multiple-alignment and a list of frames,
#		and returns sequences with the frames complementary to the specified ones  masked.
#
sub compl_mask_frames (\@\@) {

	my ($rf_align, $rf_frames) = @_;

	my $ct_seq = @{$rf_align};
	$ct_seq or (return [] );

		# fixing a fatal bug (on Feb 18 (Wed), 2004) #
	#unless (@{$rf_frames}) {		
	#	my @tmp =  @{$rf_align};
	#	return \@tmp;
	#}

	my @out_align = my @copy = ();
	foreach my $align (@{$rf_align}) {
		my $tmp = '';
		push @out_align , \$tmp;
		push @copy, ${$align};
	}

	my $len_align = length $copy[0];

	my $expected_frame = 0;
	foreach my $frame (@{$rf_frames}) {

		my $len_masked = $frame - $expected_frame;
		if ($len_masked > 0) {
			for (my $i=0; $i<$ct_seq; $i++) {
				my $tmp = substr ( $copy[$i], 0, $len_masked, '');
				$tmp =~ tr/a-zA-Z/N/;
				${$out_align[$i]} .= $tmp;
			}
				
		} elsif ($len_masked < 0) {
			print STDERR "negative masked length (expected_frame = $expected_frame, real_frame = $frame).\n";
			return undef;
		}

		for (my $i=0; $i<$ct_seq; $i++) {
			${$out_align[$i]} .= substr ($copy[$i], 0, 1, '');
		}

		$expected_frame = $frame + 1;

	}				# end of the foreach #

	for (my $i=0; $i<$ct_seq; $i++) { 

		$copy[$i] =~ tr/a-zA-Z/N/;
		${$out_align[$i]} .= $copy[$i];
	}

	return \@out_align;

}				# end of the subroutine compl_mask_frames #


# phase_column: This subroutine evaluates the phase of a column, using its residue numbers.
#
#		The subroutine returns undef if the column does not contain 2 or more nogap characters,
#		or the sequences are out of phase.
#
sub phase_column (\@) {

	my ($rf_resnums) = @_;

	@{$rf_resnums} or (return undef);

	my @resnums = @{$rf_resnums};
	my @reservoir = ();
	while (defined (my $num = shift @resnums)  ) {
		($num eq '-') or (push @reservoir, $num % 3);
	}
	(@reservoir < 2) and (return undef);

	my $first_phase = shift @reservoir;

	foreach my $phase (@reservoir) {

		($phase == $first_phase) or (return undef);
	}

	return $first_phase;
}			# end of the subroutine phase_column #


# select_in_phase: This subroutine receives lists of frames and  of residue numbers,
#		 then picks out a list of frames in which the columns are in-phase.
#		A column containing only one nogap character is regarded as 'in-phase'. 
#
# NOTE: The 1st codon corresponds to the "frame (= resnum % 3) = 1".
#
sub select_in_phase (\@\@) {

	my ($rf_resnums, $rf_frames) = @_;

	(@{$rf_frames} and @{$rf_resnums})  or (return [] );
	 my @sel_frames = ();

	foreach  my $frame (@{$rf_frames}) {

		my $phase  = phase_column (@{$rf_resnums->[$frame]}) ;
		(defined $phase)  and (push @sel_frames, $frame);

	}					# end of the foreach #

	return \@sel_frames;

}			# end of the subroutine select_in_phase #


# select_unamb_frames: This subroutine recieves an alignment and a subset of its frames,
#			chooses frames which consists solely of unambiguous bases,
#			and returns a list of those frames.
#
sub select_unamb_frames (\@\@) {

	my ($rf_align, $rf_selframe) = @_;

	(@{$rf_selframe}) or ( return []); 

	my @copy = my @ng_frames = ();
	foreach my $rf_seq (@{$rf_align}) {
		push @copy, ${$rf_seq};
	}

	my $exp_frame = 0;
	foreach my $frame (@{$rf_selframe}) {

		if ( (my $diff = $frame - $exp_frame) > 0)  {

			for (my $i=0; $i<@copy; $i++) { substr($copy[$i], 0, $diff, ''); }

		} elsif ($diff < 0) {

			print STDERR "Error in diff_malign:\n";
			print STDERR "expected frame ($exp_frame) exceeds the real one ($frame).\n";
			return undef;
		}

		my $fl_nogap = 1; 
		for (my $i=0; $i<@copy; $i++) {

			my $res = substr($copy[$i], 0, 1, '');
			($res =~ m/[acgtuACGTU]/ ) or ($fl_nogap = 0);
		}

		$fl_nogap and (push @ng_frames, $frame);
		$exp_frame = $frame + 1;
		
	}

	return \@ng_frames;

}			# end of the subroutine select_unamb_frames #


# sel_complete_codon_frames: This subroutine receives a subset of frames and a complete set of residue numbers,
#			and selects frames which, together with the adjacent frames,  form complete codons
#			in all of the sequences.
#
sub sel_complete_codon_frames (\@\@) {

	my ($rf_resnums, $rf_sb_frames) = @_;

	@{$rf_sb_frames} or (return []);
	(@{$rf_resnums} < @{$rf_sb_frames}) and (return undef);

	my $ct_seqs = @{$rf_resnums->[0]};
	my $upper_limit =  @{$rf_sb_frames} -2;
	my @sel_frames = ();

	my $l = 0;

	while ($l < $upper_limit) {

		my $frame1 = $rf_sb_frames->[$l];
		my $phase = phase_column (@{$rf_resnums->[$frame1]}) ;
		unless ( (defined $phase) and ($phase == 1) ) {	# skip untill we meet the phase '1' #
			$l++;
			next;
		}

		my $frame2 = $rf_sb_frames->[$l+1];
		my $frame3 = $rf_sb_frames->[$l+2];
		my $fl_comp_codon = 1;	
		for (my $i=0; $i<$ct_seqs; $i++) {

			my $resnum1 = $rf_resnums->[$frame1][$i];
			my $resnum2 = $rf_resnums->[$frame2][$i];
			my $resnum3 = $rf_resnums->[$frame3][$i];

				# a gap of the whole codon length or longer #
			( ($resnum1 eq '-') and ($resnum2 eq '-') and ($resnum3 eq '-') ) and next;

				# a complete codon #
			( ($resnum1 ne '-') and ($resnum2 ne '-') and ($resnum3 ne '-') ) 
			and (($resnum2 == $resnum1 +1) and ($resnum3 == $resnum2 +1) ) and next;

			$fl_comp_codon = 0;
			last;
		}					# end of the for-loop #

		if ($fl_comp_codon) {
			push @sel_frames , ($frame1, $frame2, $frame3);
			$l += 3;
		} else {
			$l++;
		}
	}						# end of the while-loop #

	return \@sel_frames;
}				# end of the subroutine sel_complete_codon_frames #


# sel_codon_frames_in_ref: This subroutine receives a subset of frames, a complete set of residue numbers,
#			and the index of a "reference" sequence,
#			and selects frames which, together with the adjacent frames,  form complete codons
#			in the reference sequence.
#
sub sel_codon_frames_in_ref (\@\@$) {

	my ($rf_resnums, $rf_sb_frames, $indx_ref) = @_;

	@{$rf_sb_frames} or (return []);
	(@{$rf_resnums} < @{$rf_sb_frames}) and (return undef);

	my $ct_seqs = @{$rf_resnums->[0]};
	my $upper_limit =  @{$rf_sb_frames} -2;
	my @sel_frames = ();

	my $l = 0;

	while ($l < $upper_limit) {

		my $frame1 = $rf_sb_frames->[$l];
		my $resnum1 = $rf_resnums->[$frame1][$indx_ref];
		#my $phase = phase_column (@{$rf_resnums->[$frame1]}) ;
		#unless ( (defined $phase) and ($phase == 1) ) {	# skip untill we meet the phase '1' #
		unless ( ($resnum1 ne '-') and ($resnum1 % 3 == 1) ) {	# skip untill we meet the phase '1' #
			$l++;
			next;
		}

		my $frame2 = $rf_sb_frames->[$l+1];
		my $frame3 = $rf_sb_frames->[$l+2];

		#my $resnum1 = $rf_resnums->[$frame1][$i];

		my $resnum2 = $rf_resnums->[$frame2][$indx_ref];
		my $resnum3 = $rf_resnums->[$frame3][$indx_ref];

				# a gap of the whole codon length or longer #
 			# the "next" is associated with the for loop which does not exist any longer.  
		#( ($resnum1 eq '-') and ($resnum2 eq '-') and ($resnum3 eq '-') ) and next;

				# a complete codon #
		my $fl_comp_codon = ( ($resnum2 ne '-') and ($resnum2 == $resnum1 +1) 
				and  ($resnum3 ne '-') and ($resnum3 == $resnum2 +1) ) ;

		if ($fl_comp_codon) {
			push @sel_frames , ($frame1, $frame2, $frame3);
			$l += 3;
		} else {
			$l++;
		}
	}						# end of the while-loop #

	return \@sel_frames;
}				# end of the subroutine sel_codon_frames_in_ref #


# extract_codon_alignment: This subroutine receives a sequence part of the multiple alignment, a complete list of residue numbers, and a subset of frames,
#			then extracts the parts composed of complete codon alignments.
#
sub extract_codon_alignment (\@\@\@) {

	my ($rf_align, $rf_resnums, $rf_sb_frames) = @_;

	@{$rf_align} or (return (undef, undef, undef) );
	@{$rf_resnums} or (return ([], [], []) );
	my $sel_frames = @{$rf_sb_frames} ? $rf_sb_frames : [0 .. $#{$rf_resnums}];

	$sel_frames = select_in_phase (@{$rf_resnums}, @{$sel_frames});
	$sel_frames = select_unamb_frames (@{$rf_align}, @{$sel_frames});
	$sel_frames = sel_complete_codon_frames (@{$rf_resnums}, @{$sel_frames});

	my $sel_columns = diff_malign (@{$rf_align}, @{$sel_frames});
	my @sel_resnums = ();
	foreach my $frame (@{$sel_frames}) { push @sel_resnums, $rf_resnums->[$frame]; }

	return ($sel_frames, $sel_columns, \@sel_resnums);

}				# end of the subroutine extract_codon_alignment #


# extract_codon_aln_with_ref: This subroutine receives a sequence part of the multiple alignment,
#			 a complete list of residue numbers, a subset of frames,
#			 and the index of a sequence which is to be used as the "reference",
#			then extracts parts of the alignment containing complete codons of the "reference"
#			and consisting solely of unambiguous nucleotides.
#
#		This subroutine does NOT care if the codons in different sequences are in-phase or not.
#		It just looks at the reference sequence.
#
sub extract_codon_aln_with_ref (\@\@\@$) {

	my ($rf_align, $rf_resnums, $rf_sb_frames, $indx_ref) = @_;

	@{$rf_align} or (return (undef, undef, undef) );
	@{$rf_resnums} or (return ([], [], []) );
	my $sel_frames = @{$rf_sb_frames} ? $rf_sb_frames : [0 .. $#{$rf_resnums}];

	#$sel_frames = select_in_phase (@{$rf_resnums}, @{$sel_frames});
	$sel_frames = select_unamb_frames (@{$rf_align}, @{$sel_frames});
	#$sel_frames = sel_complete_codon_frames (@{$rf_resnums}, @{$sel_frames});
	$sel_frames = sel_codon_frames_in_ref (@{$rf_resnums}, @{$sel_frames}, $indx_ref);

	my $sel_columns = diff_malign (@{$rf_align}, @{$sel_frames});
	my @sel_resnums = ();
	foreach my $frame (@{$sel_frames}) { push @sel_resnums, $rf_resnums->[$frame]; }

	return ($sel_frames, $sel_columns, \@sel_resnums);

}				# end of the subroutine extract_codon_aln_with_ref #


# pr_block_maln_dec: This subroutine prints out a  block-separated multiple- alignment in a recorated form.
#			It returns the count of completely identical columns in the alignment. 
#
sub pr_block_maln_dec (\@\@\@\@\*$) {

	my ($rf_bl_columns, $rf_bl_frames, $rf_resnums, $rf_seqIDs, $fh, $bnum) = @_;

	my $ct_blocks = @{$rf_bl_columns};
	my $ct_seqs = @{$rf_seqIDs};
	unless ($ct_blocks) { print $fh "\n"; return 0; }
 	unless (@{$rf_bl_frames} == $ct_blocks) {
		print STDERR "Illegal input for \"print_block_maln\".\n";
		return undef;
	}

	my ($rf_bl_starts, $rf_bl_ends) = block_starts_ends (@{$rf_resnums}, @{$rf_bl_frames});

	unless (defined $rf_bl_starts) {
		print STDERR "Subroutine \"block_starts_ends\" failed.\n";
		return undef;
	}

	my @labels = ();
	foreach my $seqID (@{$rf_seqIDs}) {

		my $label = $seqID . ' ' x (15 - length($seqID) );
		push @labels, $label;
	}

	my $id_ct = 0;

	for (my $i=0; $i < $ct_blocks; $i++) { # the outer for #

	    my $frame = $rf_bl_frames->[$i];
	    my $ct_columns = @{$frame};
	    my $frame_start = $frame->[0] + 1;
	    my $frame_end = $frame->[$#{$frame}] + 1;
	    printf $fh "> (frame: %d -- %d, %d columns) ", $frame_start, $frame_end, $ct_columns;

	    my @reservoir = my @intervals = ();
	    for (my $j=0; $j < $ct_seqs; $j++) { # the inner for #
		my $start = $rf_bl_starts->[$i][$j];
		my $end = $rf_bl_ends->[$i][$j];
		unless (defined $start) { 
			print STDERR "The starting position of block $i, sequence $rf_seqIDs->[$j] is NOT defined.\n";
			return undef;
		}
		unless (defined $end) { 
			print STDERR "The ending position of block $i, sequence $rf_seqIDs->[$j] is NOT defined.\n";
			return undef;
		}
		#$start = ($start eq '-') ? (' ' x 4 . '-' x 3) : sprintf ("%7d", $start);
		#$end = ($end eq '-') ? (' ' x 4 . '-' x 3) : sprintf ("%7d", $end);
		my $seq = '';
		for (my $k=0; $k<@{$rf_bl_columns->[$i]}; $k++) {

			unless (defined $rf_bl_columns->[$i][$k]) {
				print STDERR "The column $k of block $i is NOT defined.\n";
				return undef;
			}
			unless (defined $rf_bl_columns->[$i][$k][$j]) {
				print STDERR "The sequence $rf_seqIDs->[$j] is NOT defined at the column $k of block $i.\n";
				return undef;
			}
			$seq .= $rf_bl_columns->[$i][$k][$j];
		}

		#my $line = $labels[$j] . $start . '  ' . $seq . $end . "\n";
		#print $fh $line;
		push @reservoir, \$seq;
		push @intervals, "$start -- $end";

	    }				# end of the inner for #

	    my $id_ind = id_indicator(@reservoir);
	    $id_ct +=  (${$id_ind} =~ tr/\*//);

	    print $fh " (intervals: " . join( '; ', @intervals) . ")\n";

	   for (my $j=0; $j<$ct_seqs; $j++) { print $fh "$labels[$j]${$reservoir[$j]}\n"; }
	   print $fh '>' . ' ' x 14 . ${$id_ind} . "\n";
	   print $fh "\n";
	}				# end of the outer for #

	print $fh "\n";
	return $id_ct;
}				# end of the subroutine pr_block_maln_dec #


# prepare_oden_input:
#	This subroutine receives a multiple CDS alignment,
#	then prepares input data for the program 'dists' in the ODEN package,
#	which is a concatenated alignment of CDS regions
#	consisting only of codons that are intact, i.e.
#	contain no gaps and/or ambiguities in it.
#
# INPUT: (\@in_cds_align), where
#	${$in_cds_align[$k]} is the $k th CDS sequence embeded in the input alignment;
#
# OUTPUT: (\@bl_columns, \@bl_frames, \@nuc_resnums, \@removed_frames); 
#	$bl_columns[$r]->[$i][$k]  is the nucleotide residue that the $k th CDS
#	has at the $i th site of the $r th block of the codon alignment;
#	$bl_frames[$r]->[$i] is the frame coordinate of the $i th site in the $r th block#		(starting at 0, which represents the leftmost site);
#	$nuc_resnums[$l]->[$k] is the nucleotide residue number of the $k th sequence
#	at the $l th site in the entire input CDS alignment,
#	or '-' if the site is occupied with a gap;
#	$removed_frames[$m] is the frame coordinate of the $m th site that canNOT
#	constitute an intact codon site while consisting only of in-phase unambiguous
#	nucleotide residues.
#
sub prepare_oden_input (\@) {

	my ($in_cds_align) = @_;

	my $ct_seqs = @{$in_cds_align};
	unless ($ct_seqs>1)  {
	    print STDERR "[ERROR 'prepare_oden_input'] An INSUFFICEINT NUMBER of sequences ($ct_seqs) to construct an alignment !\n";
	    return (undef, undef, undef, undef);
	}

		# Nucleotide residue numbers within the CDSs (1 for the leftmost residue)
	my $nuc_resnums = assign_number (@{$in_cds_align});
	unless (defined $nuc_resnums) {
	    print STDERR "[ERROR 'prepare_oden_input'] The subroutine 'assign_number' FAILED !\n";
	    reuturn (undef, undef, undef, undef);
	}

	my $ct_full_clms = @{$nuc_resnums};
	if ($ct_full_clms < 3) {
	    print STDERR "[ERROR 'prepare_oden_input'] The input alignment only contains $ct_full_clms unambiguous nucleotide sites !\n";
	    return (undef, undef, undef, undef);
	}
	my @full_frames = (0 .. $ct_full_clms-1); # frame numbers of the alignment (Start = 0)

	my $unamb_frames = select_unamb_frames (@{$in_cds_align}, @full_frames);
	my $ct_unamb_clms = @{$unamb_frames};

	my ($selected_frames, $selected_columns, $sel_resnums)
	    = extract_codon_alignment (@{$in_cds_align}, @{$nuc_resnums}, @full_frames);
	unless (defined $selected_frames) {
	    print STDERR "[ERROR 'prepare_oden_input'] 'extract_codon_alignment' FAILED !\n";
	    return (undef, undef, undef, undef);
	}

	my $ct_sel_clms = @{$selected_frames};
	if ($ct_sel_clms < 3 ) {
	    print STDERR "[ERROR 'prepare_oden_input'] The input alignment canNOT keep even a codon site intact because it only has $ct_sel_clms decent sites!\n";
	    return (undef, undef, undef, undef);
	}

	(undef, my $removed_frames, my $ex_2)
		= diff_nmarray (@{$unamb_frames}, @{$selected_frames});
	if (@{$ex_2}>0) { ## Normally, this cannot happen ##
	    print STDERR "[ERROR 'prepare_oden_input'] The selected frame contains ",
		scalar (@{$ex_2}), " columns with ambiguous nucleotides !\n";
	    return (undef, undef, undef, undef);
	}

	my $thrsh_distcolumn = 1;
		# Two contiguous sites are put in the same block
		# if the difference of their frame coordinates falls within this value.

	my ($bl_columns, $bl_resnums, $bl_frames)
		= construct_blocks (@{$selected_columns}, @{$sel_resnums},
			@{$selected_frames}, $thrsh_distcolumn);

	return ($bl_columns, $bl_frames, $nuc_resnums, $removed_frames); 

}		# End of the subroutine 'prepare_oden_input' #


# extract_intact_codonaln:
#	This subroutine receives a multiple CDS alignment,
#	and extracts 'intact' codon sites, each of which
#	consists ONLY of unambiguous nucleotide residues and
#	comprises a complete codon along with the neighboring sites.
#
# INPUT: (\@in_cds_aln,  \@relv_frames, \@beg_coords, $gap), where
#	${$in_cds_aln[$m]} is the $m th CDS sequence embedded in the alignment;
#	\@relv_frames stores the relevant alignment coordinate frames.
#	the coordinate frame starts with 0 and ends with {length} - 1;
#	$beg_coords[$m] is the beginning coordinate of the $m th CDS sequence,
#	with the coordinate 1 at the translation starting site;
#
# OUTPUT: (\@out_cds_aln, \@selected_frames, \@sel_resnums), where
#	${$out_cds_aln[$m]} is the concatenated regions of the $m th sequence
#	in the 'intact' codon alignment;
#	$selected_frames[$l] is the frame coordinate of the $l th site comprising
#	the 'intact' codon alignment;
#	$sel_resnums[$l]->[$m] is the coordinate, or the nucleotide residue number,
#	of the $l th constituent site of the $m th CDS sequence
#	embedded in this alignment.
#
sub extract_intact_codonaln (\@\@\@$) {

	my ($in_cds_align, $relv_frames, $beg_coords, $gap) = @_;

	my $ct_seqs = @{$in_cds_align};
	unless ($ct_seqs>1)  {
	    print STDERR "[ERROR 'extract_intact_codonaln'] An INSUFFICEINT NUMBER of sequences ($ct_seqs) to construct an alignment !\n";
	    return (undef, undef, undef);
	}

		# Nucleotide residue numbers within the CDSs
		# (1 for the leftmost residue)
	my $nuc_resnums = assign_number_gen (@{$in_cds_align}, @{$beg_coords}, $gap);
	unless (defined $nuc_resnums) {
	    print STDERR "[ERROR 'extract_intact_codonaln'] The subroutine 'assign_number' FAILED !\n";
	    reuturn (undef, undef, undef);
	}

	my $ct_full_clms = @{$nuc_resnums};
	if ($ct_full_clms < 3) {
	    print STDERR "[ERROR 'extract_intact_codonaln'] The input alignment only contains $ct_full_clms unambiguous nucleotide sites !\n";
	    return (undef, undef, undef);
	}
	##my @full_frames = (0 .. $ct_full_clms-1); # frame numbers of the alignment (Start = 0)
	$relv_frames = (@{$relv_frames}>0) ? $relv_frames : [0 .. $ct_full_clms-1];

	my $unamb_frames = select_unamb_frames (@{$in_cds_align}, @{$relv_frames});
	my $ct_unamb_clms = @{$unamb_frames};

	my ($selected_frames, $selected_columns, $sel_resnums)
	  = extract_codon_alignment (@{$in_cds_align}, @{$nuc_resnums}, @{$relv_frames});
	unless (defined $selected_frames) {
	    print STDERR "[ERROR 'extract_intact_codonaln'] 'extract_codon_alignment' FAILED !\n";
	    return (undef, undef, undef);
	}

	my $ct_sel_clms = @{$selected_frames};
	if ($ct_sel_clms < 3 ) {
	    print STDERR "[ERROR 'extract_intact_codonaln'] The input alignment canNOT keep even a codon site intact because it only has $ct_sel_clms decent sites!\n";
	    return (undef, undef, undef);
	}

		## Concatenate the selected columns ##

	my @out_cds_aln = ();
	for (my $m=0; $m<$ct_seqs; $m++) { my $seq = ''; push @out_cds_aln, \$seq; }

	while (my $column = shift @{$selected_columns}) { # The outer while-loop #
	    for (my $m=0; $m<$ct_seqs; $m++) { # The inner for-loop #
		my $res = $column->[$m];
		${$out_cds_aln[$m]} .= ($res eq '-') ? $gap : $res;
			# (fransform back the gap character) #
	    }				# End of the inner for-loop #
	}				# End of the outer while-loop #

	return (\@out_cds_aln, $selected_frames, $sel_resnums);

}		# end of the subroutine 'extract_intact_codonaln' #


# extract_intact_cnct_codonaln:
#	This revised version of the subrouine 'extract_intact_codonaln'
#	can process concatenated CDS alignments as well.
#
# extract_intact_codonaln:
#	This subroutine receives a multiple CDS alignment,
#	and extracts 'intact' codon sites, each of which
#	consists ONLY of unambiguous nucleotide residues and
#	comprises a complete codon along with the neighboring sites.
#
# INPUT: (\@in_cds_aln,  \@relv_frames, \@beg_coords, \@end_coords, $gap), where
#	${$in_cds_aln[$m]} is the $m th CDS sequence embedded in the alignment;
#	\@relv_frames stores the relevant alignment coordinate frames.
#	the coordinate frame starts with 0 and ends with {length} - 1;
#	$beg_coords[$m] is the beginning coordinate of the $m th CDS sequence,
#	with the coordinate 1 at the translation starting site;
#
# OUTPUT: (\@out_cds_aln, \@selected_frames, \@sel_resnums), where
#	${$out_cds_aln[$m]} is the concatenated regions of the $m th sequence
#	in the 'intact' codon alignment;
#	$selected_frames[$l] is the frame coordinate of the $l th site comprising
#	the 'intact' codon alignment;
#	$sel_resnums[$l]->[$m] is the coordinate, or the nucleotide residue number,
#	of the $l th constituent site of the $m th CDS sequence
#	embedded in this alignment.
#
sub extract_intact_cnct_codonaln (\@\@\@\@$) {

	my ($in_cds_align, $relv_frames, $beg_coords, $end_coords, $gap) = @_;

	my $ct_seqs = @{$in_cds_align};
	unless ($ct_seqs>1)  {
	    print STDERR "[ERROR 'extract_intact_cnct_codonaln'] An INSUFFICEINT NUMBER of sequences ($ct_seqs) to construct an alignment !\n";
	    return (undef, undef, undef);
	}

		# Nucleotide residue numbers within the CDSs
	###my $nuc_resnums = assign_number_gen (@{$in_cds_align}, @{$beg_coords}, $gap);
	my $nuc_resnums = assign_number_cnctseqs (@{$in_cds_align}, @{$beg_coords}, @{$end_coords}, $gap);
	unless (defined $nuc_resnums) {
	    print STDERR "[ERROR 'extract_intact_cnct_codonaln'] The subroutine 'assign_number' FAILED !\n";
	    reuturn (undef, undef, undef);
	}

	my $ct_full_clms = @{$nuc_resnums};
	if ($ct_full_clms < 3) {
	    print STDERR "[ERROR 'extract_intact_cnct_codonaln'] The input alignment only contains $ct_full_clms unambiguous nucleotide sites !\n";
	    return (undef, undef, undef);
	}
	##my @full_frames = (0 .. $ct_full_clms-1); # frame numbers of the alignment (Start = 0)
	$relv_frames = (@{$relv_frames}>0) ? $relv_frames : [0 .. $ct_full_clms-1];

	my $unamb_frames = select_unamb_frames (@{$in_cds_align}, @{$relv_frames});
	my $ct_unamb_clms = @{$unamb_frames};

	my ($selected_frames, $selected_columns, $sel_resnums)
	  = extract_codon_alignment (@{$in_cds_align}, @{$nuc_resnums}, @{$relv_frames});
	unless (defined $selected_frames) {
	    print STDERR "[ERROR 'extract_intact_codonaln'] 'extract_codon_alignment' FAILED !\n";
	    return (undef, undef, undef);
	}

	my $ct_sel_clms = @{$selected_frames};
	if ($ct_sel_clms < 3 ) {
	    print STDERR "[ERROR 'extract_intact_cnct_codonaln'] The input alignment canNOT keep even a codon site intact because it only has $ct_sel_clms decent sites!\n";
	    return (undef, undef, undef);
	}

		## Concatenate the selected columns ##

	my @out_cds_aln = ();
	for (my $m=0; $m<$ct_seqs; $m++) { my $seq = ''; push @out_cds_aln, \$seq; }

	while (my $column = shift @{$selected_columns}) { # The outer while-loop #
	    for (my $m=0; $m<$ct_seqs; $m++) { # The inner for-loop #
		my $res = $column->[$m];
		${$out_cds_aln[$m]} .= ($res eq '-') ? $gap : $res;
			# (fransform back the gap character) #
	    }				# End of the inner for-loop #
	}				# End of the outer while-loop #

	return (\@out_cds_aln, $selected_frames, $sel_resnums);

}		# end of the subroutine 'extract_intact_cnct_codonaln' #


# remove_stop_codons:
#	This subroutine removes stop codons, as well as codons withOUT
#	corresponding amino acids, from an 'intact' codon alignment
#	that follows the format of the output of 'extract_intact_codonaln'.
#
# NOTE that this subroutine eats up the input alignment.
#
# INPUT: (\@in_cds_aln, \@in_frames, \@in_resnums, \%codon2aa, $stop), where
#	${$in_cds_aln[$m]} is the concatenated regions of the $m th sequence
#	in the input 'intact' codon alignment;
#	$in_frames[$l] is the frame coordinate of the $l th site comprising
#	the input 'intact' codon alignment;
#	$in_resnums[$l]->[$m] is the coordinate, or the nucleotide residue number,
#	of the $l th constituent site of the $m th CDS sequence
#	embedded in this alignment;
#	%codon2aa = ($codon => $aa, ...), with $codon is sure to expressed with
#	a set of (upper-case) characters including 'ATGC';
#	$stop represents translation stop.
#
# OUTPUT: (\@out_cds_aln, \@out_frames, \@out_resnums,
#		\@rmvd_codonaln, \@rmvd_frames, \@rmvd_resnums), where
#	@out_cds_aln, @out_frames, and @out_resnums,
#	which follow the same formats as
#	@in_cds_aln, @in_frames, and @in_resnums, respectivly,
#	give information on the alignment after removing the stop codons;
#	$rmvd_codonaln[$ll]->[$m] is the codon of the $m th sequence
#	in the $ll th removed  codon-site;
#	$rmvd_frames[$ll] is the frame of the 1st position of the $ll th removed codon-site;
#	$rmvd_resnums[$ll]->[$m] is the (nucleotide) residue number of
#	1st position of the $ll th removed codon-site for the $m th  sequence.
#
sub remove_stop_codons (\@\@\@\%$) {

	my ($in_cds_aln, $in_frames, $in_resnums, $codon2aa, $stop) = @_;

	my $ct_seqs = @{$in_cds_aln};

	($ct_seqs>0) or (return ([], [], [], [], [], []));

	my $ct_ntsites = length ${$in_cds_aln->[0]};

	unless ((@{$in_frames} == $ct_ntsites) and (@{$in_resnums} == $ct_ntsites)) {
	    print STDERR "[ERROR 'remove_stop_codons'] The number of input frames (", scalar (@{$in_frames}), ") or that of residue numbers (", scalar (@{$in_resnums}), ") CLASHES with the length of the codon alignment ($ct_ntsites) !!\n\n";
	    return (undef, undef, undef, undef, undef, undef);
	}

	my @out_cds_aln = my @out_frames = my @out_resnums = ();
	my @rmvd_codonaln  = my @rmvd_frames = my @rmvd_resnums = ();
	for (1 .. $ct_seqs) { my $init = ''; push @out_cds_aln, \$init; }

	my $ct_codons = int ($ct_ntsites /3);

	for (my $l=0; $l<$ct_codons; $l++) { # The outer for-loop #

	    my @codon_clm = my @indices_stop = ();
	    for (my $m=0; $m<$ct_seqs; $m++) { # The 1st inner for-loop #
		my $codon = substr (${$in_cds_aln->[$m]}, 0, 3, '');
		push @codon_clm, $codon;
		$codon = uc $codon;
		$codon =~ tr/U/T/ ;
		(!(defined $codon2aa->{$codon}) or ($stop eq $codon2aa->{$codon}))
			and (push @indices_stop, $m);
	    }				# End of the 1st inner for-loop #

			# Contains stop codons or dubious codons #
	    if (@indices_stop>0) {
		push @rmvd_codonaln, \@codon_clm;
		push @rmvd_frames, $in_frames->[3*$l];
		push @rmvd_resnums, $in_resnums->[3*$l];
		next;
	    }

	    	# Append decent codon columns to the output alignment #
	   for (my $m=0; $m<$ct_seqs; $m++) { # The 2nd inner for-loop #
		${$out_cds_aln[$m]} .= $codon_clm[$m];
	   }			# End of the 2nd inner for-loop #

	   push @out_frames, @{$in_frames}[3*$l, 3*$l+1, 3*$l+2];
	   push @out_resnums, @{$in_resnums}[3*$l, 3*$l+1, 3*$l+2];

	}				# End of the outer for-loop #

	return (\@out_cds_aln, \@out_frames, \@out_resnums,
		\@rmvd_codonaln, \@rmvd_frames, \@rmvd_resnums);

}		# End of the subroutine 'remove_stop_codons' #


# pr_in_oden_dec: This subroutine prints out a  block-separated multiple- alignment in a decorated
#		format which can be used as an input for ODEN's program "dists".
#		It returns the count of completely identical columns in the alignment. 
#
sub pr_in_oden_dec (\@\@\@\@\*$) {

	my ($rf_bl_columns, $rf_bl_frames, $rf_resnums, $rf_seqIDs, $fh, $offset) = @_;
	my $dis_len = 60;

	my $ct_blocks = @{$rf_bl_columns};
	my $ct_seqs = @{$rf_seqIDs};
	unless ($ct_blocks) { print $fh "\n"; return 0; }
 	unless (@{$rf_bl_frames} == $ct_blocks) {
		print STDERR "Illegal input for \"print_block_maln\".\n";
		return undef;
	}

	my $id_ct = my $tot_ct_clms = 0;

	for (my $i=0; $i < $ct_blocks; $i++) { # the outer for #

	    my $frame = $rf_bl_frames->[$i];
	    my $ct_columns = @{$frame};
	    my $frame_start = $frame->[0] + 1;
	    my $frame_end = $frame->[$#{$frame}] + 1;
	    my $block_num = $i+1+$offset;
	    printf $fh "> Block_%d (overall_frame: %d  -- %d, subtotal %d columns):\n\n", $block_num, $frame_start, $frame_end, $ct_columns;

	    $tot_ct_clms += $ct_columns;

	    my $ct_sbbl = int ( $ct_columns / $dis_len);
	    my @sbbl_columns = my @sbbl_frames = ();
	    for (my $t=0; $t<$ct_sbbl; $t++) {	# the inner for #
			my @tmp1 = @{$rf_bl_columns->[$i]}[ $t * $dis_len .. ($t +1) * $dis_len -1];
			push @sbbl_columns, \@tmp1;
			my @tmp2 = @{$frame}[$t * $dis_len .. ($t+1) * $dis_len -1];
			push @sbbl_frames, \@tmp2;
	    }						# end of the inner for #

	    if ( $ct_columns % $dis_len) {
			my @tmp1 = @{$rf_bl_columns->[$i]}[ $ct_sbbl * $dis_len .. $#{$frame} ];
			push @sbbl_columns, \@tmp1;
			my @tmp2 = @{$frame}[ $ct_sbbl * $dis_len .. $#{$frame} ];
			push @sbbl_frames, \@tmp2;
	    }

	    my $sbct_id =  pr_block_maln_dec (@sbbl_columns, @sbbl_frames, @{$rf_resnums}, @{$rf_seqIDs}, *${$fh}, $block_num);
	    unless (defined $sbct_id) {
		print STDERR "\'pr_block_maln_dec\' failed for block $block_num.\n";
		return undef;
	    }
	    $id_ct += $sbct_id;
	}

	printf $fh "> Total \#{columns} = $tot_ct_clms;  Total \#{monomorphic_columns} = $id_ct.\n";

	return $id_ct;
}				# end of the subroutine pr_in_oden_dec #

# chop_alignment:  This subroutine receives an alignment of the form:
#			[\$seq_in_aln_1, \$seq_in_aln_2, ...],
#			accompanied by a list of lengths ,
#			and returns a set of subalignments produced by 'chop'ping the
#			input alignment to the lengths specified by the length list.
#			The 2nd output is a flag which takes the value 0 if the total length
#			of the subalignments equals to the length of the input,
#			1 if  the input is longer, and -1 if it is shorter.  
#
# NOTE: This subroutine "eats up" the input alignment.
#	Therefore, you have to copy the alignment beforehand if you want to keep it.
#
sub chop_alignment (\@\@) {

	my ($alignment, $lengths)  = @_;

	my $ct_seqs = @{$alignment};

	$ct_seqs or (return ([], 0) );

	my $whole_len = length (${$alignment->[0]} );

	my @chopped = ();

	foreach my $len (@{$lengths}) {

	    my @tmp = ();
	    if ($whole_len <= $len) {
		for (my $i=0; $i<$ct_seqs; $i++) {
		   my $sbaln = ${$alignment->[$i]};
		    push @tmp, \$sbaln;
		    ${$alignment->[$i]} = '';
		}

	    } else {
		for (my $i=0; $i<$ct_seqs; $i++) {
			my $sbaln = substr ( ${$alignment->[$i]}, 0, $len, '');
			push @tmp, \$sbaln;
		}
	    }

	    push @chopped, \@tmp;
	    $whole_len -= $len;
	    ($whole_len < 0) and (return (\@chopped, -1) );

	}

	return ( \@chopped, (($whole_len == 0) ? 0 : 1) ) ;

}			# end of the subroutine chop_alignment #


1;	# a 'spell' which enables the module 'use'-able and 'require'-able #


