package MyMalnManipu_pep_spe;
#
###################################################################################
#	DENSERM_P: A package to DEtect Negative SElection on Recurrent Mutations
#			(almost exclusively via Perl modules and scripts).
#	Version 0.3: Copyright (C) 2012 Kiyoshi Ezawa
#
#    This package is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This package is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License, "GNU_GPL.txt",
#    along with this package.  If not, see <http://www.gnu.org/licenses/>.
#
#    The author can be contacted by e-mailing to <kezawa dot ezawa3 at gmail dot com>
#    (replace " dot " and " at " with "." and "@", respectively).
###################################################################################
#
# MyMalnManipu_pep_spe.pm:
#	This module is a collection of subroutines
#	that conduct special sequence manipulations. 
#
# MyMalnManipu_pep.pm:
#	This module is basically a collection of subroutines
#	each of which is forged by revising  a subroutine in the module
#	'MyMalnManipu.pm' so that it can handle peptide sequences.
#	The module only contains subroutines whose revision is necessary to cope with
#	peptide alignments.
#	Thus the module lacks subroutines which can innately handle both nucleotide
#	and peptide alignments, and those which can NOT EVEN IN PRINCIPLE convert to
#	the peptide counterparts.
#	You should find such subroutines in the original module 'MyMalnManipu.pm'.
#
# MyMalnManipu.pm:
#	This module manipulates multiple or pairwise alignments
#	in a more sophisticated manner than the module "MyMalnIO.pm".
#
# NOTE: This package is almost exclusively for nucleotide multiple-alignments.
#
# REVISED on Nov 19 (Fri), 2004:
#	The subroutine 'compare_columns', which is  a satellite of 'compare_malign',
#	no longer depends on the recursion technique.


use strict;
use warnings;
use Exporter;

use MyArrayManipu_num qw( diff_nmarray );
##use MyMalnManipu qw( block_starts_ends id_sliding_window );

our @ISA = qw( Exporter);

our @EXPORT_OK = qw(
	init_consv_hashes consv_indicator_aa
	);
	## separate_nogap_sites_pep monosite_picker_pep id_indicator_aa
	## print_block_maln_pep maln_identities_pep block_maln_identities_pep
	## block_maln_id_distribution_pep bl_id_sliding_window_pep
	## mask_frames_pep compl_mask_frames_pep select_unamb_frames_pep
	## pr_block_maln_dec_pep pep_diff2dist pep_pairwise_diffs 

our %AA2STRONG = ( # Groups of 'Strong' conservation #
		);
our %AA2WEAK = ( # Groups of 'Weak' conservation #
		);

# BE SURE to call this subroutine before using the subroutine 'consv_indicator_aa'.
#
sub init_consv_hashes () {

   my @strong = ('STA', 'NEQK', 'NHQK', 'NDEQ', 'QHRK', 'MILV', 'MILF', 'HY', 'FYW');
   my @weak = ('CSA', 'ATV', 'SAG', 'STNK', 'STPA', 'SGND', 'SNDEQK', 'NDEQHK',
		'NEQHRK', 'FVLIM', 'HFY');

   for (my $k=0; $k<@strong; $k++) { # The 1st outer for-loop #
	my @aas = split //, $strong[$k];
	foreach my $aa (@aas) { # The 1st inner for-loop #
	    if (defined $AA2STRONG{$aa}) {
		push @{$AA2STRONG{$aa}}, $k+1;
	    } else {
		$AA2STRONG{$aa} = [$k+1];
	    }
	}		# End of the 1st inner for-loop #
   }			# End of the 1st outer for-loop #

   for (my $k=0; $k<@weak; $k++) { # The 2nd outer for-loop #
	my @aas = split //, $weak[$k];
	foreach my $aa (@aas) { # The 2nd inner for-loop #
	    if (defined $AA2WEAK{$aa}) {
		push @{$AA2WEAK{$aa}}, $k+1;
	    } else {
		$AA2WEAK{$aa} = [$k+1];
	    }
	}		# End of the 2nd inner for-loop #
   }			# End of the 2nd outer for-loop #

}		# End of the subroutine 'init_consv_hashes'


# consv_indicator_aa:
#	This subroutine is a revised version of 'id_indicator_aa',
#	and the resulting line indicates the position of high
#	amino acid conservation.
#	'*' stands for perfect conservation, ':' for 'strong' conservation,
#	and '.' for 'weak' conservation.
#
# id_indicator_aa:
#	This subroutine makes a line which indicates
#	the positions of columns in the alignment
#	consisting only of one kind of amino acids.
#
# NOTE: BE SURE TO CALL the subroutine 'init_consv_hashes' before using this subroutine.
#
sub consv_indicator_aa (\@) {

	my ($rf_seqs) = @_;

	my $id_ind = '';
	my $ct_seqs = @{$rf_seqs};
	($ct_seqs) or (return \$id_ind ) ;

	my @copy = ();
	foreach my $seq (@{$rf_seqs}) { push @copy, uc ${$seq}; }
	my $seq_len = length (${$rf_seqs->[0]});
	my $first_seq = shift @copy;

	foreach my $l  (1 .. $seq_len) {

	    my $first_char = substr($first_seq, 0, 1, '');
	    #my $ct_id = ( ($first_char eq '-') or ($first_char eq 'N') ) ? 0 : 1;
	    my $ct_id = ($first_char =~ tr/ACDEFGHIKLMNPQRSTUVWY// );
	    	# NOTE: U = 'selenocystein'

	    my $cons_strong = $AA2STRONG{$first_char};
	    (defined $cons_strong) or ($cons_strong = []);
	    my $cons_weak = $AA2WEAK{$first_char};
	    (defined $cons_weak) or ($cons_weak = []);

	    for (my $i=0; $i<@copy; $i++) {
		my $char = substr ($copy[$i], 0, 1, '');
		($char eq $first_char) and $ct_id++;
		my $strong = $AA2STRONG{$char};
		my $weak = $AA2WEAK{$char};
		if ((defined $strong) and (@{$cons_strong}>0)) {
		    ($cons_strong, undef, undef) = diff_nmarray (@{$strong}, @{$cons_strong});
		} else {
		    $cons_strong = [];
		}
		if ((defined $weak) and (@{$cons_weak}>0)) {
		    ($cons_weak, undef, undef) = diff_nmarray (@{$weak}, @{$cons_weak});
		} else {
		    $cons_weak = [];
		}
	    }

	    ##$id_ind .= ($ct_id == $ct_seqs) ? '*' : ' ';
	    if ($ct_id == $ct_seqs) { $id_ind .= '*'; }
	    elsif (@{$cons_strong}>0) { $id_ind .= ':'; }
	    elsif (@{$cons_weak}>0) { $id_ind .= '.'; }
	    else { $id_ind .= ' '; }
	}

	return \$id_ind;
}				# end of the subroutine 'consv_indicator_aa' #


1;		# a 'spell' which enables the module 'use'-able and 'require'-able #

