package MyPWA_Analyses;
#
###################################################################################
#	
#	LASTPIECE_P (Local Alignment-STate Probabilities that Insertion-type and dEletion-type gaps Co-Exist, type P):
#	    A package of programs to compute the probabilities of gapped segments in each of which an insertion-type gap and a deletion-type gap co-exist, (which were referred to as case-(iv) gapped segments by Ezawa (2016a?)),
#	    under a stochastic model of sequence evolution with biologically realistic insertions/deletions.
#	   (Written almost exclusively in Perl.)
#
#	Version 0.3: Copyright (C) 2020 Kiyoshi Ezawa
#
#    This package is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This package is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License, "GNU_GPL.txt",
#    along with this package.  If not, see <http://www.gnu.org/licenses/>.
#
#    The author can be contacted by e-mailing to <kezawa dot ezawa3 at gmail dot com>
#    (replace " dot " and " at " with "." and "@", respectively).
#
###################################################################################
#
# MyPWA_Analyses.pm:
#	This Perl package consists of some subroutines that help analyze PWAs.
#	[CREATION STARTED on 2020/04/06, ENDED on 2020/mm/dd.]

use strict;
use warnings;


=Set aside for later use.
use MyFile_manipu qw(
	open_input_file2
	open_output_file
	); # ADDED 'open_output_file' in creation of 'detect_purge_cands.ver0.5.pl' (on 2019/06/17). #
	# list_sessions_ext
	# tfm_labels2indices

=cut # End of "Set aside for later use."

use Exporter;

our @ISA =qw(Exporter);

our @EXPORT_OK = qw( 
	count_config_gsegs_in_pwa
	pr_loose_2D_array
	sum_loose_2D_arrays
	readin_2D_array
	readin_1D_array
	cmpt_dbl_cumulative_2Darray
	);
	# Added 'readin_2D_array' & 'cmpt_dbl_cumulative_2Darray' on 2020/04/09. #

sub count_config_gsegs_in_pwa (\@$);
sub pr_loose_2D_array (\*\@);
sub sum_loose_2D_arrays (\@\@);
sub readin_2D_array (\*$$$); # ADDED on 2020/04/08.
sub readin_1D_array (\*$$$); # ADDED on 2020/04/12.

sub cmpt_dbl_cumulative_2Darray (\@$$); # ADDED on 2020/04/08.



		### DEFINITIONS of the Subroutines. ###

# count_config_gsegs_in_pwa:
#	This subroutine receives an input pairwise alignment (PWA),
#	and counts the frequencies of configurations of gapped segments in it.
#	[CREATION STARTED on 2020/04/06, ENDED on 2020/04/06.]
#
#	NOTE that this subroutine eats up the input PWA.
#
#	[DEFINITION]:
#	There are two types of gapped segments:
#	Each "bulk" gapped segment is flanked by gapless columns on both sides,
#	and each gapped segment "on end" is flanked by a gapless column on one side,
#	and the other side of it is on the left- or right-end of the PWA.
#
# INPUTS: (\@pwa, $gap), where
#
#	@pwa = (\$seq0, \$seq1) is an input pairwise alignment (PWA), where
#	$seq0 and $seq1 are the two sequences, possibly accommodating some gaps, in the alignment.
#	
#	$gap is the character representing a gap;
#
# OUTPUTS: (\@frqs_cnfg_gsegs_bulk, \@cnfg_gseg_le, \@cnfg_gseg_re), where
#
#	$frqs_cnfg_gsegs_bulk[$n0]->[$n1] is the count, i.e., frequency, of
#	the gapped segment with $n0 residues from $seq0 and $n1 residues from $seq1;
#	[NOTE] $frqs_cnfg_gsegs_bulk[$n0]->[$n1] is UNDEFINED if the PWA has NO such gapped segments,
#	and $frqs_cnfg_gsegs_bulk[$n0] is UNDEFINED if NO gapped segments have $n0 residues from $seq0.
#
#	@cnfg_gseg_le/re = ($ct0, $ct1)
#	is the configuration of the gapped segment on the left/right-end of the PWA, where
#	$ct0 and $ct1 are the numbers of residues from $seq0 and from $seq1, respectively.
#
sub count_config_gsegs_in_pwa (\@$) {

	my ($pwa, $gap) = @_;

	unless (@{$pwa} == 2) {
	    my $size = @{$pwa};
	    print STDOUT "[ERROR 'count_config_gsegs_in_pwa'] ILLEGAL INPUT: the size of ", '@{$pwa}', " is $size, NOT 2: $! !\n";
	    return (undef, undef, undef);
	}

	my ($seq0, $seq1) = @{$pwa};

	my $len_pwa = length ${$seq0};
	my $ct_examined = 0;

		# (1) Examine the gapped segment on the right-end. #
	my $ct0_re = my $ct1_re = 0;

	while ($ct_examined <$len_pwa) { # 1st main outer while-loop. #
	    my $site0 = substr (${$seq0}, -1, 1, '');
	    my $site1 = substr (${$seq1}, -1, 1, '');
	    $ct_examined++;

	    if ($site0 eq $gap) {
		if ($site1 eq $gap) {
		    next; # Both sequences contain gaps at this site. => Just skip the site. #
		} else {
		   $ct1_re++;
		}
	    } else { # if ($site0 ne $gap) 

		if ($site1 eq $gap) {
		    $ct0_re++;
		} else { # if ($site1 ne $gap)
		    last; # Both sequences contain residues at this site. => End the examination.
		}
	    }

	} # End of the 1st main outer while-loop. #

	my @cnfg_gseg_re = ($ct0_re, $ct1_re);
	

		# (2) Examine the gapped segment on the left-end. #
	my $ct0_le = my $ct1_le = 0;

	while ($ct_examined <$len_pwa) { # 2nd main outer while-loop. #
	    my $site0 = substr (${$seq0}, 0, 1, '');
	    my $site1 = substr (${$seq1}, 0, 1, '');
	    $ct_examined++;

	    if ($site0 eq $gap) {
		if ($site1 eq $gap) {
		    next; # Both sequences contain gaps at this site. => Just skip the site. #
		} else {
		   $ct1_le++;
		}
	    } else { # if ($site0 ne $gap) 

		if ($site1 eq $gap) {
		    $ct0_le++;
		} else { # if ($site1 ne $gap)
		    last; # Both sequences contain residues at this site. => End the examination.
		}
	    }

	} # End of the 2nd main outer while-loop. #

	my @cnfg_gseg_le = ($ct0_le, $ct1_le);

	
		# (3) Examine the gapped segment in the bulk. #

	my @frqs_cnfg_gsegs_bulk = ();
	my $ct0_bulk = my $ct1_bulk = 0;

	while ($ct_examined <$len_pwa) { # 3rd main outer while-loop. #
	    my $site0 = substr (${$seq0}, 0, 1, '');
	    my $site1 = substr (${$seq1}, 0, 1, '');
	    $ct_examined++;

	    if ($site0 eq $gap) {
		if ($site1 eq $gap) {
		    next; # Both sequences contain gaps at this site. => Just skip the site. #
		} else {
		   $ct1_bulk++;
		}
	    } else { # if ($site0 ne $gap) 

		if ($site1 eq $gap) {
		    $ct0_bulk++;
		} else { # if ($site1 ne $gap)

		    # Both sequences contain residues at this site. => Tally the configuration.

		    my $row_frqs = $frqs_cnfg_gsegs_bulk[$ct0_bulk];
		    unless (defined $row_frqs) {
			$row_frqs = $frqs_cnfg_gsegs_bulk[$ct0_bulk] = [];
		    }
		    if (defined $row_frqs->[$ct1_bulk]) {
			$row_frqs->[$ct1_bulk]++;
		    } else {
			$row_frqs->[$ct1_bulk] = 1;
		    }

		    $ct0_bulk = $ct1_bulk = 0; # Reset the counters. #
		}
	    }

	} # End of the 3rd main outer while-loop. #


		# Tally the configuration of the finally-examined gapped-segment. #

	my $row_frqs = $frqs_cnfg_gsegs_bulk[$ct0_bulk];
	unless (defined $row_frqs) {
	    $row_frqs = $frqs_cnfg_gsegs_bulk[$ct0_bulk] = [];
	}
	if (defined $row_frqs->[$ct1_bulk]) {
	    $row_frqs->[$ct1_bulk]++;
	} else {
	    $row_frqs->[$ct1_bulk] = 1;
	}


	return (\@frqs_cnfg_gsegs_bulk, \@cnfg_gseg_le, \@cnfg_gseg_re);

}	# End of the subroutine, 'count_config_gsegs_in_pwa'. #


# pr_loose_2D_array:
#	This subroutine prints out a "loose" 2D array,
#	which potentially has many UNDEFINED elements,
#	such as @frqs_cnfg_gsegs_bulk output by 'count_config_gsegs_in_pwa'.
#	[CREATION STARTED on 2020/04/06, ENDED on 2020/04/06.]
#
# INPUTS: ($fh, \@loose_2D_array), where
#
# 	$fh is the file-handle for the output file;
#
#	@loose_2D_array is the loose 2D array, where
#	$loose_2D_array[$x]->[$y] is the ($x, $y)-element of the array (if defined).
#
# OUTPUTS: $flag, where
#
#	$flag	= 1 if the subroutine ended successfully,
#		= 0 if it failed.
#
sub pr_loose_2D_array (\*\@) {

	my ($fh, $loose_2D_array) = @_;

	my $ub_x = @{$loose_2D_array};

	for (my $x=0; $x < $ub_x; $x++) { # Main outer for-loop (over $x). #

	    ##my $array_y = $loose_2D_array; # A FATAL BUG fixed on 2020/04/07. #
	    my $array_y = $loose_2D_array->[$x];

	    unless (defined $array_y) { next; }

	    my $ub_y = @{$array_y};

	    for (my $y=0; $y < $ub_y; $y++) { # Inner for-loop (over $y). #

		my $element = $array_y->[$y];

		unless (defined $element) { next; }

		print $fh join ("\t", "$x:$y", $element), "\n";

	    } # End of the inner for-loop (over $y). #

	} # End of the Main outer for-loop (over $x). #


	return 1;

}	# End of the subroutine, 'pr_loose_2D_array'. #


# sum_loose_2D_arrays:
#	This subroutine performs element-wise summations
#	of TWO "loose" 2D arrays,
#	each of which potentially has many UNDEFINED elements,
#	like @frqs_cnfg_gsegs_bulk output by 'count_config_gsegs_in_pwa'.
#	And the result is stored in the 1st array.
#	[CREATION STARTED on 2020/04/06, ENDED on 2020/04/07.]
#
# INPUTS: (\@array1, \@array2), where
#
#	@array1 and @array2 are the input "loose" 2D arrays, where
#	$array1/2[$x]->[$y] is the ($x, $y)-element of the 1st/2nd array (if defined).
#
#	[NOTE]: @array1 will be OVERWRITTEN by the results of the element-wise summations,
#		whereas @array2 will be kept intact.
#
# OUTPUTS: $flag, where
#
#	$flag	= 1 if the subroutine ended successfully,
#		= 0 if it failed.
#
sub sum_loose_2D_arrays (\@\@) {

	my ($array1, $array2) = @_;

	my $ub_x2 = @{$array2};

	for (my $x=0; $x < $ub_x2; $x++) { # Main Outer for-loop (over $x). #

	    my $row2_at_x = $array2->[$x];

	    unless (defined $row2_at_x) { next; } # Nothing needs to be done. #

	    my $row1_at_x = $array1->[$x];

	    unless (defined $row1_at_x) {
			# Just put the copy of @{$row2_at_x} into the $x th row of @{$array1}. #
		my @cp_row2_at_x = @{$row2_at_x};
		$array1->[$x] = \@cp_row2_at_x;
		next;
	    }

	    my $ub_y2 = @{$row2_at_x};

	    for (my $y=0; $y <$ub_y2; $y++) { # Main inner for-loop (over $y). #

		my $element2 = $row2_at_x->[$y];
		unless (defined $element2) { next; } # Nothing needs to be done. #

		if (defined $row1_at_x->[$y]) {
			# Add $element2 to the $y th element of $row1_at_x. #
		    $row1_at_x->[$y] += $element2;
		} else {
		    	# Insert $element2 into the $y th element of $row1_at_x. #
		    $row1_at_x->[$y] = $element2;
		}

	    } # End of the main inner for-loop (over $y). #

	} # End of the Main Outer for-loop (over $x). #

	return 1;

}	# End of the subroutine, 'sum_loose_2D_arrays'. #

		### RESTART FROM HERE ###

# readin_2D_array:
#	This subroutine reads in a 2D array, which may contain some 'undef' elements,
#	recorded in an input file, which was typically output,e.g., by 'pr_loose_2D_array' above
#	or by any of the "PreFinal" scripts in the LASTPIECE package.
#	[CREATED on 2020/04/08.]
#
#	[INPUT FORMAT] The format of the input file should be:
#	join ("\t", "$x:$y", @values), "\n",
#	where $x and $y (both non-negative integers) are the indexes for the 2D array,
#	and @values stores the values to be analyzed later.
#
# INPUTS: (\*${$fh}, $indx_sbj_fld, $str_undef, $str_end), where
#	
#	$fh is the file-handle for the input file (recording the 2D array);
#
#	$indx_sbj_fld is the index of the subject field;
#	[The index starts with 0 for the row-label (i.e., "$x:$y").]
#
#	$str_undef is a string indicating that the subject fields is undefined.
#
#	$str_end is a string (at the biginning of the line)
#	that indicates the end of the record (for the subject 2D array);
#	it is undef if such a string does not exist.
#
# OUTPUTS: (\@array_2D), where
#
#	@array_2D is the 2D array retrieved, where
#	$array_2D[$x]->[$y] = $values[$indx_sbj_fld-1] if the subject value is defined,
#			    = undef if it is undefined.
#	(NOTE: If the subject values for ALL $y's are undefined for a particular $x,
#	$array_2D[$x] itself is undefined.)
#
sub readin_2D_array (\*$$$) {

	my ($fh, $indx_sbj_fld, $str_undef, $str_end) = @_;

	local $/ = "\n"; # Set the input-record-separator to be "\n".

	my $len_str_end = (defined $str_end) ? length ($str_end) : 0;

	my @array_2D = ();

	while (my $in_line = <$fh>) { # Main while-loop (over the input lines). #

	    unless ($in_line =~ m/\S/) { next; } # Skip if the line contains no non-white-spaces. #

	    if ($len_str_end > 0) {
		my $sbstr = substr ($in_line, 0, $len_str_end);
		if ($sbstr eq $str_end) { last; }
	    }

	    my ($x, $y) = ($in_line =~ m/^(\d+)\:(\d+)\t/);
	    unless (defined $y) { next; }

	    $in_line =~ s/\s+$//; # Remove the white-spaces on the end of the line.

	    my @flds = split /\t/, $in_line;

	    unless (@flds > $indx_sbj_fld) { next; }

	    my $value = $flds[$indx_sbj_fld];

	    unless (defined $value) { next; }
	    if ($value eq $str_undef) { next; }

		# Store the $value as the ($x, $y)-element of @array_2D. #

	    my $row_values = $array_2D[$x];
	    unless (defined $row_values) {
		$row_values = $array_2D[$x] = [];
	    }

	    $row_values->[$y] = $value;

	} # End of the Main while-loop (over the input lines). #


	return \@array_2D;

}	# End of the subroutine, 'readin_2D_array'. #


# readin_1D_array:
#	This subroutine reads in a 1D array, which may contain some 'undef' elements,
#	recorded in an input file, which was typically output,e.g.,
#	by the script 'cmpt_muP_case_i2iii.pl' in the LASTPIECE package.
#	It was created by slightly modifying 'readin_2D_array' above.
#	[CREATED on 2020/04/12.]
#
# readin_2D_array:
#	This subroutine reads in a 2D array, which may contain some 'undef' elements,
#	recorded in an input file, which was typically output,e.g., by 'pr_loose_2D_array' above
#	or by any of the "PreFinal" scripts in the LASTPIECE package.
#	[CREATED on 2020/04/08.]
#
#	[INPUT FORMAT] The format of the input file should be:
#	join ("\t", "$x:$y", @values), "\n",
#	where $x and $y (both non-negative integers) are the indexes for the 2D array,
#	and @values stores the values to be analyzed later.
#
# INPUTS: (\*${$fh}, $indx_sbj_fld, $str_undef, $str_end), where
#	
#	$fh is the file-handle for the input file (recording the 2D array);
#
#	$indx_sbj_fld is the index of the subject field;
#	[The index starts with 0 for the row-label (i.e., "$x:$y").]
#
#	$str_undef is a string indicating that the subject fields is undefined.
#
#	$str_end is a string (at the biginning of the line)
#	that indicates the end of the record (for the subject 2D array);
#	it is undef if such a string does not exist.
#
# [ OUTPUTS: (\@array_2D), where ] ... for the predecessor, 'readin_2D_array'.
#
# OUTPUTS: (\@array_1D), where
#
#	@array_1D is the 1D array retrieved, where
#	$array_2D[$x] = $values[$indx_sbj_fld-1] if the subject value is defined,
#			    = undef if it is undefined.
#
# [ sub readin_2D_array (\*$$$) { ] ... for the predecessor, 'readin_2D_array'.
#
sub readin_1D_array (\*$$$) {

	my ($fh, $indx_sbj_fld, $str_undef, $str_end) = @_;

	local $/ = "\n"; # Set the input-record-separator to be "\n".

	my $len_str_end = (defined $str_end) ? length ($str_end) : 0;

	my @array_1D = ();

	while (my $in_line = <$fh>) { # Main while-loop (over the input lines). #

	    unless ($in_line =~ m/\S/) { next; } # Skip if the line contains no non-white-spaces. #

	    if ($len_str_end > 0) {
		my $sbstr = substr ($in_line, 0, $len_str_end);
		if ($sbstr eq $str_end) { last; }
	    }

			# Major Modification in creation of 'readin_1D_array'. (1) #
	    my ($x) = ($in_line =~ m/^(\d+)\t/);
	    unless (defined $x) { next; }
	    ##my ($x, $y) = ($in_line =~ m/^(\d+)\:(\d+)\t/);
	    ##unless (defined $y) { next; }
			# End of "Major Modification in creation of 'readin_1D_array'. (1)" #

	    $in_line =~ s/\s+$//; # Remove the white-spaces on the end of the line.

	    my @flds = split /\t/, $in_line;

	    unless (@flds > $indx_sbj_fld) { next; }

	    my $value = $flds[$indx_sbj_fld];

	    unless (defined $value) { next; }
	    if ($value eq $str_undef) { next; }


			# Major Modification in creation of 'readin_1D_array'. (2) #

		# Store the $value as the $x-element of @array_1D. #

	    $array_1D[$x] = $value;

=For the predecessor, 'readin_2D_array'.
		# Store the $value as the ($x, $y)-element of @array_2D. #

	    my $row_values = $array_2D[$x];
	    unless (defined $row_values) {
		$row_values = $array_2D[$x] = [];
	    }

	    $row_values->[$y] = $value;
=cut # End of "For the predecessor, 'readin_2D_array'.
			# END of "Major Modification in creation of 'readin_1D_array'. (2)" #

	} # End of the Main while-loop (over the input lines). #


	return \@array_1D;

}	# End of the subroutine, 'readin_1D_array'. #


# cmpt_dbl_cumulative_2Darray:
#	This subroutine receives a (possibly "loose") 2D array,
#	and computes its double-cumulative counterpart, which is defined below.
#	[CREATION STARTED on 2020/04/08, ENDED on 2020/04/09; DEBUGGED on 2020/04/11]
#
#	[DEFINITION]:
#	A double-cumulative 2D array, @dcum_array_2D, of a 2D-array, @array_2D,
#	is defined as:
#	$dcum_array_2D[$x]->[$y] = Sum_{$x2 = $x}^{$max_x} Sum_{$y2 = $y}^{$max_y} $array_2D[$x2]->[$y2],
#	where $max_x and $max_y are the "maximum" values of $x2 and $y2, respectively, 
#	which are specified by the user.
#
#	[Algorithm]:
#	To speed up the computation, the elements of the @dcum_array_2D will be computed by iteration
#	(from $x = $max_x to $x = 0 and from $y = $max_y to $y = 0):
#
#	$dcum_array_2D[$max_x]->[$max_y] = $array_2D[$max_x]->[$max_y],
#
#	$dcum_array_2D[$max_x]->[$y] = $array_2D[$max_x]->[$y] + $dcum_array_2D[$max_x]->[$y+1] (for $y < $max_y),
#
#	$dcum_array_2D[$x]->[$max_y] = $array_2D[$x]->[$max_y] + $dcum_array_2D[$x+1]->[$max_y] (for $x < $max_x),
#	and
#	$dcum_array_2D[$x]->[$y] = $array_2D[$x]->[$y]
#			 + $dcum_array_2D[$x]->[$y+1] + $dcum_array_2D[$x+1]->[$y]
#				 - $dcum_array_2D[$x+1]->[$y+1] (for $x < $max_x and $y < $max_y).
#
# 	[NOTE]: This subroutine regards the undefined elements of the input 2D array as 0 (zero).
#
#
# INPUTS: (\@array_2D, $max_x, $max_y), where
#
#	@array_2D is an input 2D array, which may possibly be "loose" (i.e., containing some undefined elements);
#	$array_2D[$x]->[$y] is the ($x, $y)-element of the 2D array.
#
#	$max_x and $max_y are the maximum values (inclusive) of the $x and $y, respectively.
#
# OUTPUTS: (\@dcum_array_2D), where
#
#	@dcum_array_2D is the 2D array storing the "double-cumulative" counterparts of the elements of @array_2D.
#	See the [DEFINITION] and [Algorithm] above for more details.
#
sub cmpt_dbl_cumulative_2Darray (\@$$) {

	my ($array_2D, $max_x, $max_y) = @_;

		# Initialize the output. #
	my @dcum_array_2D = ();
	for (my $x=0; $x <= $max_x; $x++) { $dcum_array_2D[$x] = []; }


		# Main Process: Compute the double-cumulative 2D array. #

	my $raw_at_max_x = $array_2D->[$max_x];
	my $dcum_at_max_x = $dcum_array_2D[$max_x];


			# At ($max_x, $max_y). #

	$dcum_at_max_x->[$max_y] = ((defined $raw_at_max_x) and (defined $raw_at_max_x->[$max_y])) ? $raw_at_max_x->[$max_y] : 0;


			# On $x = $max_x.

	for (my $y = $max_y-1; $y >= 0; $y--) { # 1st main for-loop (over $y). #

	    my $raw_element = 0;
	    if ((defined $raw_at_max_x) and (defined $raw_at_max_x->[$y])) {
		$raw_element = $raw_at_max_x->[$y];
	    }

	    $dcum_at_max_x->[$y] = $raw_element + $dcum_at_max_x->[$y+1];

	} # End of the 1st main for-loop (over $y). #


			# On $y = $max_y. #

	for (my $x = $max_x-1; $x >= 0; $x--) { # 2nd main for-loop (over $x). #

	    my $raw_at_x = $array_2D->[$x];
	    my $dcum_at_x = $dcum_array_2D[$x];
	    my $dcum_at_xpls1 = $dcum_array_2D[$x+1];

	    my $raw_element = 0;
	    if ((defined $raw_at_x) and (defined $raw_at_x->[$max_y])) {
		$raw_element = $raw_at_x->[$max_y];
	    }

	    $dcum_at_x->[$max_y] = $raw_element + $dcum_at_xpls1->[$max_y];

	} # End of the 2nd main for-loop (over $x). #


			# The remaining (bulk). #

	for (my $x = $max_x-1; $x >= 0; $x--) { # 3rd main outer for-loop (over $x). #

	    my $raw_at_x = $array_2D->[$x];
	    my $dcum_at_x = $dcum_array_2D[$x];
	    my $dcum_at_xpls1 = $dcum_array_2D[$x+1];

	    for (my $y = $max_y-1; $y >= 0; $y--) { # Inner for-loop (over $y). #

		my $raw_element = 0;

				# REVISED to Fix a BUG (on 2020/04/11). #
		##if ((defined $raw_at_x) and (defined $raw_at_x->[$max_y])) {
		##    $raw_element = $raw_at_x->[$max_y];
		##}
		if ((defined $raw_at_x) and (defined $raw_at_x->[$y])) {
		    $raw_element = $raw_at_x->[$y];
		}
				# End of "REVISED to Fix a BUG (on 2020/04/11)." #

		$dcum_at_x->[$y] = $raw_element
 			+ $dcum_at_x->[$y+1] + $dcum_at_xpls1->[$y]
				- $dcum_at_xpls1->[$y+1];

	    } # End of the inner for-loop (over $y). #

	} # End of the 3rd main outer for-loop (over $x). #


	return \@dcum_array_2D;

}	# End of the subroutine, 'cmpt_dbl_cumulative_2Darray'. #


1;		# a "spell" which makes the module "use"-able and/or "require"-able #


